/* $Id: mod_macs_atc.c,v 1.3 2002/04/09 15:38:14 santana Exp $
 *
 * An implementation of the atc piece of MACS. Starting off from a perl
 * module with the following comments
 *
 * ---------
 * WACA.pm
 *
 * This module is the heart of WACA and performs all the dificult "in
 * the server" type stuff, any serious problem here could crash the
 * whole server, so be careful!
 *
 * 05-04-99 Blake Mitchell
 * ---------
 *
 * Installation and Use
 * --------------------
 *
 * Make sure the mod_macs_atc.so is in your apache's modules/ dir,
 * that the following line is in httpd.conf:
 *   LoadModule macs_atc_module modules/mod_macs_atc.so
 *
 * Configure by specifying the MacsConfigFile directive in your base
 * and/or <virtualhost> sections, and the MacsResource directive in
 * your <dir>, <location>, or <file> sections:
 *   MacsConfigFile /path/to/cf
 *   <location /appone>
 *   MacsResource mycompany/appone
 *   </location>
 *   <location /appone/subdir/apptwo>
 *   MacsResource mycompany/apptwo
 *   </location>
 *   <location /appone/subdir/notamacsresource>
 *   MacsResource clear
 *   </location
 * The format of the MACS configuration file is out of the scope
 * of these comments, see the docs for MACS configuration.
 *
 * TODO
 * ----
 *
 * thread-safe for when that day comes (plus there'll be thread-safe
 *   atc code to reuse)
 */

 
/******************
 ** begin header **
 ******************/
 
/* identify yourself */
const char macs_atc_rcs_id[] = "$Id: mod_macs_atc.c,v 1.3 2002/04/09 15:38:14 santana Exp $";

/* apxs includes */
#include "httpd.h"
#include "http_log.h"
#include "http_config.h"
#include "http_core.h"
#include "http_protocol.h"
#include "ap_config.h"

#define MACS_ATC_BUFSIZ 1024
#define MACS_ATC_DEFAULT_PORT 0
#define MACS_ATC_DEFAULT_TIMEOUT 3

/* sockets to ats servers */
struct macs_atc_sock_struct {
  char *host;
  int port;
  int isup;
  int sock;
  int in_use;
  struct macs_atc_sock_struct *next;
};
typedef struct macs_atc_sock_struct macs_atc_sock;

/* server configuration data */
typedef struct {
  int port,              /* default port (not used)      */
    timeout;             /* timeout for various ops      */
  char *uidhdr;          /* header/cookie w/uid from auc (e.g.,
			    r->headers_in); also used as
			    MACSAuthenticate flag in r->notes */
  macs_atc_sock *socks;  /* linked list of ATC sockets   */
} macs_atc_cfg;

/* directory configuration data */
typedef struct {
  char *root_resource; /* the macs resource mapped to             */
  char *root_path;     /* the path mapped from                    */
  char *resource;      /* root_resource + (this_path - root_path) */
} macs_atc_dir_cfg;

/* atc initializations (logs a line) */
void macs_atc_init(server_rec *,pool *);

/* Initialize macs_atc_cfg
 * There won't be any *server* merging routines because if we give
 *   different servers different config files, we're plugging that
 *   server into a different, isolated instance of macs.
 */
void *macs_atc_svrcfg_init(pool *,server_rec *);
void *macs_atc_dircfg_init(pool *,char *);
void *macs_atc_dircfg_merge(pool *,void *,void *);

/* Configuration directive handler to deal with config file
 *   populate ats list
 *   connect and disconnect at child startup
 */
const char *macs_atc_cfdirective(cmd_parms *,void *,char *);

/* Parse a config line
 * errors into arg1 w/ rv != 0
 */
int macs_atc_parse_cfline(char *,cmd_parms *,macs_atc_cfg *);

/* Configuration directive handler to map this dir */
const char *macs_atc_map_dir(cmd_parms *,void *,const char *);

/* add an ATS
 * errors into arg1 w/ rv != 0
 */
int macs_atc_add_ats(char *,cmd_parms *,macs_atc_cfg *);
 
/* returns a null-initted macs_atc_sock */
macs_atc_sock *new_macs_atc_sock(pool *);
 
/* connect to the ATS specs */
void macs_atc_child_init(server_rec *,pool *);

/* close all ATS connections so garbage collection goes smoothly */
void macs_atc_child_exit(server_rec *,pool *);

/* if url is not mapped, return DECLINED
 * if no uid is found, return UNAUTHORIZED
 * unmap url and ask ATS (operation READ)
 * if ATS says yes(*), return OK
 * else return UNAUTHORIZED
 *
 * (*) If during this stage, no ATS's are up, we rerun
 *     macs_atc_child_init
 */
int macs_atc_authorize(request_rec *);

/* unmap uri to macs resource path */
int macs_atc_mkresource(macs_atc_dir_cfg *,const char *);

/* asks the ATS: CAN uid READ resource_path */
int macs_atc_query_ats( const char *, request_rec *,
			macs_atc_dir_cfg *, macs_atc_cfg * );

/****************
 ** end header **
 ****************/

/* Config directive */
static const command_rec macs_atc_commands[] = {
  { "MACSConfigFile", macs_atc_cfdirective, NULL, RSRC_CONF,  TAKE1,
    "MACS configuration file: MacsConfigFile /path/to/macs.cfg" },
  { "ATCConfigFile", macs_atc_cfdirective, NULL, RSRC_CONF,  TAKE1,
    "MACS ATC-specific configuration file: ATCConfigFile /path/to/atc.cfg" },
  { "MACSResource", macs_atc_map_dir, NULL, OR_AUTHCFG, TAKE1,
    "Map this directory to a MACS resource path: MACSResource root/path\nThe special resource `CLEAR' unmaps the current directory from any macs resource path." },
  {NULL}
};

/* Dispatch list for API hooks */
module MODULE_VAR_EXPORT macs_atc_module = {
  STANDARD_MODULE_STUFF, 
  macs_atc_init,         /* module initializer                  */
  macs_atc_dircfg_init,  /* create per-dir    config structures */
  macs_atc_dircfg_merge, /* merge  per-dir    config structures */
  macs_atc_svrcfg_init,  /* create per-server config structures */
  NULL,                  /* merge  per-server config structures */
  macs_atc_commands,     /* table of config file commands       */
  NULL,                  /* [#8] MIME-typed-dispatched handlers */
  NULL,                  /* [#1] URI to filename translation    */
  NULL,                  /* [#4] validate user id from request  */
  NULL,                  /* [#5] check if the user is ok _here_ */
  /* FIXME: this *should* be in #5, but that seems to be hard to work
     with */
  macs_atc_authorize,    /* [#3] check access by host address   */
  NULL,                  /* [#6] determine MIME type            */
  NULL,                  /* [#7] pre-run fixups                 */
  NULL,                  /* [#9] log a transaction              */
  NULL,                  /* [#2] header parser                  */
  macs_atc_child_init,   /* child_init                          */
  macs_atc_child_exit,   /* child_exit                          */
  NULL                   /* [#0] post read-request              */
};

void macs_atc_init(server_rec *s,pool *p) {
  ap_add_version_component("macs/atc");
} /* macs_atc_init */


/*****************************************************************
                  config block initializers
 *****************************************************************/
void *macs_atc_svrcfg_init(pool *p, server_rec *s) {
  macs_atc_cfg *c = (macs_atc_cfg *)ap_palloc(p,sizeof(macs_atc_cfg));
  c->port = MACS_ATC_DEFAULT_PORT;
  c->timeout = MACS_ATC_DEFAULT_TIMEOUT;
  c->uidhdr = NULL;
  c->socks = NULL;
  return (void *)c;
} /* macs_atc_svrcfg_init */

void *macs_atc_dircfg_init(pool *p, char *dir) {
  macs_atc_dir_cfg *dc = (macs_atc_dir_cfg *)ap_palloc(p,sizeof(macs_atc_dir_cfg));
  dc->root_resource = dc->root_path = dc->resource = NULL;
  fprintf(stderr, "\natc asked to init dircfg for %s)\n\n", dir);
  return (void *)dc;
} /* macs_atc_dircfg_init */

void *macs_atc_dircfg_merge(pool *p, void *base, void *new) {
  macs_atc_dir_cfg *b = (macs_atc_dir_cfg *)base,
    *n = (macs_atc_dir_cfg *)new;
  /* FIXME: how to log w/o access to server or request record? */
  /*    ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, cd->r->server,
		 "asked to merge parent (root_path:%s, root_resource:%s, resource:%s) with child (root_path:%s, root_resource:%s, resource:%s)",
		 b->root_path,b->root_resource,b->resource,
		 n->root_path,n->root_resource,n->resource ); */
  fprintf(stderr, "\natc asked to merge parent (root_path:%s, root_resource:%s, resource:%s) with child (root_path:%s, root_resource:%s, resource:%s)\n\n",
	  b->root_path,b->root_resource,b->resource,
	  n->root_path,n->root_resource,n->resource );

    return new;
} /* macs_atc_dircfg_merge */


/*****************************************************************
             directive handlers & co
 *****************************************************************/

/* make this directory/location secure (or not) */
const char *macs_atc_map_dir( cmd_parms *parms, void *mconfig,
			      const char *resource ) {
  macs_atc_dir_cfg *dc = (macs_atc_dir_cfg *)mconfig;
  if ( parms->path == NULL || mconfig == NULL ) {
    return "MACSResource is only valid in a directory context";
  }
  if ( !strcasecmp(resource,"clear") ) {
    ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, parms->server,
		 "Unmapping path %s from resource %s",
		 parms->path, resource);
    return dc->root_resource = dc->root_path = dc->resource = NULL;
  }
  dc->root_resource = ap_pstrdup(parms->pool,resource);
  dc->root_path = ap_pstrdup(parms->pool,parms->path);
  ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, parms->server,
	       "Mapping path %s to resource %s",
	       parms->path, resource);
  return NULL;
} /* macs_atc_map_dir */

/* go through the macs-style config file */
const char *macs_atc_cfdirective( cmd_parms *parms, 
				  void *mconfig, 
				  char *cfname ) {
  FILE *cf;
  macs_atc_cfg *c;
  char *line;

  c = (macs_atc_cfg *)
    ap_get_module_config(parms->server->module_config,&macs_atc_module);
  /* we should inherit a macs_atc_cfg from the server-init by now */
  if (!c) return "Internal error initializing configs";

  cf = ap_pfopen(parms->temp_pool,cfname,"r");
  line = (char *)ap_palloc(parms->temp_pool,sizeof(char)*MACS_ATC_BUFSIZ);
  if (!cf) {
    ap_snprintf(line,MACS_ATC_BUFSIZ,"Error opening %s: %s",
		cfname,strerror(errno));
    return line;
  }
  while (fgets(line,MACS_ATC_BUFSIZ,cf)) {
    switch (macs_atc_parse_cfline(line,parms,c)) {
    case 0: /* good */
      break;
    default: /* bad, errs in line */
      return line;
      break;
    } /* switch parse line */
  } /* while fgets */
  if ( ferror(cf) && !feof(cf) ) {
    /* bad */
    ap_snprintf(line,MACS_ATC_BUFSIZ,"Error reading from %s: %s",
		cfname,strerror(errno));
    return line;
  }
  return (char *)NULL;
} /* macs_atc_parse_cf */

/* changes the config block, or feeds a line to add_ats */
int macs_atc_parse_cfline(char *line, cmd_parms *parms, macs_atc_cfg *c) {
  char *cdir, *val;

  cdir = ap_getword_white_nc(parms->temp_pool,&line);
  if (cdir[0] == '#')
    return 0; /* skip comments */
  /* there *has* to be a better way to do this, but this isn't a
   *   speed-sensitive section...
   * for now we'll just check each directive against the ones we care
   *   about
   */
  /* ATS host:port  OR  ATS host port */
  if (!strcasecmp(cdir,"ATS")) {
    val = ap_getword_white_nc(parms->temp_pool,&line);
    if (!val) {
      ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, parms->server,
		   "Null ATS spec in config: %s",line);
      return 1;
    }
    if ( ( *line ) &&
	 ( cdir = ap_getword_white_nc(parms->temp_pool,&line) ) &&
	 ( cdir[0] != '#' ) )
      val = ap_pstrcat(parms->temp_pool,val,":",cdir,NULL);
    return macs_atc_add_ats(val,parms,c);
  }
  if (!strcasecmp(cdir,"HTTP_UID_HEADER")) {
    c->uidhdr = ap_getword_conf_nc(parms->pool,&line);
    if (!c->uidhdr) {
      ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, parms->server,
		   "Null uid header in config: %s",line);
      return 1;
    }
    ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, parms->server,
		 "The emulated cookie&header name is %s\n",c->uidhdr);
    return 0;
  }
  if (!strcasecmp(cdir,"TIMEOUT")) {
    val = ap_getword_conf_nc(parms->pool,&line);
    if (!val) {
      ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, parms->server,
		   "Null timeout spec in config: %s",line);
      return 1;
    }
    c->timeout = atoi(val);
    ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, parms->server,
		 "The general timeout is %d seconds",c->timeout);
    return 0;
  }
  ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, parms->server,
	       "Skipping irrelevant config directive: %s",cdir);
  return 0;
} /* macs_atc_parse_cfline */


/*****************************************************************
                     config-block modifiers
 *****************************************************************/

/* adds an ats to the config block */
int macs_atc_add_ats(char *hostparm, cmd_parms *parms, macs_atc_cfg *c) {
  char *host, *charpt;
  long port;
  macs_atc_sock *curr;
  /* parse hostparm */
  charpt = host = ap_pstrdup(parms->temp_pool,hostparm);
  if (!host) {
    /* malloc error */
    hostparm--;
    ap_snprintf(hostparm,MACS_ATC_BUFSIZ,"Memory allocation error: %s",
		strerror(errno));
    return -1;
  }
  while (charpt && *charpt != ':') charpt++;
  if (charpt) {
    *charpt = (char)NULL;
    port = atol(++charpt);
  } else {
    port = 4480;
    ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, parms->server,
		 "No port in config line: %s",hostparm);
  }
  if (!c->socks) {
    /* first item in list */
    curr = c->socks = new_macs_atc_sock(parms->pool);
  } else {
    /* append an item to list */
    for (curr=c->socks; curr->next; curr=curr->next) ;
    curr = curr->next = new_macs_atc_sock(parms->pool);
  }
  if (!curr) {
    /* malloc error */
    hostparm--;
    ap_snprintf(hostparm,MACS_ATC_BUFSIZ,"Memory allocation error: %s",
		strerror(errno));
    return -1;
  }
  /* init the bare necessities; the client init actually connects */
  curr->host = ap_pstrdup(parms->pool,host);
  if (!curr->host) {
    /* malloc error */
    hostparm--;
    ap_snprintf(hostparm,MACS_ATC_BUFSIZ,"Memory allocation error: %s",
		strerror(errno));
    return -1;
  }
  curr->port = port;
  ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, parms->server,
	       "Configured ATS: %s:%d",curr->host,curr->port);

  return 0;
} /* macs_atc_add_ats */

macs_atc_sock *new_macs_atc_sock(pool *p) {
  macs_atc_sock *s = (macs_atc_sock *)ap_palloc(p,sizeof(macs_atc_sock));
  s->host = NULL;
  s->port = 0;
  s->isup = 0;
  s->sock = -1;
  s->in_use = 0;
  s->next = NULL;
  return s;
} /* new_macs_atc_sock */


/*****************************************************************
                       child startup/exit
 *****************************************************************/

/* connect to ATSs on child startup (and at need) */
void macs_atc_child_init(server_rec *r, pool *p) {
  macs_atc_cfg *c = (macs_atc_cfg *)
    ap_get_module_config(r->module_config, &macs_atc_module);    
  macs_atc_sock *s;
  struct hostent *h;
  struct sockaddr_in addr;
  unsigned int host_addr;
  /* we don't get any real tmp pools, so
     we'll just do this buffer by hand */
  char *buf = (char *)malloc(sizeof(char)*MACS_ATC_BUFSIZ);
  /* I hate inheriting this... */
  errno = 0;
  /* multi-process rather than multi-threaded...  this might change
   * someday, but til then we don't want each child to connect to
   * every ATS. One is probably enough.
   * 
   * FIXME: there should be a way to evenly spread which server entry
   * the children connect to. Right now, the first entry gets hit til
   * it won't accept any more, then the second, etc.
   */  
  for (s = c->socks; s; s = s->next) {
    if ( s->in_use ) {
      /* in use means ok */
      ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		   "ATS %s:%i in use",s->host,s->port);
      continue;
    }
    s->in_use = 1;
    errno = 0;
    /* let's just ping and see if we get a response */
    if ( s->isup && ! ( write(s->sock,"PING\n",5) == 5   &&
			read(s->sock,buf,MACS_ATC_BUFSIZ)     &&
			strncasecmp(buf,"PONG\n",5) == 0 ) ) {
      ap_log_error(APLOG_MARK, APLOG_DEBUG, r,
		   "ATS %s:%i down", s->host, s->port);
      s->isup = 0;
    } else if ( s->isup ) {
      ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		   "Already connected to ATS %s:%i",s->host,s->port);
      s->in_use = 0;
      /* this shouldn't happen in multi-process mode */
      break;
    }
    /* if down, bring back up */
    close(s->sock);
    if ( !(h = gethostbyname(s->host)) ) {
      ap_log_error(APLOG_MARK, APLOG_WARNING, r,
		   "Couldn't get hostname for %s", s->host);
      s->in_use = 0;
      continue;
    }
    host_addr = *(unsigned int *)h->h_addr;
    addr.sin_family = AF_INET;
    addr.sin_port = htons(s->port);
    addr.sin_addr.s_addr = host_addr;
    s->sock = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);
    if ( s->sock < 0 ) {
      ap_log_error(APLOG_MARK, APLOG_WARNING, r,
		   "Couldn't open socket to %s:%i", s->host,s->port);
      s->in_use = 0;
      continue;
    }
    if ( connect(s->sock,(struct sockaddr *)&addr,sizeof(addr)) < 0 ) {
      ap_log_error(APLOG_MARK, APLOG_DEBUG, r,
		   "Failed to connect to ATS %s:%i", s->host, s->port);
      close(s->sock);
      s->in_use = 0;
      continue;
    }
    /* let's just ping and see if we get a response */
    if ( s->isup && ! ( write(s->sock,"PING\n",5) == 5   &&
			read(s->sock,buf,MACS_ATC_BUFSIZ)     &&
			strncasecmp(buf,"PONG\n",5) == 0 ) ) {
      ap_log_error(APLOG_MARK, APLOG_WARNING, r,
		   "Communication with ATS %s:%i failed", s->host, s->port);
      s->in_use = 0;
      continue;
    } else {
      ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		   "Connected to ATS %s:%i", s->host, s->port);
      s->isup = 1;
      s->in_use = 0;
      /* we got one, that's enough */
      break;
    }
  } /* foreach socket */
  free(buf);
  return;
} /* macs_atc_child_init */

void macs_atc_child_exit(server_rec *r, pool *p) {
  macs_atc_cfg *c = (macs_atc_cfg *)
    ap_get_module_config(r->module_config, &macs_atc_module);    
  macs_atc_sock *s;
  for (s = c->socks; s; s = s->next) {
    errno = 0;
    if (s->sock)
      close(s->sock);
    ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		 "Disconnecting from %s:%i", s->host, s->port);
    s->in_use = 1;
    s->isup = 0;
  }
  return;
} /* macs_atc_child_exit */


/*****************************************************************
          the actual authorizaton routines (finally)
 *****************************************************************/
int macs_atc_authorize(request_rec *r) {
  const char *uid;
  macs_atc_cfg *c;
  macs_atc_dir_cfg *dc;
  /*if ( r->main && r->main != r ) {
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		  "ignoring subrequest %s (main request %s)",
		  r->uri, r->main->uri);
    return DECLINED;
  } */
  dc = (macs_atc_dir_cfg *)
    ap_get_module_config(r->per_dir_config, &macs_atc_module);
  c = (macs_atc_cfg *)
    ap_get_module_config(r->server->module_config, &macs_atc_module);
  if ( !ap_table_get(r->notes,c->uidhdr) ) {
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		  "MACSAuthenticate off (no %s note), no authorization",
		  c->uidhdr);
    return OK;
  }
  if ( !(uid = ap_table_get(r->headers_in,c->uidhdr)) ) {
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		  "no uid, but MACSAuthenticate is on. problem with auc?");
    return DECLINED;
  }
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		"atzing %s w/ dircfg (root_path:%s, root_resource:%s, resource:%s)",
		r->uri,dc->root_path,dc->root_resource,dc->resource);
  if ( !dc->resource ) {
    /* calculate macs resource for this uri */
    if ( !macs_atc_mkresource(dc,r->uri) ) {      
      ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
		    "failed to make a resource");
      return SERVER_ERROR;
    }
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		  "resource now %s", dc->resource);
  }
  /* prelims done, check against ATS */
  return macs_atc_query_ats(uid,r,dc,c);
} /* macs_atc_authorize */

int macs_atc_mkresource(macs_atc_dir_cfg *dc, const char *uri) {
  if ( strncmp(dc->root_path,uri,strlen(dc->root_path)) ) {
    fprintf(stderr,"\nuri %s not a subdir of dcroot %s\n\n",
	    uri,dc->root_path); 
    return 0;
  }
  dc->resource = malloc( sizeof(char)* (strlen(dc->root_resource)+
					strlen(uri)-strlen(dc->root_path)
					+ 1) );
  strcpy( dc->resource, dc->root_resource );
  strcat( dc->resource, &uri[strlen(dc->root_path)] );
  return 1;
} /* macs_atc_mkresource */

int macs_atc_query_ats( const char *uid, request_rec *r,
			macs_atc_dir_cfg *dc, macs_atc_cfg *c ) {
  macs_atc_sock *s = c->socks;
  int upsvrs, rv;
  char *query = (char *)ap_palloc(r->pool,sizeof(char)*MACS_ATC_BUFSIZ),
    *reply;
 
  /* create the query */
  rv = ap_snprintf(query,MACS_ATC_BUFSIZ,"CAN %s READ %s\n",
		   uid, dc->resource);
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		"query will be %s",query);
  /* if there are servers up, keep trying them until one shows up
   * unused, then grab it
   */
  for (upsvrs=0; s||upsvrs; s=s->next) {
    if ( !s ) {
      /* start over */
      s = c->socks;
      upsvrs = 0;
    }
    if ( s->isup )
      upsvrs++;
    else
      continue;
    if ( !s->in_use ) {
      s->in_use = 1;
      break;
    }
  }
  if ( !s ) {
    /* no ATS's up! */
    if ( ap_table_get(r->notes,"macs_atc_childreinit") ) {
      ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
		    "No ATS's up after reinit!");
      return SERVER_ERROR;
    }
    ap_log_rerror(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, r,
		  "No ATS's up! Reinitting");
    ap_table_setn(r->notes,"macs_atc_childreinit","1"); /* only reinit once */
    macs_atc_child_init(r->server,(pool *)NULL);
    return macs_atc_query_ats(uid,r,dc,c);
  }
 
  /* send the query to the ATS */
  /* FIXME: infinite loop possible, e.g. in case of network problems? */
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		"writing query");
  if ( (upsvrs=write(s->sock,query,rv)) != rv ) {
    ap_log_rerror(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, r,
		  "Retry on error writing to ATS %s:%i: %s",
		  s->host, s->port, strerror(errno));
    s->isup = 0;
    s->in_use = 0;
    return macs_atc_query_ats(uid,r,dc,c);
  }
 
  /* get the reply */
  rv = upsvrs = errno = 0;
  reply = (char *)ap_palloc(r->pool,sizeof(char)*MACS_ATC_BUFSIZ);
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		"reading reply");
  while (1) {
    /* upsvrs := bytes read */
    upsvrs = read(s->sock,&reply[rv],MACS_ATC_BUFSIZ-rv);
    if ( upsvrs > 0 )
      rv += upsvrs;
    if ( reply[rv-1] == '\n' ) {
      break;
    }
    switch (errno) {
    case 0: /* fine, read more */
      break;
    case EAGAIN: /* temporary error condition - clear & read more */
      errno = 0;
      break;
    case ETIMEDOUT: /* timed out waiting for response */
      ap_log_rerror(APLOG_MARK, APLOG_WARNING, r,
		    "Timeout waiting for ATS %s:%i", s->host, s->port);
    default: /* read error */
      ap_log_rerror(APLOG_MARK, APLOG_WARNING, r,
		    "Retry on read error from ATS %s:%i", s->host, s->port);
      s->isup = 0;
      s->in_use = 0;
      return macs_atc_query_ats(uid,r,dc,c);
    }
  } /* read reply loop */
  s->in_use = 0;

  ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		"reply is .%s.",reply);
  
  if ( !strncasecmp("YES",reply,3) ) {
    /* strip that pesky newline */
    for ( query=reply; *query; query++ )
      if ( *query=='\n' )
	*query=(char)NULL;
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		  "Successful authorization for uid %s, resource %s",
		  uid, dc->resource);
    return OK;
  }
  if ( !strncasecmp("NO\n",reply,3) ) {
    /* no good */
    ap_log_rerror(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, r,
		  "Unsuccessful authorizaton for uid %s, resource %s",
		  uid, dc->resource);
    return HTTP_UNAUTHORIZED;
  }
  /* error string from server */
  ap_log_rerror(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, r,
		"Error string from ATS: %s", reply);
  return SERVER_ERROR;
} /* macs_atc_query_ats */
