#!/usr/bin/perl -w
# $Id: ATSd.pl,v 1.9 2002/04/14 08:42:34 blake Exp $
#

package ATSd;
use strict;
use lib "../lib";
use Tie::DistHash;
use Net::muxDaemon;
use URI::Escape;
use Text::ParseWords;
# retrieve all the info from the SQL DB
use DBI qw(:sql_types);

use constant PATHSEP => '/';

# the permissions flags
use constant DISABLED => 0x01;	# 00000001
use constant OWNER => 0x02;	# 00000010
use constant ADMIN => 0x04;	# 00000100
use constant WRITE => 0x08;	# 00001000
use constant READ => 0x10;	# 00010000

use constant ALL_PERMS => DISABLED|OWNER|ADMIN|WRITE|READ;


#
# Get configured
use Config::IniFiles;
my($cfg,%cfg);
die("Couldn't configure!")
  unless $cfg = tie(%cfg, 'Config::IniFiles',
		    -file => '../macs.cfg',
		    -default => 'default');
for ( keys %{$cfg{env}} ) {
  $ENV{$_} = $cfg{env}{$_};
}


#
# write a config file for the auc
open(CCF, '>../atc.cfg');
print CCF 'HTTP_UID_HEADER ',$cfg{AUS}{http_uid_header},"\n";
print CCF 'TIMEOUT ',$cfg{AUS}{timeout},"\n";
#
# The hosts in our distributed setup
my($hosts) = [];
for ( $cfg->val('ATS','p') ) {
  print CCF 'ATS ',$_,"\n";
  push(@$hosts, {addr => $_});
}
close CCF;

# make sure we're running in the background
FORK: {
  # the process id variable
  my($pid);
  # spawn a child
  if ($pid = fork()) {
    # parent process here, bye bye
    exit(0);
  } elsif (defined $pid) {
    # child process here, we continue
    last FORK;
  } elsif ($! =~ m/[Nn]o more process/) {
    # can't fork, sleep and try again
    sleep(2);
    redo FORK;
  } else {
    # unrecoverable fork error
    die("Couldn't fork: $!\n");
  }
}
# has the muxDemon opened syslog yet?
my($syslog_open) = 0;

# db vars
my( $dbh, $lsgroups_sth, $lsroots_sth, $get_gperm_sth,
    $lschildren_sth, $groups_of_uid_sth, $groups_of_gid_sth );

#
# the permissions hashes
my($gperms) = {};
#my($gperms_h) = tie( %$gperms, 'Tie::DistHash',
#		     ckptfile => "ATSgperms.dat",
#		     sync => $hosts );
my($uperms) = {};
#my($uperms_h) = tie( %$uperms, 'Tie::DistHash',
#		     ckptfile => "ATSuperms.dat",
#		     sync => $hosts );

# connect to the database
db_setup();

# starting net::demon, syslog will be open
$syslog_open = 1;
# vroom...
new Net::muxDaemon( POOL => $hosts,
		    INITHOOK => \&build_gperms,
		    LOGTAG => "ATSd",
		    PIDFILE => "ATSd.pid" );
die("Net::muxDaemons don't return!\n\t$!\n\t$@\n");


###################
# Initializations #
###################

# set up the database connection
sub db_setup {
  syslog('debug','Initializing gperms');
  # make sure the database isn't already connected
  eval { $dbh->disconnect(); };
  # create the database handle
  $dbh = DBI->connect($cfg{db}{d}, $cfg{db}{l}, $cfg{db}{p});
  # make sure the database is connected
  unless($dbh) {
    if ($syslog_open) {
      syslog("err", "Couldn't connect: $DBI::errstr");
      $KILLED = 1;
    } else {
      die("Couldn't connect: $DBI::errstr");
    }
  }
  # the database should wine if there are any errors
  $dbh->{RaiseError} = 1;
  # prepare the query to retrieve group information
  $@ = "";
  # trap errors
  eval {
    $lsgroups_sth = $dbh->prepare($cfg{'sql ATS'}{list_groups});
    $lsroots_sth = $dbh->prepare($cfg{'sql ATS'}{list_roots});
    $get_gperm_sth = $dbh->prepare($cfg{'sql ATS'}{get_group_perms});
    $lschildren_sth = $dbh->prepare($cfg{'sql ATS'}{get_child_nodes});
    $groups_of_uid_sth = $dbh->prepare($cfg{'sql ATS'}{groups_of_uid});
    $groups_of_gid_sth = $dbh->prepare($cfg{'sql ATS'}{groups_of_gid});
  };
  # exit if there are errors
  if ($@) {
    if ($syslog_open) {
      syslog("err", "$@");
      $KILLED = 1;
    } else {
      die("$@");
    }
  }
}				# db_setup

# build the groups' permissions hash
sub build_gperms {
  my($groups,$roots);
  eval {
    $lsgroups_sth->execute();
    # $groups->[x] := [ gid, name ]
    $groups = $lsgroups_sth->fetchall_arrayref();
    $lsroots_sth->execute();
    # $roots->[x] := [ node_id, name ]
    $roots = $lsroots_sth->fetchall_arrayref();
  };
  # exit if there are errors
  if ($@) {
    if ($syslog_open) {
      syslog("err", "$@");
      $KILLED = 1;
    } else {
      die("$@");
    }
  }

  for (@$roots) {
    walk_gperms($_,$groups);
  }
} # sub build_gperms

# walks a hierarchy, filling in group perms info
sub walk_gperms {
  my($node,$groups,$parent_name) = (shift(),shift(),shift());
  syslog("debug","Working on $node->[1]");
  # if $node is not a root node...
  if(defined($parent_name)) {
    syslog("debug","(child of $parent_name)");
    # prepend a separation character to the name (BAM: why?)
    # FIXME: the separatin character should be drawn from the config file
    $node->[1] = '/' . $node->[1];
  } else {
    syslog("debug",'(root node)');
    # set the parent node name to an empty string
    $parent_name = '';
  }
  my($children,$parent_perms);
  eval {
    foreach my $group (@$groups) {
      # set up all the group perms for this path
      $get_gperm_sth->bind_param(1, $group->[0], { TYPE => SQL_INTEGER });
      $get_gperm_sth->bind_param(2, $node->[0], { TYPE => SQL_INTEGER });
      $get_gperm_sth->execute();
      # $gp->[x] := [ disabled,is_owner,is_admin,can_read,can_write ]
      my($gp) = $get_gperm_sth->fetchall_arrayref();
      # there should only be one entry per group per node
      die('Duplicate gid/resource_id in getting gperms')
	if $#$gp > 0;
      syslog("debug","group " . $group->[1] . "(" . $group->[0] . ")");
      # if there is a parent...
      if ( $parent_name ) {
	# get it's permissions
	$parent_perms = $gperms->{$group->[0]}{$parent_name};
	die "no parent perms $group->[0]:$parent_name"
	  unless defined($parent_perms);
	syslog("debug",'parent_perms ' .
	       expand_perms($parent_perms) .
	       "($parent_perms)");
      } else {
	$parent_perms = 0;
      }
      if ( $#$gp == -1 ) {
	# no perms defined, just inherit
	syslog("debug",
	       "inheriting $parent_perms (into $group->[0]:$parent_name:$node->[1])");
	$gperms->{$group->[0]}{$parent_name.$node->[1]} = $parent_perms;
      } else {
	# inherit and override
	syslog("debug",
	       "overriding $parent_perms (into $group->[0]:$parent_name:$node->[1]) ");
	$gperms->{$group->[0]}{$parent_name.$node->[1]} =
	  mkbitfld($gp->[0],$parent_perms);
	syslog("debug",
	       "becoming $gperms->{$group->[0]}{$parent_name.$node->[1]} :" .
	       expand_perms($gperms->{$group->[0]}{$parent_name.$node->[1]}) .
	       ":");
      }
    }
    $lschildren_sth->bind_param(1, $node->[0], { TYPE => SQL_INTEGER });
    $lschildren_sth->execute();
    # $children->[x] := [ node_id, name ]
    $children = $lschildren_sth->fetchall_arrayref();
  };
  # exit if there are errors
  if ($@) {
    if ($syslog_open) {
      syslog("err", "$@");
      $KILLED = 1;
    } else {
      die("$@");
    }
  }
  # I know, recursion sucks, but this won't be too deep (10? 50?)
  # BAM: recursion rocks!
  for (@$children) {
    walk_gperms($_, $groups, $parent_name.$node->[1]);
  }
} # sub walk_gperms

# makes a bitmask out of the array of permissions
#   (in the db, 0 is no change, 1 is on, 2 is off)
sub mkbitfld {
  my($perms_ary, $parent_perms) = (shift(),shift());

  if ( $perms_ary->[0] == 1 ) {
    #print (STDERR "DISABLED on ");
    $parent_perms |= DISABLED;
    #print (STDERR "($parent_perms) ");
  } elsif ( $perms_ary->[0] == 2 ) {
    #print (STDERR "DISABLED off ");
    # unary ~ is bitwise not
    $parent_perms &= ~DISABLED;
    #print (STDERR "($parent_perms) ");
  }

  if ( $perms_ary->[1] == 1 ) {
    #print (STDERR "OWNER on ");
    $parent_perms |= OWNER;
    #print (STDERR "($parent_perms) ");
  } elsif ( $perms_ary->[1] == 2 ) {
    #print (STDERR "OWNER off ");
    $parent_perms &= ~OWNER;
    #print (STDERR "($parent_perms) ");
  }

  if ( $perms_ary->[2] == 1 ) {
    #print (STDERR "ADMIN on ");
    $parent_perms |= ADMIN;
    #print (STDERR "($parent_perms) ");
  } elsif ( $perms_ary->[2] == 2 ) {
    #print (STDERR "ADMIN off ");
    $parent_perms &= ~ADMIN;
    #print (STDERR "($parent_perms) ");
  }

  if ( $perms_ary->[3] == 1 ) {
    #print (STDERR "READ on ");
    $parent_perms |= READ;
    #print (STDERR "($parent_perms) ");
  } elsif ( $perms_ary->[3] == 2 ) {
    #print (STDERR "READ off ");
    $parent_perms &= ~READ;
    #print (STDERR "($parent_perms) ");
  }

  if ( $perms_ary->[4] == 1 ) {
    #print (STDERR "WRITE on ");
    $parent_perms |= WRITE;
    #print (STDERR "($parent_perms) ");
  } elsif ( $perms_ary->[4] == 2 ) {
    #print (STDERR "WRITE off ");
    $parent_perms &= ~WRITE;
    #print (STDERR "($parent_perms) ");
  }

  return $parent_perms;
}				# sub mkbitfld

# expand the perms to the supplied fh
sub expand_perms {
  my($p) = shift();
  my($count,$r) = (0,'');
  if ( $p & DISABLED ) {
    $r .= '|' if $count;
    $r .= 'DISABLED';
    $count++;
  }
  if ( $p & OWNER ) {
    $r .= '|' if $count;
    $r .= 'OWNER';
    $count++;
  }
  if ( $p & ADMIN ) {
    $r .= '|' if $count;
    $r .= 'ADMIN';
    $count++;
  }
  if ( $p & READ ) {
    $r .= '|' if $count;
    $r .= 'READ';
    $count++;
  }
  if ( $p & WRITE ) {
    $r .= '|' if $count;
    $r .= 'WRITE';
    $count++;
  }
  return ($count ? $r : 'NONE');
}				# sub expand_perms


################
# CAN handling #
################

# returns an array of gids this uid is in
sub get_groups_of {
  my($uid) = shift();
  my($g) = [];
  my($groups);
  # get user's groups
  eval {
    $groups_of_uid_sth->bind_param(1, $uid, { TYPE => SQL_INTEGER });
    $groups_of_uid_sth->execute();
    $groups = $groups_of_uid_sth->fetchall_arrayref();
  };
  # warn about errors
  if ($@) {
    syslog("err", "$@");
    return @$g
  }

  for (@$groups) {
    push(@$g,$_->[0]);
  }
  # get groups those groups are in
  for (@$g) {
    get_child_groups($g,$_);
  }
  return @$g;
} # get_groups_of

# appends to the list of groups all groups those gid is in, making
# sure not to recurse infinitely by refusing to process groups already
# in the list.
sub get_child_groups {
  my($g,$gid) = (shift(),shift());
  my($groups);
  eval {
    $groups_of_gid_sth->bind_param(1, $_, { TYPE => SQL_INTEGER });
    $groups_of_gid_sth->execute();
    $groups = $groups_of_gid_sth->fetchall_arrayref();
  };
  # warn about errors
  if ($@) {
    syslog("err", "$@");
    return @$g
  }
  # check each child group
 GROUP: for my $n (@$groups) {
    # skip it if it's already in the group list
    for my $p (@$g) {
      next GROUP if $p == $n->[0];
    }
    # add the group too the group list
    push(@$g,$n->[0]);
    # find it's children
    get_child_groups($g,$n->[0]);
  }
} # sub get_child_groups

#
#
#
sub process_command {
  my($this,$buf) = (shift(),shift());
  if ( $buf =~ m/^\s*$/ ) {
    syslog("debug","Exiting process_command: OK NOOP");
    return "OK NOOP\n";
  }  
  # used all over
  my($reply, $operation);
  # the operation to perform
  ($operation,$buf) = split(' ', $buf, 2);
  $operation = uc($operation);
  $buf =~ s/^\s+|\s+$//g;	# eat {lead,trail}ing whitespace

  syslog("debug","Entering process_command: $operation $buf");

 OP: {

    # CAN uid operation resource
    if ($operation eq "CAN") {
      my($uid, $op, $path) = split(' ', $buf, 3);
      unless ( defined($uid) && defined($op) && defined($path) ) {
	$reply = 'ERR CAN: not enough arguments';
	last OP;
      }
      my($perm);
      unless (exists($uperms->{$uid}) && defined($uperms->{$uid})) {
	$reply = 'ERR CAN: unknown uid';
	last OP;
      }
      my(@path) = parse_line('/',0,$path);
      my($count);
      syslog("debug","Looking through $path (",join(',',@path),"):");
      for ($count = $#path;
	   $count && !exists($uperms->{$uid}{$path});
	   $count--) {
	pop(@path);
	$path=join('/',@path);
	syslog("debug","Shortening to $path");
      }
      unless (defined($perm = $uperms->{$uid}{$path})) {
	$reply = 'ERR CAN: unknown path';
	last OP;
      }
      if ($perm & DISABLED) {
	$reply = 'NO disabled';
	last OP;
      }
      if ($op =~ /READ/i) {
	$op = READ;
      } elsif ($op =~ /WRITE/i) {
	$op = WRITE;
      } elsif ($op =~ /ADMIN/i) {
	$op = ADMIN;
      } elsif ($op =~ /OWN/i) {
	$op = OWNER;
      } else {
	$reply = "ERR CAN: invalid operation: $op";
	last OP;
      }
      $reply = ($perm & $op ? 'YES' : 'NO');
    }

    # LOGIN uid
    elsif ($operation eq 'LOGIN') {
      my($uid) = $buf;
      unless (defined($uid)) {
	$reply = 'ERR LOGIN: not enough arguments';
	last OP;
      }
      unless ($uid =~ /^[0-9]+$/) {
	$reply = 'ERR LOGIN: uid not valid';
	last OP;
      }
      syslog("debug","logging in UID: $uid");
      my(@groups) = get_groups_of($uid);
      unless ( @groups ) {
	$reply = 'NO user not in any groups';
	last OP;
      }
      # bitwise _OR_ the permissions for each group to get all user permissions
      for my $g (@groups) {
	syslog("debug","for GID: $g");
	for my $p (keys %{$gperms->{$g}}) {
	  syslog("debug","found permissions for path: $p");
	  $uperms->{$uid}{$p} = 0
	    unless exists($uperms->{$uid}{$p});
	  $uperms->{$uid}{$p} |= $gperms->{$g}{$p};
	}
      }
      $reply = 'OK '.join(' ',@groups);
    }

    # LOGOUT uid
    elsif ( $operation eq 'LOGOUT' ) {
      my($uid) = $buf;
      unless ( defined($uid) ) {
	$reply = 'ERR LOGOUT: not enough arguments';
	last OP;
      }
      unless (exists($uperms->{$uid})) {
	$reply = "ERR LOGOUT: user not logged in: $uid";
	last OP;
      }
      delete $uperms->{$uid};
      $reply = 'OK';
    }

    # PING
    elsif ( $operation eq "PING" ) {
      $reply = "PONG";
    }

    # EVAL
    # **DANGEROUS**
    elsif ( $operation eq "EVAL" ) {
      $reply = "OK ";
      $reply .= eval $buf;
    }

    # QUIT
    elsif ( $operation eq "QUIT" ) {
      unlink("AppCache.pid");
      $reply = "OK";
      $KILLED = 1;
    }

    # menospeakie
    else {
      $reply = "ERR unknown command: $operation $buf";
    }

  } # OP

  syslog("debug","Exiting process_command: $reply");
  return "$reply\n";
} # sub process_command
