/* $Id: mod_macs_auc.c,v 1.5 2002/04/09 15:41:23 santana Exp $
 *
 * 12.99,
 * .mario
 *
 * An implementation of the auc piece of MACS. Starting off from a perl
 * module with the following comments
 *
 * ---------
 * WACA.pm
 *
 * This module is the heart of WACA and performs all the dificult "in
 * the server" type stuff, any serious problem here could crash the
 * whole server, so be careful!
 *
 * 05-04-99 Blake Mitchell
 * ---------
 *
 * Installation and Use
 * --------------------
 *
 * Make sure the mod_macs_auc.so is in your apache's modules/ dir,
 * that the following line is in httpd.conf:
 *   LoadModule macs_auc_module modules/mod_macs_auc.so
 *
 * There are two ways to configure this auc. The best is to use
 * the MacsConfigFile directive in your base and/or <virtualhost>
 * sections, and the MacsAuthenticate directive in your <dir>,
 * <location>, or <file> sections:
 *   MacsConfigFile /path/to/cf
 *   <location /secureapps>
 *   MacsAuthenticate on
 *   </location>
 *   <location /secureapps/unsecure>
 *   MacsAuthenticate off
 *   </location>
 * The format of the MACS configuration file is out of the scope
 * of these comments, see the docs for MACS configuration.
 *
 * The old WacaConfigFile directive still works, too. Support for
 * this might (will) be taken out in a future release, so don't
 * use it. The uri lookups are fairly fast but redundant. You just
 * tell the directive where it can find an old-style config file:
 *   WacaConfigFile /path/to/auc.conf
 * A good auc.conf looks something like this:
 *   # auc.conf
 *   saucsrv1:1234
 *   saucsrv2:4321
 *   u/home*private
 *   u/some/secure/app
 * The "auc.conf" file should contain no blank lines. If the first
 * character of a line is '#', the line is ignored; if it is 'u', the
 * line is interpreted as a uri pattern, and if it is 's', the line is
 * interpreted as an AUS server (host:port). When a request matches
 * any of the specified patterns, authentication takes place to one of
 * the specified servers. Note the lack of spaces...
 *
 *
 * TODO
 * ----
 *
 * thread-safe for when that day comes (plus there'll be thread-safe
 *   auc code to reuse)
 */

 
/******************
 ** begin header **
 ******************/
 
/* identify yourself */
const char macs_auc_rcs_id[] = "$Id: mod_macs_auc.c,v 1.5 2002/04/09 15:41:23 santana Exp $";

/* apxs includes */
#include "httpd.h"
#include "http_log.h"
#include "http_config.h"
#include "http_core.h"
#include "http_protocol.h"
#include "ap_config.h"

#define MACS_AUC_BUFSIZ 1024
#define MACS_AUC_DEFAULT_PORT 0
#define MACS_AUC_DEFAULT_TIMEOUT 3

/* sockets to auth servers */
struct macs_auc_sock_struct {
  char *host;
  int port;
  int isup;
  int sock;
  int in_use;
  struct macs_auc_sock_struct *next;
};
typedef struct macs_auc_sock_struct macs_auc_sock;

/* secured uri's */
struct macs_auc_securi_struct {
  char *uri;
  int isxp;
  struct macs_auc_securi_struct *next;
};
typedef struct macs_auc_securi_struct macs_auc_securi;

/* server configuration data */
struct macs_auc_cfg_struct {
  int port,          /* default port (not used)                 */
    timeout;         /* timeout for various ops                 */
  char *cookie_name, /* name of the skey cookie                 */
    *cookie_domain,  /* skey cookie's domain                    */
    *uidhdr,         /* emulate cookie&header from client w/uid
			also set in r->notes if MACSAuthenticate */
    *lia;            /* url to the LogIn Authority              */
  macs_auc_securi *securis;   /* linked list of secure uris              */
  macs_auc_sock *socks;       /* linked list of AUC sockets              */
};
typedef struct macs_auc_cfg_struct macs_auc_cfg;

/* directory configuration data */
struct macs_auc_dir_cfg_struct {
  int secure;        /* this directory is secured */
};
typedef struct macs_auc_dir_cfg_struct macs_auc_dir_cfg;

/* auc initializations (logs a line) */
void macs_auc_init(server_rec *,pool *);

/* Initialize macs_auc_cfg
 * There won't be any *server* merging routines because if we give
 *   different servers different config files, we're plugging that
 *   server into a different, isolated instance of macs; and if we
 *   mark a <dir> as (un)secure, it will override any parent
 *   setting.
 */
void *macs_auc_svrcfg_init(pool *,server_rec *);
void *macs_auc_dircfg_init(pool *,char *);

/* New-style configuration directive handler
 *   same as old-style (see below), plus
 *   set global variables (cookie name/domain, lia location)
 */
const char *macs_auc_cfdirective(cmd_parms *,void *,char *);

/* Parse a config line
 * errors into arg1 w/ rv != 0
 */
int macs_auc_parse_cfline(char *,cmd_parms *,macs_auc_cfg *);

/* Configuration directive handler for old-style auc config file
 *   populate securi and ausspec lists
 *   connect and disconnect at child startup/exit
 */
const char *macs_auc_parse_oldcf(cmd_parms *,void *,char *);

/* Configuration directive handler to make this directory secure */
const char *macs_auc_secure_dir(cmd_parms *,void *,int);

/* Parse a config line
 * errors into arg1 w/ rv != 0
 */
int macs_auc_parse_oldcfline(char *,cmd_parms *,macs_auc_cfg *);

/* add a secured uri or an AUS server
 * errors into arg1 w/ rv != 0
 */
int macs_auc_add_securi(char *,cmd_parms *,macs_auc_cfg *);
int macs_auc_add_aus(char *,cmd_parms *,macs_auc_cfg *);
 
/* returns a null-initted macs_auc_sock/macs_auc_securi */
macs_auc_sock *new_macs_auc_sock(pool *);
macs_auc_securi *new_macs_auc_securi(pool *);
 
/* connect to the AUS specs */
void macs_auc_child_init(server_rec *,pool *);

/* close all AUS connections so garbage collection goes smoothly */
void macs_auc_child_exit(server_rec *,pool *);

/* if url is not secure, return DECLINED
 * if no valid skey cookie was sent, redirect to LIA
 * if AUS validates skey(*), return OK
 * else redirect to LIA
 *
 * (*) If during this stage, no AUS's are up, we rerun
 *     macs_auc_child_init
 */
int macs_auc_authenticate(request_rec *);

/* is arg1 matched by anything in the securi list? */
int macs_auc_uri_issecure(char *,macs_auc_cfg *);

/* asks the AUS about the key */
int macs_auc_auth_skey(char *,request_rec *,macs_auc_cfg *);

/* returns the skey */
char *macs_auc_get_skey(request_rec *,macs_auc_cfg *);

/* redirect to the LIA */
int macs_auc_tolia(request_rec *,macs_auc_cfg *);
 
/****************
 ** end header **
 ****************/

/* Config directive */
static const command_rec macs_auc_commands[] = {
  { "MACSConfigFile", macs_auc_cfdirective, NULL, RSRC_CONF,  TAKE1,
    "MACS configuration file: MacsConfigFile /path/to/macs.cfg" },
  { "AUCConfigFile", macs_auc_cfdirective, NULL, RSRC_CONF,  TAKE1,
    "MACS AUC-specific configuration file: AUCConfigFile /path/to/auc.cfg" },
  { "WacaConfigFile", macs_auc_parse_oldcf, NULL, RSRC_CONF,  TAKE1,
    "WACA-style MACS config file: WacaConfigFile /path/to/configfile" },
  { "MACSAuthenticate", macs_auc_secure_dir, NULL, OR_AUTHCFG, FLAG,
    "Secure this directory with MACS: MacsAuthenticate {on|off}" },
  {NULL}
};

/* Dispatch list for API hooks */
module MODULE_VAR_EXPORT macs_auc_module = {
  STANDARD_MODULE_STUFF, 
  macs_auc_init,         /* module initializer                  */
  macs_auc_dircfg_init,  /* create per-dir    config structures */
  NULL,                  /* merge  per-dir    config structures */
  macs_auc_svrcfg_init,  /* create per-server config structures */
  NULL,                  /* merge  per-server config structures */
  macs_auc_commands,     /* table of config file commands       */
  NULL,                  /* [#8] MIME-typed-dispatched handlers */
  NULL,                  /* [#1] URI to filename translation    */
  NULL,                  /* [#4] validate user id from request  */
  NULL,                  /* [#5] check if the user is ok _here_ */
  macs_auc_authenticate, /* [#3] check access by host address   */
  NULL,                  /* [#6] determine MIME type            */
  NULL,                  /* [#7] pre-run fixups                 */
  NULL,                  /* [#9] log a transaction              */
  NULL,                  /* [#2] header parser                  */
  macs_auc_child_init,   /* child_init                          */
  macs_auc_child_exit,   /* child_exit                          */
  NULL                   /* [#0] post read-request              */
};

void macs_auc_init(server_rec *s,pool *p) {
  ap_add_version_component("macs/auc");
} /* macs_auc_init */


/*****************************************************************
                  config block initializers
 *****************************************************************/
void *macs_auc_svrcfg_init(pool *p, server_rec *s) {
  macs_auc_cfg *c = (macs_auc_cfg *)ap_palloc(p,sizeof(macs_auc_cfg));
  c->port = MACS_AUC_DEFAULT_PORT;
  c->timeout = MACS_AUC_DEFAULT_TIMEOUT;
  c->cookie_name = c->cookie_domain = c->lia = NULL;
  c->securis = NULL;
  c->socks = NULL;
  return (void *)c;
} /* macs_auc_svrcfg_init */

void *macs_auc_dircfg_init(pool *p, char *dir) {
  macs_auc_dir_cfg *dc = (macs_auc_dir_cfg *)ap_palloc(p,sizeof(macs_auc_dir_cfg));
  dc->secure = 0;
  return (void *)dc;
} /* macs_auc_dircfg_init */


/*****************************************************************
             new-style directive handlers & co
 *****************************************************************/

/* make this directory/location secure (or not) */
const char *macs_auc_secure_dir(cmd_parms *parms, void *mconfig, int s) {
  macs_auc_dir_cfg *dc = (macs_auc_dir_cfg *)mconfig;
  dc->secure = s;
  return NULL;
} /* macs_auc_secure_dir */

/* go through the macs-style config file */
const char *macs_auc_cfdirective(cmd_parms *parms, 
				 void *mconfig, 
				 char *cfname) {
  FILE *cf;
  macs_auc_cfg *c;
  char *line;

  c = (macs_auc_cfg *)
    ap_get_module_config(parms->server->module_config,&macs_auc_module);
  /* we should inherit a pre-configured macs_auc_cfg by now */
  if ( !c )
    return "Internal error initializing configs";

  cf = ap_pfopen(parms->temp_pool,cfname,"r");
  line = (char *)ap_palloc(parms->temp_pool,sizeof(char)*MACS_AUC_BUFSIZ);
  if ( !cf ) {
    ap_snprintf(line,MACS_AUC_BUFSIZ,"Error opening %s: %s",
		cfname,strerror(errno));
    return line;
  }
  while (fgets(line,MACS_AUC_BUFSIZ,cf)) {
    switch (macs_auc_parse_cfline(line,parms,c)) {
    case 0:
      /* good */
      break;
    default:
      /* bad, errs in line */
      return line;
      break;
    } /* switch parse line */
  } /* while fgets */
  if ( ferror(cf) && !feof(cf) ) {
    /* bad */
    ap_snprintf(line,MACS_AUC_BUFSIZ,"Error reading from %s: %s",
		cfname,strerror(errno));
    return line;
  }
  return (char *)NULL;
} /* macs_auc_parse_cf */

/* changes the config block, or feeds a line to add_securi or add_aus */
int macs_auc_parse_cfline(char *line, cmd_parms *parms, macs_auc_cfg *c) {
  char *cdir, *val;

  cdir = ap_getword_white_nc(parms->temp_pool,&line);
  if (cdir[0] == '#')
    return 0; /* skip comments */
  /* there *has* to be a better way to do this, but this isn't a
   *   speed-sensitive section...
   * for now we'll just check each directive against the ones we care
   *   about
   */
  if (!strcasecmp(cdir,"LIA")) {
    c->lia = ap_getword_conf_nc(parms->pool,&line);
    if (!c->lia || !ap_is_url(c->lia)) {
      ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, parms->server,
		   "Null or invalid url for LIA: %s",line);
      return 1;
    }
    ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, parms->server,
		 "The LIA's url is %s\n",c->lia);
    return 0;
  }
  /* AUS host:port  OR  AUS host port */
  if (!strcasecmp(cdir,"AUS")) {
    val = ap_getword_white_nc(parms->temp_pool,&line);
    if (!val) {
      ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, parms->server,
		   "Null AUS spec in config: %s",line);
      return 1;
    }
    if ( ( *line ) &&
	 ( cdir = ap_getword_white_nc(parms->temp_pool,&line) ) &&
	 ( cdir[0] != '#' ) )
      val = ap_pstrcat(parms->temp_pool,val,":",cdir,NULL);
    return macs_auc_add_aus(val,parms,c);
  }
  if (!strcasecmp(cdir,"COOKIE_NAME")) {
    c->cookie_name = ap_getword_conf_nc(parms->pool,&line);
    if (!c->cookie_name) {
      ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, parms->server,
		   "Null cookie name in config: %s",line);
      return 1;
    }
    ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, parms->server,
		 "The cookie's name is %s\n",c->cookie_name);
    return 0;
  }
  if (!strcasecmp(cdir,"COOKIE_DOMAIN")) {
    c->cookie_domain = ap_getword_conf_nc(parms->pool,&line);
    if (!c->cookie_domain) {
      ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, parms->server,
		   "Null cookie domain in config: %s",line);
      return 1;
    }
    ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, parms->server,
		 "The cookie's domain is %s\n",c->cookie_domain);
    return 0;
  }
  if (!strcasecmp(cdir,"HTTP_UID_HEADER")) {
    c->uidhdr = ap_getword_conf_nc(parms->pool,&line);
    if (!c->uidhdr) {
      ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, parms->server,
		   "Null uid header in config: %s",line);
      return 1;
    }
    ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, parms->server,
		 "The emulated cookie&header name is %s\n",c->uidhdr);
    return 0;
  }
  if (!strcasecmp(cdir,"TIMEOUT")) {
    val = ap_getword_conf_nc(parms->pool,&line);
    if (!val) {
      ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, parms->server,
		   "Null timeout spec in config: %s",line);
      return 1;
    }
    c->timeout = atoi(val);
    ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, parms->server,
		 "The general timeout is %d seconds",c->timeout);
    return 0;
  }
  ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, parms->server,
	       "Skipping irrelevant config directive: %s",cdir);
  return 0;
} /* macs_auc_parse_cfline */


/*****************************************************************
             waca-style directive handler & parser
 *****************************************************************/

/* feeds cf to macs_auc_parse_oldcfline one line at a time */
const char *macs_auc_parse_oldcf(cmd_parms *parms,void *mconfig,char *cfname) {
  FILE *cf;
  macs_auc_cfg *c = (macs_auc_cfg *)
    ap_get_module_config(parms->server->module_config,&macs_auc_module);
  char *line = (char *)
    ap_palloc(parms->temp_pool,sizeof(char)*MACS_AUC_BUFSIZ);
  /* we should inherit a pre-configured macs_auc_cfg by now */
  if (c == NULL)
    return "Internal error initializing configs";
  cf = ap_pfopen(parms->temp_pool,cfname,"r");
  if ( !cf ) {
    ap_snprintf(line,MACS_AUC_BUFSIZ,"Error opening %s: %s",
		cfname,strerror(errno));
    return line;
  }
  while ( fgets(line,MACS_AUC_BUFSIZ,cf) ) {
    switch ( macs_auc_parse_oldcfline(line,parms,c) ) {
    case 0:
      /* good */
      break;
    default:
      /* bad, errs in line */
      return line;
      break;
    } /* switch parse line */
  } /* while fgets */
  if ( ferror(cf) && !feof(cf) ) {
    /* bad */
    ap_snprintf(line,MACS_AUC_BUFSIZ,"Error reading from %s: %s",
		cfname,strerror(errno));
    return line;
  }
  return (char *)NULL;
} /* macs_auc_parse_oldcf */

/* feeds each line to add_securi or add_aus */
int macs_auc_parse_oldcfline(char *line, cmd_parms *parms, macs_auc_cfg *c) {
  switch ( *line ) {
  case 'u': case 'U':
    return macs_auc_add_securi(&line[1],parms,c);
    break;
  case 's': case 'S':
    return macs_auc_add_aus(&line[1],parms,c);
    break;
  case '#': /* comment */
    break;
  default:  /* ignore unknown directives */
    break;
  }
  return 0;
} /* macs_auc_parse_oldcfline */


/*****************************************************************
                     config-block modifiers
        (used for both old- and new-style config directives)
 *****************************************************************/

/* adds a secure uri to the config block */
int macs_auc_add_securi(char *uri, cmd_parms *parms, macs_auc_cfg *cfg) {
  macs_auc_securi *curr = cfg->securis;
  char *c;
  if ( !cfg->securis ) {
    /* first item in list */
    curr = cfg->securis = new_macs_auc_securi(parms->pool);
  } else {
    /* append an item to list */
    while (curr->next && curr->next->uri) curr = curr->next;
    if ( curr->next )
      curr = curr->next;
    else
      curr = curr->next = new_macs_auc_securi(parms->pool);
  }
  if ( !curr ) {
    /* malloc error */
    uri--;
    ap_snprintf(uri,MACS_AUC_BUFSIZ,"Memory allocation error: %s",
		strerror(errno));
    return -1;
  }
  /* strip newline from securi spec */
  for ( c = curr->uri = ap_pstrdup(parms->pool,uri); *c; c++ );
  c--; if ( *c == '\n' ) *c = (char)NULL;  
  /* is it a valid expression? */
  curr->isxp = ap_is_matchexp(uri); 
  return 0;
} /* macs_auc_add_securi */

macs_auc_securi *new_macs_auc_securi(pool *p) {
  macs_auc_securi *s = (macs_auc_securi *)ap_palloc(p,sizeof(macs_auc_securi));
  s->uri = NULL;
  s->isxp = -1;
  s->next = NULL;
  return s;
} /* new_macs_auc_securi */

/* adds an aus to the config block */
int macs_auc_add_aus(char *hostparm, cmd_parms *parms, macs_auc_cfg *c) {
  char *host, *charpt;
  long port;
  macs_auc_sock *curr;
  /* parse hostparm */
  charpt = host = ap_pstrdup(parms->temp_pool,hostparm);
  if ( !host ) {
    /* malloc error */
    hostparm--;
    ap_snprintf(hostparm,MACS_AUC_BUFSIZ,"Memory allocation error: %s",
		strerror(errno));
    return -1;
  }
  while ( charpt && *charpt != ':' )
    charpt++;
  if ( charpt ) {
    *charpt = (char)NULL;
    port = atol(++charpt);
  } else {
    port = 42020;
    ap_log_error(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, 
		 parms->server,"No port in config line: %s",hostparm);    
  }
  if ( !c->socks ) {
    /* first item in list */
    curr = c->socks = new_macs_auc_sock(parms->pool);
  } else {
    /* append an item to list */
    for ( curr = c->socks; curr->next; curr = curr->next );
    curr = curr->next = new_macs_auc_sock(parms->pool);
  }
  if ( !curr ) {
    /* malloc error */
    hostparm--;
    ap_snprintf(hostparm,MACS_AUC_BUFSIZ,"Memory allocation error: %s",
		strerror(errno));
    return -1;
  }
  /* init the bare necessities; the client init actually connects */
  curr->host = ap_pstrdup(parms->pool,host);
  if ( !curr->host ) {
    /* malloc error */
    hostparm--;
    ap_snprintf(hostparm,MACS_AUC_BUFSIZ,"Memory allocation error: %s",
		strerror(errno));
    return -1;
  }
  curr->port = port;
  return 0;
} /* macs_auc_add_aus */

macs_auc_sock *new_macs_auc_sock(pool *p) {
  macs_auc_sock *s = (macs_auc_sock *)ap_palloc(p,sizeof(macs_auc_sock));
  s->host = NULL;
  s->port = 0;
  s->isup = 0;
  s->sock = -1;
  s->in_use = 0;
  s->next = NULL;
  return s;
} /* new_macs_auc_sock */


/*****************************************************************
                       child startup/exit
 *****************************************************************/

/* connect to AUSs on child startup (and at need) */
void macs_auc_child_init(server_rec *r, pool *p) {
  macs_auc_cfg *c = (macs_auc_cfg *)
    ap_get_module_config(r->module_config, &macs_auc_module);    
  macs_auc_sock *s;
  struct hostent *h;
  struct sockaddr_in addr;
  unsigned int host_addr;
  /* we don't get any real tmp pools, so
     we'll just do this buffer by hand */
  char *buf = (char *)malloc(sizeof(char)*MACS_AUC_BUFSIZ);
  /* I hate inheriting this... */
  errno = 0;
  /* multi-process rather than multi-threaded...  this might change
   * someday, but til then we don't want each child to connect to
   * every AUS. One is probably enough.
   * 
   * FIXME: there should be a way to evenly spread which server entry
   * the children connect to. Right now, the first entry gets hit til
   * it won't accept any more, then the second, etc.
   */  
  for (s = c->socks; s; s = s->next) {
    if ( s->in_use ) {
      /* in use means ok */
      ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		   "AUS %s:%i in use",s->host,s->port);
      continue;
    }
    s->in_use = 1;
    errno = 0;
    /* let's just ping and see if we get a response */
    if ( s->isup && ! ( write(s->sock,"PING\n",5) == 5   &&
			read(s->sock,buf,MACS_AUC_BUFSIZ)     &&
			strncasecmp(buf,"PONG\n",5) == 0 ) ) {
      ap_log_error(APLOG_MARK, APLOG_DEBUG, r,
		   "AUS %s:%i down", s->host, s->port);
      s->isup = 0;
    } else if ( s->isup ) {
      ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		   "Already connected to AUS %s:%i",s->host,s->port);
      s->in_use = 0;
      /* this shouldn't happen in multi-process mode */
      break;
    }
    /* if down, bring back up */
    close(s->sock);
    if ( !(h = gethostbyname(s->host)) ) {
      ap_log_error(APLOG_MARK, APLOG_WARNING, r,
		   "Couldn't get hostname for %s", s->host);
      s->in_use = 0;
      continue;
    }
    host_addr = *(unsigned int *)h->h_addr;
    addr.sin_family = AF_INET;
    addr.sin_port = htons(s->port);
    addr.sin_addr.s_addr = host_addr;
    s->sock = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);
    if ( s->sock < 0 ) {
      ap_log_error(APLOG_MARK, APLOG_WARNING, r,
		   "Couldn't open socket to %s:%i", s->host,s->port);
      s->in_use = 0;
      continue;
    }
    if ( connect(s->sock,(struct sockaddr *)&addr,sizeof(addr)) < 0 ) {
      ap_log_error(APLOG_MARK, APLOG_DEBUG, r,
		   "Failed to connect to AUS %s:%i", s->host, s->port);
      close(s->sock);
      s->in_use = 0;
      continue;
    }
    /* let's just ping and see if we get a response */
    if ( s->isup && ! ( write(s->sock,"PING\n",5) == 5   &&
			read(s->sock,buf,MACS_AUC_BUFSIZ)     &&
			strncasecmp(buf,"PONG\n",5) == 0 ) ) {
      ap_log_error(APLOG_MARK, APLOG_WARNING, r,
		   "Communication with AUS %s:%i failed", s->host, s->port);
      s->in_use = 0;
      continue;
    } else {
      ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		   "Connected to AUS %s:%i", s->host, s->port);
      s->isup = 1;
      s->in_use = 0;
      /* we got one, that's enough */
      break;
    }
  } /* foreach socket */
  free(buf);
  return;
} /* macs_auc_child_init */

void macs_auc_child_exit(server_rec *r, pool *p) {
  macs_auc_cfg *c = (macs_auc_cfg *)
    ap_get_module_config(r->module_config, &macs_auc_module);    
  macs_auc_sock *s;
  for (s = c->socks; s; s = s->next) {
    errno = 0;
    if (s->sock)
      close(s->sock);
    ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		 "Disconnecting from %s:%i", s->host, s->port);
    s->in_use = 1;
    s->isup = 0;
  }
  return;
} /* macs_auc_child_exit */


/*****************************************************************
          the actual authentication routines (finally)
 *****************************************************************/
int macs_auc_authenticate(request_rec *r) {
  char *skey;
  macs_auc_cfg *c;
  macs_auc_dir_cfg *dc;
  if ( r->main && r->main != r ) {
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		  "ignoring subrequest %s (main request %s)",
		  r->uri, r->main->uri);
    return DECLINED;
  }
  dc = (macs_auc_dir_cfg *)
    ap_get_module_config(r->per_dir_config, &macs_auc_module);
  c = (macs_auc_cfg *)
    ap_get_module_config(r->server->module_config, &macs_auc_module);
  skey = macs_auc_get_skey(r,c);
  /* check both the dir_cfg and waca-style places to see if this uri
   * is secured -- the waca-style check is much slower (prolly 'cause
   * it's redundant) and it will prolly be cut out soon.
   */  
  if ( !dc->secure && !macs_auc_uri_issecure(r->uri,c) ) {
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		  "uri %s NOT SECURE", r->uri);
    /* not secure, but we want to expose uid if we have it */
    if (skey) macs_auc_auth_skey(skey,r,c);
    return DECLINED;
  } else {
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		  "authenticating securi %s", r->uri);
    /* mark that we're authenticating */
    ap_table_set(r->notes,c->uidhdr,"y");
    if (!skey) {
      ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		    "No skey cookie");
      return macs_auc_tolia(r,c);
    }
    /* prelims done, check against AUS */
    if (macs_auc_auth_skey(skey,r,c) == OK) return OK;
    return macs_auc_tolia(r,c);
  }
} /* macs_auc_authenticate */

int macs_auc_uri_issecure(char *uri, macs_auc_cfg *c) {
  macs_auc_securi *s;
  for (s = c->securis; s; s = s->next) {
    if ( s->isxp ) {
      if ( !ap_strcmp_match(uri,s->uri) )
	return 1; /* shexp matched */
    } else {
      if ( !strcmp(uri,s->uri) )
	return 1; /* literal matched */
    }
  } /* foreach securi */
  return 0;  /* no match */
} /* macs_auc_uri_issecure */

int macs_auc_auth_skey(char *skey, request_rec *r, macs_auc_cfg *c) {
  macs_auc_sock *s = c->socks;
  int upsvrs, rv;
  char *query = (char *)ap_palloc(r->pool,sizeof(char)*MACS_AUC_BUFSIZ),
    *reply;
 
  /* create the query */
  rv = ap_snprintf(query,MACS_AUC_BUFSIZ,"ATH %s %s\n",
		   skey, r->connection->remote_ip);
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		"query will be %s",query);

  /* if there are servers up, keep trying them until one shows up
   * unused, then grab it
   */
  for (upsvrs=0; s||upsvrs; s=s->next) {
    if ( !s ) {
      /* start over */
      s = c->socks;
      upsvrs = 0;
    }
    if ( s->isup )
      upsvrs++;
    else
      continue;
    if ( !s->in_use ) {
      s->in_use = 1;
      break;
    }
  }
  if ( !s ) {
    /* no AUS's up! */
    if ( ap_table_get(r->notes,"macs_auc_childreinit") ) {
      ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, r,
		    "No AUS's up after reinit!");
      return macs_auc_tolia(r,c);
    }
    ap_log_rerror(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, r,
		  "No AUS's up! Reinitting");
    ap_table_setn(r->notes,"macs_auc_childreinit","1"); /* only reinit once */
    macs_auc_child_init(r->server,(pool *)NULL);
    return macs_auc_auth_skey(skey,r,c);
  }
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		"writing query");

  /* send the query to the AUS */
  if ( (upsvrs=write(s->sock,query,rv)) != rv ) {
    ap_log_rerror(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, r,
		  "Retry on error writing to AUS %s:%i: %s",
		  s->host, s->port, strerror(errno));
    s->isup = 0;
    s->in_use = 0;
    return macs_auc_auth_skey(skey,r,c);
  }
 
  /* get the reply */
  rv = upsvrs = errno = 0;
  reply = (char *)ap_palloc(r->pool,sizeof(char)*MACS_AUC_BUFSIZ);
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		"reading reply");
  while (1) {
    /* upsvrs := bytes read */
    upsvrs = read(s->sock,&reply[rv],MACS_AUC_BUFSIZ-rv);
    if ( upsvrs > 0 )
      rv += upsvrs;
    if ( reply[rv-1] == '\n' ) {
      break;
    }
    switch (errno) {
    case 0: /* fine, read more */
      break;
    case EAGAIN: /* temporary error condition - clear & read more */
      errno = 0;
      break;
    case ETIMEDOUT: /* timed out waiting for response */
      ap_log_rerror(APLOG_MARK, APLOG_WARNING, r,
		    "Timeout waiting for AUS %s:%i", s->host, s->port);
    default: /* read error */
      ap_log_rerror(APLOG_MARK, APLOG_WARNING, r,
		    "Retry on read error from AUS %s:%i", s->host, s->port);
      s->isup = 0;
      s->in_use = 0;
      return macs_auc_auth_skey(skey,r,c);
    }
  } /* read reply loop */
  s->in_use = 0;

  ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		"reply is %s",reply);

  if ( !strncasecmp("YES ",reply,4) ) {
    /* strip that pesky newline */
    for ( query=reply; *query; query++ )
      if ( *query=='\n' )
	*query=(char)NULL;
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		  "Successful authen for uid %s",&reply[4]);
    /* woo-hoo  -- add the header */
    ap_table_set(r->headers_in,c->uidhdr,&reply[4]);
    ap_table_set(r->subprocess_env,c->uidhdr,&reply[4]);
    /* and add what looks like a cookie from the client */
    query = (char *)ap_table_get(r->headers_in,"Cookie");
    ap_table_unset(r->headers_in,"Cookie");
    ap_table_set(r->headers_in,"Cookie",
		 ap_pstrcat(r->pool, query, "; ",
			    c->uidhdr, "=", &reply[4],
			    "\0", NULL));
    return OK;
  }
  if ( !strncasecmp("NO\n",reply,3) ) {
    /* no good */
    ap_log_rerror(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, r,
		  "Unsuccessful authen for skey %s", skey);
    return AUTH_REQUIRED;
  }
  /* error string from server */
  ap_log_rerror(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, r,
		"Error string from AUS: %s", reply);
  return AUTH_REQUIRED;
} /* macs_auc_auth_skey */

char *macs_auc_get_skey(request_rec *r, macs_auc_cfg *c) {
  const char *cookie, *name, *skey;

  /* careful not to modify the strings in the tables */
  cookie = ap_table_get(r->headers_in,"Cookie");
  if (!cookie) return NULL; /* no cookies */
  while ( *cookie && (skey = ap_getword(r->pool,&cookie,';')) ) {
    name = ap_getword(r->pool,&skey,'=');
    if ( strcmp(name,c->cookie_name) ) {
      /* this ain't it, line up the next cookie and redo */
      while (*cookie == ' ' || *cookie == '\t')
	cookie++;
      continue;
    }
    return (char *)skey; /* found it */
  }
  return NULL; /* no skey cookie */
} /* macs_auc_get_skey */

int macs_auc_tolia(request_rec *r, macs_auc_cfg *c) {
  char *ptr, *unescaped_uri, *scheme, /* *user, *password, */
    *loc, *escaped_uri;
  unsigned int port;

  loc = (char *)ap_palloc(r->pool,sizeof(char)*MACS_AUC_BUFSIZ);
  escaped_uri = (char *)ap_palloc(r->pool,sizeof(char)*MACS_AUC_BUFSIZ);

  /* we'll embed where we are now in the url we're redirecting to, but
   *   that means any '&' or '=' chars might be processed as part of
   *   the request over there. We'll escape '&', '?' and '=' so it'll
   *   keep until the far side unescapes it to send our user back
   *   here...
   * note that there's no predone way to figure all this stuff out,
   *   unless we're serving a proxying request... i wish it weren't so
   *   ugly
   */
  ptr = escaped_uri;
  for (unescaped_uri = r->unparsed_uri; *unescaped_uri; unescaped_uri++ ) {
    switch (*unescaped_uri) {
    case '&':
      *ptr = '%'; ptr++;
      *ptr = '2'; ptr++;
      *ptr = '6'; ptr++;
      break;
    case '=':
      *ptr = '%'; ptr++;
      *ptr = '3'; ptr++;
      *ptr = 'D'; ptr++;
      break;
    case '?':
      *ptr = '%'; ptr++;
      *ptr = '3'; ptr++;
      *ptr = 'F'; ptr++;
      break;
    default:
      *ptr = *unescaped_uri; ptr++;
      break;
    }
  }
  /* dont't forget to terminate the string (BAM) */
  *ptr = (char)NULL;
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		"our url is %s",escaped_uri);
 
  /* FIXME: we need some better way of figuring out the scheme */
  switch ( port = ap_get_server_port(r) ) {
  case 443:
    scheme = ap_pstrdup(r->pool,"https");
    port = 0;
    break;
  case 80:
    scheme = ap_pstrdup(r->pool,"http");
    port = 0;
    break;
  default:
    scheme = ap_pstrdup(r->pool,"http");
    break;    
  }
  /* FIXME: we should take care of username/passwords here */
  if ( port ) {
    ap_snprintf(loc, MACS_AUC_BUFSIZ, "%s?url=%s://%s:%i%s", c->lia,
		scheme, /*ap_get_server_name(r)*/r->hostname, 
		port, escaped_uri);
  } else {
    ap_snprintf(loc, MACS_AUC_BUFSIZ, "%s?url=%s://%s%s", c->lia, scheme,
		/*ap_get_server_name(r)*/r->hostname, escaped_uri);
  }    
  ap_table_set(r->headers_out,"Location",loc);
  r->status_line = ap_pstrdup(r->pool,"302 Login");
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		"Redirecting to LIA: %s", loc);
  return REDIRECT;
} /* macs_auc_tolia */
