/* $Id: auc.c,v 1.1.1.1 2000/09/03 03:39:58 santana Exp $
 *
 * An implementation of the auc piece of ACS. Starting off from a perl
 * module with the following comments
 *
 * ---------
 * WACA.pm
 *
 * This module is the heart of WACA and performs all the dificult
 * "in the server" type stuff, any serious problem here could crash
 * the whole server, so be careful!
 *
 * 05-04-99 Blake Mitchell
 * ---------
 *
 * Place the following lines in obj.conf
 *
 * Init fn="load-modules" shlib="auc.so" funcs="auc-init,auc-auth"
 * Init fn="auc-init" cf="auc.conf"
 * PathCheck fn="auc-auth"
 *
 * The "auc.conf" file might look like this:
 *
 * # auc.conf
 * s aucsrv1:1234
 * s aucsrv2:4321
 * u /home*private
 * u /some/secure/app
 *
 * The "auc.conf" file should contain no blank lines. If the first
 * character of a line is '#', the line is ignored; if it is 'u', the
 * line is interpreted as a uri pattern, and if it is 's', the line is
 * interpreted as an AUS server (host:port). When a request matches
 * any of the specified patterns, authentication takes place to one of 
 * the specified servers.
 */


/******************
 ** begin header **
 ******************/

/* identify yourself */
const char auc_rcs_id[] = "$Id: auc.c,v 1.1.1.1 2000/09/03 03:39:58 santana Exp $";

/* all the includes for nsapi are here */
#include <nsapi.h>

/* FIXME: we should try to use nsapi calls only */
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>

#define AUC_DEF_PORT 4322
#define AUC_BUFSIZ 512
#define AUC_AUS_TO 3
#define AUC_SW_SLEEP_MIN .5
#define AUC_COOKIE_NAME "test-macs-skey"
#define AUC_LIA_URL "This auc needs to be recoded to use the new config scheme"
#define AUC_ACS_UID_HDR "test-macs-uid"

/* sockets to auth servers */
struct auc_sock_struct {
  char *host;
  int port;
  int isup;
  /* FIXME: we should try to use NSAPI networking */
  int sock;
  int in_use;
  struct auc_sock_struct *next;
};
typedef struct auc_sock_struct auc_sock;
static auc_sock *auc_socks = NULL;

/* secured uri's */
struct auc_securi_struct {
  char *uri;
  int isxp;
  struct auc_securi_struct *next;
};
typedef struct auc_securi_struct auc_securi;
static auc_securi *auc_securis = NULL;

/* we can only pass one arg to the socket-watcher */
struct snrq_st {
  Session *sn;
  Request *rq;
};

/* the socket-watcher thread */
SYS_THREAD auc_swthr;

/* for getting an unused socket */
static CRITICAL auc_sox_lock = NULL;
/* one init/cleanup at a time */
static CRITICAL auc_init_lock = NULL;

/* initializations: secure uri list, AUS socket pool,
 * socket-watcher thread
 * (Init SAF)
 */ 
int auc_init (pblock *, Session *, Request *);

/* if url is not secure, return REQ_NOACTION
 * if no valid skey cookie was sent, redirect to LIA
 * if AUS validates skey, return REQ_PROCEED
 * else redirect to LIA
 * (PathCheck SAF)
 */
int auc_auth (pblock *, Session *, Request *);

/* close all open sockets gracefully
 * free all structs
 * terminate socket-watcher thread
 */
int auc_cleanup(void);

/* returns a null-initted auc_sock/auc_securi */
auc_sock *new_auc_sock(void);
auc_securi *new_auc_securi(void);

/* recursives for releasing the linked lists */
void auc_empty_securis(auc_securi *);
void auc_empty_socks(auc_sock *);

/* put cf info in the init pblock */
int auc_parse_cf(filebuffer *, Session *, Request *);

/* add a secured uri or an AUS server */
int auc_add_securi(char *,Session *,Request *);
int auc_add_aus(char *,Session *,Request *);

/* redirect to the LIA */
int auc_tolia(Session *,Request *);

/* authenticate an skey with the AUS */
int auc_authenticate_cookie(char *,Session *,Request *);

/* socket-watcher code */
void *auc_socket_watcher(void *);

/****************
 ** end header **
 ****************/

int auc_init (pblock *pb, Session *sn, Request *rq) {
  char *cfname, *buf = (char *)MALLOC( sizeof(char) * AUC_BUFSIZ );
  struct stat cfstat;
  SYS_FILE scf;
  filebuffer *cf;
  auc_sock *current_sock;
  struct snrq_st snrq;

  /*log_error(LOG_INFORM, "auc-init", sn, rq,
    "Entering auc-init");*/
  /* init locks */
  if ( !auc_init_lock )
    auc_init_lock= crit_init();
  crit_enter(auc_init_lock);
  if ( !auc_sox_lock )
    auc_sox_lock = crit_init();
  crit_enter(auc_sox_lock);

  /* read config file, init secured uri & AUS host lists */
  cfname = pblock_findval("cf", pb);
  
  if ( !cfname ) {
    log_error(LOG_MISCONFIG, "auc-init", sn, rq,
	      "No cf parameter in obj.conf!");
    return REQ_ABORTED;
  }

  /*log_error(LOG_INFORM, "auc-init", sn, rq,
    "Statting cf: %s",cfname);*/
  if ( stat(cfname,&cfstat) ) {
    log_error(LOG_FAILURE, "auc-init", sn, rq,
	      "Trouble with %s: %s",cfname,rq->staterr);
    return REQ_ABORTED;
  } 

  /*log_error(LOG_MISCONFIG, "auc-init", sn, rq,
    "Got stats");*/
  if ( !(((cfstat.st_mode ^ S_IRUSR) ^ S_IRGRP) ^ S_IROTH) ) {
    log_error(LOG_FAILURE, "auc-init", sn, rq,
	      "Not enough perms opening cf: %s", cfname);
    return REQ_ABORTED;
  } 

  /*log_error(LOG_MISCONFIG, "auc-init", sn, rq,
    "Opening");*/
  if ( !(scf=system_fopenRO(cfname)) ) {
    log_error(LOG_FAILURE, "auc-init", sn, rq,
	      "Could not open cf: %s",cfname);
    return REQ_ABORTED;
  } 

  /*log_error(LOG_MISCONFIG, "auc-init", sn, rq,
    "buffering");*/
  if ( !(cf=filebuf_open_nostat(scf,AUC_BUFSIZ,&cfstat)) ) {
    log_error(LOG_FAILURE, "auc-init", sn, rq,
	      "filebuf_open of cf failed!");
    return REQ_ABORTED;
  } 

  /*log_error(LOG_INFORM, "auc-init", sn, rq,
    "Parsing cf");*/
  if ( auc_parse_cf(cf,sn,rq) == -1 ) {
    filebuf_close(cf);
    return REQ_ABORTED;
  }

  /* start the socket-watcher thread */
  snrq.sn = sn; snrq.rq = rq;
  /*log_error(LOG_INFORM, "auc-init", sn, rq,
    "Starting socket_watcher");*/
  auc_swthr = systhread_start(1,0,(void *)auc_socket_watcher,(void *)&snrq);
  if ( auc_swthr == SYS_THREAD_ERROR ) {
    log_error(LOG_FAILURE, "auc-init", sn, rq,
	      "Error starting sw thread!");
    return REQ_ABORTED;
  }

  /* attach the cleanup code */
  /*log_error(LOG_INFORM, "auc-init", sn, rq,
    "Attaching cleanup code");*/
  daemon_atrestart((void *)auc_cleanup,NULL);

  /* we out */
  crit_exit(auc_init_lock);
  crit_exit(auc_sox_lock);
  return REQ_PROCEED;
} /* auc_init */

int auc_auth (pblock *pb, Session *sn, Request *rq) {
  char *uri, *cookie;
  auc_securi *curr;
  pblock *cookies_pb = pblock_create(10);
  int val;

  /* is this a secured uri? */
  uri = pblock_findval("uri",rq->reqpb);
  if ( !uri ) {
    log_error(LOG_FAILURE, "auc-auth", sn, rq,
	      "Error determining uri");
    return REQ_ABORTED;
  }
  /*log_error(LOG_INFORM, "auc-auth", sn, rq,
    "Entering auc-auth w/ %s",uri);*/
  for (curr=auc_securis; curr; curr=curr->next ) {
    /*log_error(LOG_INFORM, "auc-auth", sn, rq,
      "checking securi %s isxp %d",curr->uri,curr->isxp);*/
    /* FIXME: not a NSAPI call */    
    if ( !(curr->isxp) && !(strcmp(uri,curr->uri)) )
      break;
    /*log_error(LOG_INFORM, "auc-auth", sn, rq,
      "checking securi %s",curr->uri);*/
    if ( curr->isxp && !(shexp_cmp(uri,curr->uri)) )
      break;
    /*log_error(LOG_INFORM, "auc-auth", sn, rq,
      "checking securi %s",curr->uri);*/
  }
  if ( !curr ) {
    /* not a secured uri */
    /*log_error(LOG_INFORM, "auc-auth", sn, rq,
      "Letting %s pass",uri);*/
    return REQ_NOACTION;
  }

  /*log_error(LOG_INFORM, "auc-auth", sn, rq,
    "about to get cookies");*/
  /* Do they have a valid skey cookie? */
  val = request_header("cookie",&cookie,sn,rq);
  /*log_error(LOG_INFORM, "auc-auth", sn, rq,
    "just got cookies");*/
  switch ( val ) {
  case REQ_EXIT:
    /* problem with client */
    log_error(LOG_INFORM, "auc-auth", sn, rq,
	      "problem with client");
    return REQ_EXIT;
  case REQ_ABORTED:
    /* no cookies */
    /*log_error(LOG_INFORM, "auc-auth", sn, rq,
      "no cookies for %s",uri);*/
    return auc_tolia(sn,rq);
  case REQ_PROCEED:
    if (!cookie) { /* no cookies */
      /*log_error(LOG_INFORM, "auc-auth", sn, rq,
	"no cookies for %s",uri);*/
      return auc_tolia(sn,rq);
    }
    /*log_error(LOG_INFORM, "auc-auth", sn, rq,
      "About to parse: %s",cookie);*/
  } /* grab cookies */
  if ( pblock_str2pblock(cookie,cookies_pb) == -1 ) {
    log_error(LOG_FAILURE, "auc-auth", sn, rq,
	      "Error parsing cookies: %s",cookie);
    return auc_tolia(sn,rq);
  } /* parse cookies */
  /*log_error(LOG_INFORM, "auc-auth", sn, rq,
    "just parsed");*/
  cookie = pblock_findval(AUC_COOKIE_NAME,cookies_pb);
  log_error(LOG_INFORM, "auc-auth", sn, rq,
	    "just found: %s",cookie);
  if ( !cookie ) {/* not logged in */
    /*log_error(LOG_INFORM, "auc-auth", sn, rq,
      "not logged in %s",uri);*/
    return auc_tolia(sn,rq);
  }
  /* does the cookie work? */
  switch ( auc_authenticate_cookie(cookie,sn,rq) ) {
  case 0:
    /* woo-hoo */
    log_error(LOG_INFORM, "auc-auth", sn, rq,
	      "woo-hoo %s",uri);
    return REQ_PROCEED;
  case 1:
    /* no good */
    log_error(LOG_INFORM, "auc-auth", sn, rq,
	      "no good %s",uri);
    return auc_tolia(sn,rq);
  case 2:
    /* no cookies?! */
    return REQ_ABORTED;
  case -1:
    /* FIXME: what to do if there's no AUS */
    log_error(LOG_FAILURE, "auc-auth", sn, rq,
	      "no AUS! %s",uri);
    return auc_tolia(sn,rq);
  } /* aus validates skey? */

  log_error(LOG_FAILURE, "auc-auth", sn, rq,
	    "huh? %s",uri);
  return REQ_NOACTION;
} /* auc_auth */

int auc_cleanup(void) {
  crit_enter(auc_init_lock);
  systhread_terminate(auc_swthr);
  auc_empty_securis(auc_securis);
  auc_empty_socks(auc_socks);
} /* auc_cleanup */

void auc_empty_securis(auc_securi *u) {
  if ( u->next )
    auc_empty_securis(u->next);
  FREE(u->uri);
  FREE(u);
  u = NULL;
} /* auc_empty_securis */

void auc_empty_socks(auc_sock *u) {
  if ( u->next )
    auc_empty_socks(u->next);
  close(u->sock);
  FREE(u->host);
  FREE(u);
  u = NULL;
} /* auc_empty_socks */

int auc_parse_cf(filebuffer *cf, Session *sn, Request *rq) {
  int lineno=1,action;
  char buf[AUC_BUFSIZ];
  /*log_error(LOG_INFORM,"auc-parse-cf", sn, rq,
    "Entering parse_cf");*/
  while (lineno) {
    action = util_getline(cf,lineno,AUC_BUFSIZ,buf);
    /*log_error(LOG_INFORM,"auc-parse-cf", sn, rq,
      "Line %i, Adding %c",lineno,*buf);*/
    switch ( action ) {
    case -1:
      /* error reading */
      log_error(LOG_FAILURE,"auc-init", sn, rq,
		"Error reading config file: %s",buf);
      return -1;
    case 1:
      /* eof, last line */
      lineno = -1;
    case 0:
      /* parse this line */
      switch ( buf[0] ) {
      case '#':
	/* comment */
	break;
      case 'u':
      case 'U':
	if ( auc_add_securi(&buf[1],sn,rq) == -1 )
	  return -1;
	break;
      case 's':
      case 'S':
	if ( auc_add_aus(&buf[1],sn,rq) == -1 )
	  return -1;
	break;
      } /* parse a line */
      break;
    } /* getline(cf)... */
    lineno++;
  } /* ...accompanying while */
} /* auc_parse_cf */

int auc_tolia (Session *sn, Request *rq) {
  char *rurl = (char *)MALLOC(sizeof(char)*AUC_BUFSIZ),
    *url,
    *host = NULL,
    *port = NULL;
  int count;
  
  url = http_uri2url_dynamic("","",sn,rq);
  for (count=0; url[count]; count++) {
    if ( !host ) {
      if ( url[count] == '/' && url[count+1] == '/' ) {
	count++; count++;
	host = &url[count];
      }
    } else if ( url[count] == ':' ) {
      url[count++] = NULL;
      port = &url[count];
      break;
    }
  }
  if ( port ) {
    log_error(LOG_INFORM, "auc_tolia", sn, rq,
	      "%s?server=%s&port=%s&uri=%s",
	      AUC_LIA_URL,host,port,pblock_findval("uri",rq->reqpb));
    util_snprintf(rurl,AUC_BUFSIZ,"%s?server=%s&port=%s&uri=%s",
		  AUC_LIA_URL,host,port,pblock_findval("uri",rq->reqpb));
  } else {
    log_error(LOG_INFORM, "auc_tolia", sn, rq,
	      "Redirecting %s?server=%s&uri=%s",
	      AUC_LIA_URL,host,pblock_findval("uri",rq->reqpb));
    util_snprintf(rurl,AUC_BUFSIZ,"%s?server=%s&uri=%s",
		  AUC_LIA_URL,host,pblock_findval("uri",rq->reqpb));
  }
  pblock_nvinsert("url",rurl,rq->vars);
  protocol_status(sn,rq,PROTOCOL_REDIRECT,"Please log in.");
  FREE(url);
  FREE(rurl);
  return REQ_ABORTED;
} /* auc_tolia */

int auc_authenticate_cookie (char *c, Session *sn, Request *rq) {
  auc_sock *s = auc_socks;
  int upsvrs, rv;
  char *query = (char *)MALLOC(sizeof(char)*AUC_BUFSIZ),
    *reply = (char *)MALLOC(sizeof(char)*AUC_BUFSIZ),
    *cookie;
  pb_param *cookie_pp;

  /*log_error(LOG_INFORM, "authenticate", sn, rq,
    "Entering auc_authenticate_cookie(%s)",c);*/
  /* create the query */
  rv = util_snprintf(query,AUC_BUFSIZ,"ATH %s %s\n",
		     c, pblock_findval("ip",sn->client));
  /*log_error(LOG_INFORM, "authenticate", sn, rq,
    "created query: %s",query);*/

  /* if there are servers up, keep trying them until one shows up
   * unused, then grab it
   */
  crit_enter(auc_sox_lock);
  for (upsvrs=0; s||upsvrs; s=s->next) {
    if ( !s ) {
      /* start over */
      s = auc_socks;
      upsvrs = 0;
    }
    if ( s->isup )
      upsvrs++;
    else
      continue;
    if ( !s->in_use ) {
      s->in_use = 1;
      break;
    }
  }
  /*log_error(LOG_INFORM, "authenticate", sn, rq,
    "found server %s:%s",s->host,s->port);*/
  crit_exit(auc_sox_lock);
  if ( !s ) {
    /* no AUS's up! */
    log_error(LOG_FAILURE, "auc-auth", sn, rq,
	      "Could not find an AUS up");
    FREE(query);
    FREE(reply);
    return -1;
  }

  /* send the query to the AUS */
  if ( (upsvrs=write(s->sock,query,rv)) != rv ) {
    log_error(LOG_FAILURE, "auc-auth", sn, rq,
	      "Error while talking to AUS %s:%i - %s",
	      s->host, s->port, strerror(errno));
    s->isup = 0;
    s->in_use = 0;
    FREE(query);
    FREE(reply);
    return auc_authenticate_cookie(c,sn,rq);
  }
  /*log_error(LOG_INFORM, "authenticate", sn, rq,
    "wrote");*/

  /* get the reply */
  rv = upsvrs = errno = 0;
  while (1) {
    upsvrs += read(s->sock,&reply[rv],AUC_BUFSIZ-rv/*,AUC_AUS_TO*/);
    if ( upsvrs > 0 )
      rv += upsvrs;    
    if ( reply[rv-1] == '\n' ) {
      /*log_error(LOG_INFORM, "authenticate", sn, rq,
	"found it!");*/
      reply[rv-1] = NULL;
      break;
    }
    /*log_error(LOG_INFORM, "authenticate", sn, rq,
      "Read so far: %s",reply);*/
    
    switch (errno) {
    case 0:
      break;
    case EAGAIN:
      /*log_error(LOG_FAILURE, "auc-auth", sn, rq,
	"EAGAIN Error while reading %s(%d) from AUS %s:%i - %s",
	reply,rv,s->host,s->port,strerror(errno));*/
      errno = 0;
      break;
    case ETIMEDOUT:
      log_error(LOG_FAILURE, "auc-auth", sn, rq,
		"Timed out reading %s(%i) from AUS %s:%i",
		reply,rv,s->host,s->port);
    default:
      /* FIXME: strerror is not NSAPI */
      log_error(LOG_FAILURE, "auc-auth", sn, rq,
		"Error while reading %s(%i) from AUS %s:%i - %s",
		reply,rv,s->host,s->port,strerror(errno));
      s->isup = 0;
      s->in_use = 0;
      FREE(query);
      FREE(reply);
      return auc_authenticate_cookie(c,sn,rq);
    } /* switch errno */
  } /* while read reply */
  s->in_use = 0;
  /*log_error(LOG_INFORM, "authenticate", sn, rq,
    "read %s", reply);*/
  
  /* FIXME: not a NSAPI call */
  if ( !strncasecmp("YES ",reply,4) ) {
    /* woo-hoo  -- add the header */
    pblock_nvinsert(AUC_ACS_UID_HDR,&reply[4],rq->headers);
    /* and add what looks like a cookie from the client */
    if ( !(cookie_pp = pblock_remove("cookie",rq->headers)) ) {
      /* no cookies */
      log_error(LOG_FAILURE, "auc-auth", sn, rq,
		"no cookies after we already found them");
      return 2;
    }
    cookie = (char *)MALLOC(sizeof(char)*AUC_BUFSIZ);
    util_snprintf(cookie,AUC_BUFSIZ,"%s; %s=%s",
		  cookie_pp->value, AUC_ACS_UID_HDR, &reply[4]);
    pblock_nvinsert("cookie",cookie,rq->headers);
    param_free(cookie_pp);
    FREE(cookie);
    FREE(query);
    FREE(reply);
    return 0;
  }
  /* FIXME: not a NSAPI call */
  if ( !strncasecmp("NO\n",reply,3) ) {
    /* no good */
    FREE(query);
    FREE(reply);
    return 1;
  }
  /* error string from server */
  log_error(LOG_FAILURE, "auc-auth", sn, rq,
	    "AUS error: %s",reply);
  FREE(query);
  FREE(reply);
  return -1;
} /* auc_authenticate_cookie */

int auc_add_securi(char *uri,Session *sn, Request *rq) {
  int rv;
  auc_securi *curr = auc_securis;

  /*log_error(LOG_INFORM, "auc-add", sn, rq,
    "Entering auc_add_securi w/ %s",uri);*/
  if ( !auc_securis ) {
    /* first item in list */
    curr = auc_securis = new_auc_securi();
  } else {
    /* append an item to list */
    while ( curr->next && curr->next->uri )
      curr = curr->next;
    if ( curr->next )
      curr = curr->next;
    else
      curr = curr->next = new_auc_securi();
  }
  
  /* is it a valid expression? */
  switch ( shexp_valid(uri) ) {
  case NON_SXP:
    log_error(LOG_INFORM, "auc-add", sn, rq,
	      "Adding secure uri literal: %s",uri);
    curr->uri = STRDUP(uri);
    curr->isxp = 0;
    break;;
  case VALID_SXP:
    log_error(LOG_INFORM, "auc-add", sn, rq,
	      "Adding secure uri pattern: %s",uri);
    curr->uri = STRDUP(uri);
    curr->isxp = 1;
    break;;
  case INVALID_SXP:
    log_error(LOG_MISCONFIG, "auc-add", sn, rq,
	      "Invalid URI spec: %s",uri);
    break;;
  default:
    /* huh? */
    log_error(LOG_FAILURE, "auc-add", sn, rq,
	      "Failed to validate uri: %s",uri);
    return -1;
  } /* validate uri */

  return 0;
} /* auc_add_securi */

int auc_add_aus(char *hostparm,Session *sn,Request *rq) {
  char *host, *charpt;
  long port;
  auc_sock *curr;

  /* parse hostparm */
  charpt = host = STRDUP(hostparm);
  while ( charpt && charpt[0] != ':' )
    charpt = charpt + 1;
  if ( charpt ) {
    charpt[0] = NULL;
    charpt = charpt + 1;
    /* FIXME: can we do this w/o resorting to libc? */
    port = atol(charpt);
  } else {
    port = AUC_DEF_PORT;
  }

  if ( !auc_socks ) {
    /* first item in list */
    curr = auc_socks = new_auc_sock();
  } else {
    /* append an item to list */
    while ( curr->next && curr->next->sock &&
	    curr->next->host )
      curr = curr->next;
    if ( curr->next )
      curr = curr->next;
    else 
      curr = curr->next = new_auc_sock();
  }

  /* init the bare necessities; the sockwatcher thread does the rest */
  log_error(LOG_INFORM,"auc-add", sn, rq,
	    "Adding AUS: %s:%i",host,port);
  curr->host = STRDUP(host); FREE(host);
  curr->port = port;
  curr->sock = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);

  return 0;
} /* auc_add_aus */

auc_sock *new_auc_sock(void) {
  auc_sock *s = (auc_sock *)MALLOC(sizeof(auc_sock));
  s->host = NULL;
  s->port = 0;
  s->isup = 0;
  s->sock = NULL;
  s->in_use = 0;
  s->next = NULL;
  return s;
} /* new_auc_sock */

auc_securi *new_auc_securi(void) {
  auc_securi *s = (auc_securi *)MALLOC(sizeof(auc_securi));
  s->uri = NULL;
  s->isxp = -1;
  s->next = NULL;
  return s;
} /* new_auc_securi */

void *auc_socket_watcher(void *_snrq) {
  auc_sock *s;
  struct sockaddr_in addr;
  unsigned int host_addr;
  struct hostent *h;
  char buf[AUC_BUFSIZ];
  int rv;
  struct snrq_st *snrq = (struct snrq_st *)_snrq;
  Session *sn = snrq->sn;
  Request *rq = snrq->rq;

  log_error(LOG_INFORM, "auc-sw", sn, rq,
	    "Starting to watch sockets");
  
  while (1) {
    for (s = auc_socks;
	 s;
	 s = s->next) {
      crit_enter(auc_sox_lock);

      if ( s->in_use ) {
	/* in use means ok */
	crit_exit(auc_sox_lock);
	continue;
      }
      s->in_use = 1;
      crit_exit(auc_sox_lock);
      
      errno = 0;
      /* let's just ping and see if we get a response */
      if ( s->isup && ! ( (rv=write(s->sock,"PING\n",5)) == 5 &&
			  read(s->sock,buf,AUC_BUFSIZ/*,AUC_AUS_TO*/) &&
			  /* FIXME: not a NSAPI call */
			  strncasecmp(buf,"PONG\n",5) == 0
			  ) ) {
	log_error(LOG_WARN, "auc-sw", sn, rq,
		  "Failed to PING AUS %s:%i - %s(%i) - %s",
		  s->host,s->port,buf,rv,strerror(errno));
	s->isup = 0;
      } else if ( s->isup ) {
	/*log_error(LOG_INFORM, "auc-sw", sn, rq,
	  "PING'd AUS. Host: %s:%i",
	  s->host,s->port);*/
	s->in_use = 0;
	continue;
      } /*else {
	log_error(LOG_WARN, "auc-sw", sn, rq,
		  "Connecting to AUS %s:%i",
		  s->host,s->port);
		  }*/

      /* if down, bring back up */
      close(s->sock);
      /* FIXME: not a NSAPI call */
      if ( !(h = gethostbyname(s->host)) ) {
	log_error(LOG_FAILURE, "auc-sw", sn, rq,
		  "Couldn't get host info for %s",s->host);
	s->in_use = 0;
	continue;
      }
      host_addr = *(unsigned int *)h->h_addr;
      addr.sin_family = AF_INET;
      addr.sin_port = htons(s->port);
      addr.sin_addr.s_addr = host_addr;
      s->sock = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);
      if ( s->sock < 0 ) {
	log_error(LOG_FAILURE, "auc-sw", sn, rq,
		  "Error creating socket!");
	s->in_use = 0;
	continue;
      }
      /*log_error(LOG_INFORM, "auc-sw", sn, rq,
	"Attempting connection...");*/
      if ( connect(s->sock,(struct sockaddr *)&addr,sizeof(addr)) < 0 ) {
	log_error(LOG_FAILURE, "auc-sw", sn, rq,
		  "Error connecting to AUS %s:%i - %s",
		  s->host,s->port,strerror(errno));
	close(s->sock);
	s->in_use = 0;
	continue;
      }
      log_error(LOG_INFORM, "auc-sw", sn, rq,
		"Connected to AUS %s:%i", s->host, s->port);
      s->isup = 1;
      s->in_use = 0;
      
    } /* foreach socket */
    systhread_sleep(AUC_SW_SLEEP_MIN*1000*60);

  } /* while 1 */
} /* auc_socket_watcher */
