#!/opt/perl/bin/perl -w
# $Id: muxDaemon.pm,v 1.4 2001/08/18 01:04:37 santana Exp $

package Net::muxDaemon;
require 5.004;

=pod

=head1 NAME

Net::muxDaemon -- a generic framework for network daemons.

=head1 SYNOPSIS

  use Net::muxDaemon;
  sub my_cmdprocessor {
    return $_[1]."\n";
  }
  new Net::muxDaemon(PROCESS_COMMAND => \&my_cmdprocessor);
  die "Huh? Net::muxDaemons aren't supposed to return!\n$!\n";

=head1 DESCRIPTION

C<Net::muxDaemon> implements all the boring, nasty details that go into
making a network daemon. It implements everything about a simple
message-response server except the processing of messages into
responses. This functionality, the I<guts> of the server, is
implemented by the user in a function.

The gist of C<Net::muxDaemon>' usage is simple. See L</SYNOPSIS> for all
the code you would need to implement an echo server. The user's
function will run as a C<Net::muxDaemon> method. It will have all the power
of any of other method with none of the worries. 

Note that C<Net::muxDaemon> is really only good for command
processing, almost RPC-like functionality. The same client can't have
more the one message being processed at a time without running some
serious risks. (Actually, the same client I<could>, but not on the
same socket.) These risks can be avoided by designing the protocol
around certain limitations; however, save yourself headaches and
stress by making sure all clients wait for responses after each
command.

=head1 EXPORTS

The C<syslog> function and C<$KILLED> are exported by default. To
export $LOGTAG and $LOGFACILITY, use the C<:syslog> export tag.

=over 4

=item C<syslog($priority, $format, @args)>

See L<Sys::Syslog>. This function does not log anything unless
C<$Net::muxDaemon::LOGTAG> is set to something. This C<Net::muxDaemon>
global is exported, so you can do C<$LOGTAG = 'myLogLine';> anytime
after C<use>ing C<Net::muxDaemon> to start logging. Alternatively,
if you can wait to start syslogging (as is typical if until you launch
the C<Net::muxDaemon> you can likely log to STDERR) you can pass
C<Net::muxDaemon> the LOGTAG option.

Basically, just call it all over the place. Then, if you want to log,
pass C<Net::muxDaemon> a LOGTAG option. Turn logging off from the
command_processor by doing C<undef($LOGTAG)>. Turn logging back on by
setting $LOGTAG to something.

Note that if you just want to change your syslog tag, changing $LOGTAG
is not enough. Do this: 

  undef $LOGTAG; 
  syslog(); 
  $LOGTAG = 'newtag';

=item C<$LOGTAG>

When this is undef, C<syslog> does nothing, otherwise it becomes the
tag used in loglines. May be initialized by the LOGTAG option. See
L</OPTIONS>.

=item C<$LOGFACILITY>

The syslog facility used for logging. May be initialized by the
LOGFACILITY option. See L</OPTIONS>. Defaults to 'daemon', but you can
set it to any valid value before you start logging. To change it after
having logged, do this:

  my($olt) = $LOGTAG; 
  undef $LOGTAG; 
  syslog(); 
  $LOGFACILITY = 'newfacility'; 
  $LOGTAG = $olt;

=item C<$KILLED>

Make this true and the Net::muxDaemon will exit nicely after the current
round of message processing.

=head1 CONSTRUCTOR

=over 4

=item new ( [Hash-o-options] )

Creates a C<Net::muxDaemon>. Options are passed by name. Never returns.

=back

=head1 ATTRIBUTES

=over 4

=item $this->{OPTION}

Options are stored in the hash, and changing some of them has no
effect after the object initializes. Some options, however, are safe
to change and begin effecting behavior immediately after any
changes. The initialize-only options will be marked
below. PROCESS_COMMAND is B<NOT> an initialize-only option. If you set
$this->{PROCESS_COMMAND}, the new setting will be in effect
immediately. Of course, the current function will run until it exits
as usual.

=item $this->{inbuffer}, $this->{outbuffer}, $this->{ready}, $this->{close}

These are for internal use. See L</getmore> and L</getreq> for a use
for $this->{inbuffer}. The others could conceivably be useful, but
these won't be documented except by the code that uses them.

=back

=head1 OPTIONS

=over 4

=item PROCESS_COMMAND

Code reference. If this option is undefined C<Net::muxDaemon> will
look for a sub named 'process_command', first in the caller's
namespace, then in the C<Net::muxDaemon> namespace. This might be
useful if you want your routine to I<live> in C<Net::muxDaemon>'s
namespace, but the benefits of this are obscure. The arguments to this
function will be C<($this,$request,$client)>, where:

=over 4

=item S<    C<$this>>

A reference to our own Net::muxDaemon object. This is where the
command processing routine gets access to static information.  See
L</ATTRIBUTES> for a list of useful attributes.

=item S<    C<$request>>

A scalar, containing the message read from the client.

=item S<    C<$client>>

A socket connected to the client. It is not necessary to send replies
to the client directly. This should be useful, for example, when more
complex communication needs to happen than a simple message-response.

=back

=item

The sub should return a string response to be sent back to the
client. If undef, nothing will be written to the
socket. Alternatively, it could return a hashref to tune the
response. Right now, the following keys are recognized:

=over 4

=item S<    C<reply>>

The reply string.

=item S<    C<CLOSE>>

If this is set, the socket will be closed after the reply is sent.

=back

Because perl's anonymous subroutines act as closures with respect to
lexical variables, you could conceivably use them to create your own
static workspace for your command processor. Because C<Net::muxDaemon>
provides you with everything you should need (Ha!), this is is left as
an exercise for the reader.

=item BASEDIR

Working directory - defaults to C<($ENV{'PWD'} || "/tmp")>. This is an
initialize-only option.

=item LOGTAG

Prepended to all log lines - logging is off unless set. This is an
initialize-only option. See the C<$LOGTAG> export above.

=item LOGFACILITY

Which facility to log to - defaults to "daemon". This is an
initialize-only option. See the C<$LOGFACILITY> export above.

=item PASSWD --NOT USED--

The admin password - defaults to "abc123"

=item INITHOOK

If set, and is a CODE ref, the referenced sub will be run just before
listening for clients.

=item DIEHOOK

If set, and is a CODE ref, the referenced sub will be run at exit.

=item HOOKSUB

If set, and is a CODE ref, the referenced sub will be run every HOOKINT
seconds.

=item HOOKINT

The interval at which the subroutine referenced by the HOOKSUB option
is run - defaults to 120.

=item TIMEOUT

Timeout in seconds for various operations - defaults to 3.

=item EOC

Command separator or command length. If this attribute is numeric, it
will determine the length of commands. Commands will be passed to the
PROCESS_COMMAND routine in chunks of EOC length. Otherwise it will
determine the end-of-command string. (Actually, it's an end-of-command
regex.)  Defaults to "\n".

=item GROUPNAME

Which group to run as - defaults to current gid. Initialize-only.

=item PORT

Which port to run on - defaults to 4321. Initialize-only.

=item HOST

Which host to listen on - defaults to '*'. Initialize-only.

=item PIDFILE

This file contains the process id - defaults to "NetDaemon.pid".
BASEDIR will be prepended unless "/" is the first
character. Initialize-only. B<Do not attempt to set this option from
the command processing function!>

=item POOL

An arryref to a hash of hosts. Unless you want to cluster your
C<Net::muxDaemon>, just use C<HOST> and C<PORT>. Takes the form of:

 POOL => [ { host => "hostname1",
             addr => "1.1.1.1:1234",
             sync_to => 10
           },
           { host => "hostname2",
             port => 1235,
             addr => "1.1.1.2",
             sync_to => 10
           }
         ]

The contents of {addr} will overwrite those of {host} (and {port}, if
that part of {addr} is supplied.) For C<Net::muxDaemon>, only {host},
{addr}, and {port} are used, and only those of the first locally
bindable spec. However, these keys of the hashes in the aryref are the
same as the C<Tie::CacheHash> option {sync} (See L<Tie::CacheHash>),
so you may just want to fill it in so you can send it to both modules.

C<POOL> overrides the C<PORT> and C<HOST> options, if present. At
least one of the host specs I<must> be bindable-to. This allows you to
specify more than one server on the same machine. The servers will use
C<Tie::CacheHash> syncing to keep each other up to date with local
changes. C<Tie::CacheHash> syncs on the udp port specified, while
C<Net::muxDaemon> listens on the tcp port of the same number.

=back

=head1 METHODS

These methods are not likely to be called except from the command
processing function.

=over 4

=item C<getmore(socket, bytecount, ref_to_scalar)>

Retrieves C<bytecount> bytes from C<socket> and places them in the
scalar referenced by C<ref_to_scalar>.

=item C<getreq(socket, ref_to_scalar)>

Retrieves a request from C<socket> and places it in the scalar
referenced by C<ref_to_scalar>. The request will not be sent to the
command processor after this. (Unless it is pushed into the %ready
attribute manually by the caller.)

=back

=cut


use strict;

# import method
use Exporter;
# OO socket interface
use IO::Socket;
# OO file interface
use IO::File();
# contains socket constants
use POSIX;
# so we can multiplex sockets
use IO::Select;
# allows us to put the sockets into not blocking mode
use Fcntl;
# I forgot what this does but it must be kewl
use Sys::Syslog ();
# Use hashes with refs as keys; we use sockets as keys
use Tie::RefHash;

# private lexical for Net::muxDaemon::syslog
my($LOGGING) = 0;

# Global vars for my Net::muxDaemon::syslog & Exporter.
use vars qw($KILLED
	    $LOGTAG $LOGFACILITY
	    $VERSION @ISA @EXPORT @EXPORT_OK %EXPORT_TAGS);
$KILLED = 0;
$LOGTAG = undef;
$LOGFACILITY = undef;
$VERSION = do {my @r=(q$Revision: 1.4 $ =~ /\d+/g);sprintf("%d."."%02d" x $#r,@r)}; # must be all one line, for MakeMaker
@ISA = qw(Exporter);
@EXPORT = qw(syslog $KILLED);
@EXPORT_OK = qw(syslog $KILLED $LOGTAG $LOGFACILITY);
%EXPORT_TAGS = ( syslog => [ qw(syslog $KILLED $LOGTAG $LOGFACILITY) ] );

# constructor
sub new {
  my($pkg) = shift();
  my(%this) = @_;
  my($this) =  %this?\%this:{};
  bless($this,$pkg);
  
  # this package
  $this->{class} = $pkg;
  # working directory
  $this->{BASEDIR} = ($ENV{'PWD'} || '/tmp')
      unless $this->{BASEDIR};
  # which facility to log to
  $this->{LOGFACILITY} = 'daemon'
      unless $this->{LOGFACILITY};
  $LOGFACILITY = $this->{LOGFACILITY};
  # my Net::muxDaemon::syslog needs this to be 
  #  available without $this
  $LOGTAG = $this->{LOGTAG} if exists($this->{LOGTAG});
  # the admin password
#  $this->{PASSWD} = "abc123"
#    unless $this->{PASSWD};
  # Command separator -OR- command length
  $this->{EOC} = "\n"
      unless $this->{EOC};
  # run this sub every {HOOKINT} seconds
  $this->{HOOKSUB} = undef
      unless ref($this->{HOOKSUB}) eq 'CODE';
  # the interval at which to run {HOOKSUB}
  $this->{HOOKINT} = 120
      unless $this->{HOOKINT};
  # run this sub at init
  $this->{INITHOOK} = sub { ;; }
      unless ref($this->{INITHOOK}) eq 'CODE';
  # run this sub at exit
  $this->{DIEHOOK} = sub { ;; }
      unless ref($this->{DIEHOOK}) eq 'CODE';
  # the timeout for various operations
  $this->{TIMEOUT} = 3
      unless $this->{TIMEOUT};
  # which port to run on
  $this->{PORT} = 4321
      unless $this->{PORT};
  # which host to listen on
  $this->{HOST} = undef
      unless $this->{HOST};
  # aryref to server pool host specs
  if ( $this->{POOL} ) {
    $this->{HOST} = $this->{PORT} = undef;
  } else {
    $this->{POOL} = undef;
  }
  # this file contains the process id
  $this->{PIDFILE} = "$this->{BASEDIR}/NetDaemon.pid"
      unless $this->{PIDFILE};
  
  # begin with empty buffers
  $this->{inbuffer}  = {};
  $this->{outbuffer} = {};
  $this->{ready}     = {};
  $this->{close}     = {};
  tie %{$this->{ready}}, 'Tie::RefHash';
  tie %{$this->{close}}, 'Tie::RefHash';
  
  # some process initializations
  $this->_init();  
  
  my @caller = caller();
  # syslog gets set up in _init()
  if ( exists($this->{PROCESS_COMMAND}) && ref($this->{PROCESS_COMMAND}) eq 'CODE' ) {    
    syslog('debug','Installing user-defined command processor');
  } elsif ( defined(&process_command)  ) {
    $this->{PROCESS_COMMAND} = \&process_command;
    syslog('debug','Installing command processor in Net::muxDaemon namespace');
  } elsif ( defined(&{$caller[0].'::process_command'}) ) {
    $this->{PROCESS_COMMAND} = \&{$caller[0]."::process_command"};
    syslog('debug',"Installing command processor in $caller[0] namespace");
  } else {
    $this->{PROCESS_COMMAND} = sub { $KILLED = 1 if $_[1] =~ /QUIT/; return $_[1]."\n"; };
    syslog('warning','Using default PROCESS_COMMAND for new Net::muxDaemon');
  }
  
  # vroom..
  $this->_run();
  
} # sub new

# setup
sub _init {
  my($this) = shift();
  
  # handle sigs
  $SIG{TERM} = sub { $KILLED = 1; };
  $SIG{HUP}  = sub { $KILLED = 1; };
  
  # run the init_hook
  &{$this->{INITHOOK}};
  
  # open the listener
  if ( $this->{PORT} ) {
    unless ( $this->{server} =
	     new IO::Socket::INET( LocalAddr => $this->{HOST},
				   LocalPort => $this->{PORT},
				   Proto => 'tcp',
				   Listen => SOMAXCONN,
				   Reuse => 1) ) {
      die 'Error creating socket';
    }
    
  } elsif ( $this->{POOL} ) {
    # go through the POOL, bind to one of them
    syslog('debug','Searching through POOL');
  SERVER: foreach my $server ( @{$this->{POOL}} ) {
    die('No "host" or "addr" in sync server spec!')
	unless exists($server->{addr}) || exists($server->{host});
    $server->{host} = gethostbyaddr($server->{addr},AF_INET)
	unless exists($server->{host});
    $server->{port} = $this->{sync_port}
	if !exists($server->{port});
    if ( exists($server->{addr}) ) {
      if ( $server->{addr} =~ /([^:]+):([0-9]+)/ ) {
	$server->{addr} = $1;
	$server->{port} = $2;
      }
    } else {
      my(@hostent,@addrs);
      @hostent = gethostbyname($server->{host});
      @addrs = $hostent[4];
      if ( !$addrs[0] ) {
	die("No host address for $server->{host} - @hostent");
      }
      $server->{addr} = inet_ntoa($addrs[0]);
    }

    eval {
      local $SIG{__WARN__} = sub { ;; };
      $this->{server} = new IO::Socket::INET
	  ( LocalAddr => $server->{addr},
	    LocalPort => $server->{port},
	    Proto     => 'tcp',
	    Listen    => SOMAXCONN,
	    Reuse     => 1 );
    };
    if ( $this->{server} && $this->{server}->isa('IO::Socket::INET') ) {
      last SERVER;
    } else {
      $this->{server} = undef;
    }
  } # SERVER
    die('Couldn\'t bind to any spec in the POOL option')
	unless $this->{server} && $this->{server}->sockport();
    
  } else {
    die('Uhh? No PORT nor POOL?!?');
  } # open listener
  
  syslog( 'debug','Connected to ' .
	  $this->{server}->sockhost().':'.$this->{server}->sockport() );
  
  # change out of the root UID/GID
  $> = scalar(getpwnam($this->{USERNAME}))
      if $this->{USERNAME};
  $) = scalar(getgrnam($this->{GROUPNAME}))
      if $this->{GROUPNAME};
  
  # write the process id to the pid file
  my($file) = new IO::File;
  unless ( $file->open(">$this->{PIDFILE}") ) {
    syslog('warning',"Cannot open pid file $this->{PIDFILE}: %m");
  } else {
    $file->print("$$\n");
    $file->close;
  }
  
  # put the listener in nonblocking mode
  nonblock($this->{server});
  # add the listener to a new select object
  $this->{sels} = IO::Select->new($this->{server});
} # sub _init  

# this should never return
sub _run {
  my($this) = shift();
  
  # initalize the save interval time
  my($hook_time) = time;
  
  
  # Main loop: check reads/accepts, check writes, check ready to process
  my($client);
 EVENT_LOOP: while (1) {
   
   # Because we keep all sockets non-blocking, we can be event-driven
   #   %ready contains requests ready to process as    
   #     ( socket => "command" ) where "command" includes
   #     $this->{EOC}     
   #   %outbuffer contains replies ready to write -- *nothing* is
   #     changed in what's returned by $this->{PROCESS_COMMAND},
   #     except that undef's are turned into ''
   #   %inbuffer contains (possibly partial) commands recently read
   #   socket events are checked for based on socket read/write
   #     availability 
   
   # anything to read or accept?
   $this->_read_sockets();
   # Any complete requests to process?
   $this->_handle($client);
   # Buffers to flush?
   $this->_write_sockets();
   # Sockets to close?
   $this->_close_sockets();
   
 } continue {
   # non-socket events are dealt with here
   
   # if we have been killed
   if( $KILLED ) {
     &{$this->{DIEHOOK}};
     defined($this->{server}) && $this->{server}->close;
     $this->{server} = undef;
     #syslog('info','Restarting.');
     #$this->_init;
     unlink($this->{PIDFILE});
     syslog('debug','Exiting');
     exit(0);
   }
   
   # if the save interval has passed
   if ( $this->{HOOKSUB} && $hook_time + $this->{SAVEINT} < time ) {
     # run the hook
     &{$this->{HOOKSUB}};
     $hook_time = time;
   }
 } # EVENT_LOOP
  
} # sub _run

# _read_socket($this) checks for sockets to read from
sub _read_sockets {
  my($this) = shift();
 READABLE_SOCKET: foreach my $client ($this->{sels}->can_read(1)) {
   
   if ($client == $this->{server}) {
     # accept a new connection
     $client = $this->{server}->accept();
     $this->{sels}->add($client);
     nonblock($client);
     syslog("debug","Accepted a new connection");
     next READABLE_SOCKET;
   }

   my($data) = '';
   my($rv) = $client->sysread($data, POSIX::BUFSIZ);
   unless ( defined($rv) && length($data) == $rv ) {
     # This would be an error
     $this->{close}{$client} = 'Error reading socket. Closing.';
     next READABLE_SOCKET;
   }
   unless ( $rv ) {
     # This would be eof
     $this->{close}{$client} = '';
     next READABLE_SOCKET;
   }       
   $this->{inbuffer}{$client} .= $data;
   
   # test whether the data in the buffer or the data we
   # just read means there is a complete request waiting
   # to be fulfilled.  If there is, set $ready{$client}
   # to the requests waiting to be fulfilled.
   # NOTE: read only one request at a time to allow for
   # protocols with arbitrary data after EOC, e.g.,
   # IMG filename\nimage-data -- in this case the app
   # has to make sure and leave the beginning of the next
   # command on the socket...
   if ( $this->{EOC} =~ m/^[0-9]+$/ ) {
     if ( $this->{inbuffer}{$client} =~ s/^(.{$this->{EOC}})// ) {
       push(@{$this->{ready}{$client}}, $1);
     }
   } else {
     if ( $this->{inbuffer}{$client} =~ s/^(.*?$this->{EOC})// ) {
       push(@{$this->{ready}{$client}}, $1);
     }
   }
 } # READABLE_SOCKET
} # sub _read_sockets  

# _write_sockets($this) checks for writeable sockets, and writes to
# them if there's anything to say
sub _write_sockets {
  my($this) = shift();
  
 WRITEABLE_SOCKET: foreach my $client ($this->{sels}->can_write(1)) {
   # Skip this client if we have nothing to say
   next WRITEABLE_SOCKET
       unless exists $this->{outbuffer}{$client};
   unless ( defined($this->{outbuffer}{$client}) &&
	    length($this->{outbuffer}{$client}) ) {
     delete $this->{outbuffer}{$client};
     next WRITEABLE_SOCKET;
   }
   my($rv) = $client->syswrite($this->{outbuffer}{$client},
			       length($this->{outbuffer}{$client}));
   unless ($rv) {
     $this->{close}{$client} = "Failed write on writeable socket?";
     next WRITEABLE_SOCKET;
   }
   # FIXME: terrible error handling
   if ( $rv == length($this->{outbuffer}{$client}) ||
	$! == POSIX::EWOULDBLOCK ||
	(syslog("debug","ERRNO at send: ".$!) && $! == 0) ) {
     substr($this->{outbuffer}{$client}, 0, $rv) = '';
     delete $this->{outbuffer}{$client}
         unless length($this->{outbuffer}{$client});
   } else {
     # Write failure for a reason other than EWOULDBLOCK
     $this->{close}{$client} = 'Write failed. len=' .
	    length($this->{outbuffer}{$client}) . ' rv=' . $rv .
	    'Shutting down socket.';
   }
 } # WRITEABLE_SOCKET
} # sub _write_sockets

# _close_sockets($this) closes sockets so scheduled
sub _close_sockets {
  my($this) = shift();

 CLOSEABLE_SOCKET: foreach my $client ( keys %{$this->{close}} ) {
   syslog("info",$this->{close}{$client})
       if $this->{close}{$client};
   delete($this->{inbuffer}{$client});
   delete($this->{outbuffer}{$client});
   delete($this->{ready}{$client});
   delete($this->{close}{$client});
   $this->{sels}->remove($client);
   $client->close;
 } # CLOSEABLE_SOCKET
} # sub _close_sockets

# _handle($socket) deals with all pending requests for $client
sub _handle {
  # requests are in $ready{$client}
  # send output to $outbuffer{$client}
  my ($this) = shift();
  my ($client,$request,$reply);
  
  CLIENT: foreach $client (keys %{$this->{ready}}) {
    REQUEST: foreach $request (@{$this->{ready}{$client}}) {
      # $request is the text of the request
      # put text of reply into $outbuffer{$client}
      $reply = $this->{PROCESS_COMMAND}->($this,$request,$client);
      if ( ref($reply) eq 'HASH' ) {
	$this->{close}{$client} = $reply->{CLOSE}
	    if exists($reply->{CLOSE});
	$this->{outbuffer}{$client} .= $reply->{reply}
	    if exists($reply->{reply});
	next REQUEST;
      }
      $this->{outbuffer}{$client} .= $reply
	  if defined($reply);
    } # REQUEST
    delete $this->{ready}{$client};
  } # CLIENT
} # sub _handle

# nonblock($socket) puts socket into nonblocking mode
sub nonblock {
  my $socket = shift;
  my $flags;
  
  $socket->autoflush(1);
  syslog("warning","Can't get flags for socket: %m")
      unless $flags = fcntl($socket, F_GETFL, 0);
  syslog("warning","Can't make socket nonblocking: %m")
      unless fcntl($socket, F_SETFL, $flags | O_NONBLOCK | O_SYNC);
} # sub nonblock

# utility for command processors to get another request from the socket
sub getreq {
  my($this,$socket,$bufref) = (shift(),shift(),shift());
  my($rv,$now);
  
  $$bufref = '';
  
  # is there a req waiting for us in %ready?
  if ( $this->{ready}{$socket}[1] ) {
    $$bufref = $this->{ready}{$socket}[1];
    $this->{ready}{$socket}[1] = undef;
    return 1;
  }
  
  # check out %inbuffer
  if ( $this->{EOC} =~ m/^[0-9]+$/ ) {
    if ($this->{inbuffer}{$socket} =~ s/(.{$this->{EOC}})//) {
      # only peel *one* command from %inbuffer
      $$bufref = $1;
      return 1;
    } else {
      # %inbuffer contains less than one command
      $$bufref = $this->{inbuffer}{$socket};
      $this->{inbuffer}{$socket} = '';
    }
  } else {
    if ($this->{inbuffer}{$socket} =~ s/(.*$this->{EOC})//) {
      # only peel *one* command from %inbuffer
      $$bufref = $1;
      return 1;
    } else {
      # %inbuffer contains less than one command
      $$bufref = $this->{inbuffer}{$socket};
      $this->{inbuffer}{$socket} = '';
    }
  }
  
  # will the client give us the rest?
  # put a debug routine here when/if you have the time
  $now = time;
 READ_SOCK: while(time < $now + $this->{TIMEOUT}) {
   $socket->sysread($$bufref,1,length($$bufref));
   if ( $this->{EOC} =~ m/^[0-9]+$/ ) {
     if ($$bufref =~ m/.{$this->{EOC}}/) {
       last READ_SOCK;
     }
   } else {
     if ($$bufref =~ m/.*$this->{EOC}/) {
       last READ_SOCK;
     }
   }
 } # READ_SOCK
  return 1;
} # sub getreq


# utility for command processors to get more data from the socket
sub getmore {
  my($this,$socket,$bc,$bufref) = (shift(),shift(),shift(),shift());
  my($rv,$flags,$now,$count,$cmd);
  
  $$bufref = '';

  # FIXME: look through %ready first...
  # is what we're looking for in %inbuffer?
  $$bufref = substr($this->{inbuffer}{$socket},0,$bc);
  substr($this->{inbuffer}{$socket},0,$bc) = '';
  $rv = length($$bufref);
  if ( $rv < $bc ) {
    # is the rest of it in %ready?
    # start at 1, 0 is the one we're doing...
    for ($count=1; $count <= $#{$this->{ready}{$socket}}; $count++ ) {
      $now = $this->{ready}{$socket}[$count];
      $$bufref .= substr($now,0,$bc-$rv);
      substr($now,0,$bc-$rv) = '';
      shift(@{$this->{ready}{$socket}})
	  if $now eq '';
      $rv = length($$bufref);
      last
	  unless $rv < $bc;
    }
  }
  if ( $rv < $bc ) {
    # will the client give us the rest?
    syslog("warning","Can't get flags for socket: %m")
	unless $flags = fcntl($socket, F_GETFL, 0);
    syslog("warning","Can't make socket blocking: %m")
	unless fcntl($socket, F_SETFL, O_NONBLOCK ^ $flags);
    # put a debug routine here when/if you have the time
    $now = time;
    while($rv < $bc && time < $now + $this->{TIMEOUT}) {
      $rv += $socket->sysread($$bufref, $bc - $rv, $rv);
    }
    syslog("warning","Can't make socket nonblocking: %m")
	unless fcntl($socket, F_SETFL, $flags);
    unless (defined($rv) && length $$bufref) {
      # This would be the end of file, so close the client
      $this->{close}{$socket} = '';
      return 0;
    }
  }
  return 1;
} # sub getmore

# this wraps around Sys::Syslog::syslog so we can turn
# debugging on and off
sub syslog {
  if ( defined($LOGTAG) ) {
    if ( ! $LOGGING ) {
      &Sys::Syslog::openlog($LOGTAG,'nowait',$LOGFACILITY);
      &Sys::Syslog::setlogsock('unix');
      &Sys::Syslog::syslog('debug','Starting to log');
      $LOGGING = 1;
    }
    return &Sys::Syslog::syslog(@_);
  }
  
  else {
    if ( $LOGGING ) {
      &Sys::Syslog::closelog();
      $LOGGING = 0;
    }
    return undef;
  }
} # sub syslog


1;
