#!/bin/false
# $Id: DistHash.pm,v 1.3 2001/08/24 23:12:22 santana Exp $
#
# Tie::DistHash needs to be thread-safed before Net::Daemon is
# threaded. See Net::Daemon source for and thoughts on implementing
# threads there.
#
# NOTE: The need for threading may die when (if?) we move the "real"
#       Net::Daemon 
#
#     NOTES on threading Tie::DistHash
#
# 1. flock all files before using (just to be safe -- this might be
#    overkill sometimes. We shouldn't be using files often anyway...)
#
# 2. row ops (read/modify existing row) locking:
#    {lock $this->{_row_locks}; $this->{_row_locks}++;}
#    {lock $this->{index}{$key}; ..operate...}
#    {lock $this->{_row_locks}; $this->{_row_locks}--;}
#
# 3. append (new row) locking:
#    {lock $this->{_append_locking}; ..operate...}
#
# 4. database-wide ops (clear,destroy,init,etc) locking:
#    {
#      lock $this->{_append_lock};
#      lock $this->{_row_locks};
#      while ($this->{_row_locks}) {yeild;}
#      ..operate...
#    }
#
package Tie::DistHash;

=pod

=head1 NAME

C<Tie::DistHash> -- a C<tie> implementation facilitating a shared
cache.

=head1 SYNOPSIS

  use Tie::DistHash;
  my(%hash);
  my($cache) = tie(%hash,'Tie::DistHash',
		   ckptfile => 'mycache.dat',
		   gc_at_ckpt => 0);
  $hash{session_key} = {arbitrary=>[$data,\%structure],including=>$objects}; 
  ...keys %hash...values %hash...
  foreach my($key,$item_as_string) ( $cache->eachString ) {
    print "$key likely has objects!\n"
      if $item_as_string =~ m/bless\(/;
  }


=head1 DESCRIPTION

C<Tie::DistHash> is a C<tie()> implementation with some useful extras
for implementing a cache, like checkpointing, garbage collection, and
parsing of arbitrary data structures to strings that can be passed
to eval.

=head1 CONSTRUCTOR

Tying a hash to this class will return an initialized C<DistHash>
object. Use this object to run methods (see L</METHODS>) and set
attributes (see L</ATTRIBUTES>). Using the hash will change the
object. 

=head1 METHODS

Most of what is likely to get done with DistHash is as
straightforward as using a normal hash (once that hash has been
C<tie>d). Some things that are likely to get done fairly regularly,
though, are difficult to implement purely through a C<tie>
interface. So it becomes a handy-dandy thing that C<tie> returns an
object, which can have methods, etc, etc. Tying to DistHash returns
an object with the following methods:

=over 4

=item C<set_timeout $timeout [,$key]>

Sets the expiration interval to $timeout, the default for items
subsequently added, or for just one item if $key is supplied.

=item C<get_timeout [$key]>

Returns the default expiration interval, or the expiration interval
for just one item if $key is supplied.

=item C<get_atime $key>

Returns the last time C<$key> was accessed. If C<time> is greater than
C<get_atime($key)> and C<get_timeout($key)> combined, the next garbage
collection will remove C<$key>.


=item C<checkpoint [options]>

Performs a checkpoint. C<options> is a hash that can override default
attributes affecting checkpoints. See L</ATTRIBUTES>.

=item C<collect_garbage>

Goes through garbage_collection. Usually taken care of during
checkpoints.

=item C<integrity_check>

Performs an integrity check. This should only be necessary during
development, or conceivably once in a while for mission-critical apps. 

=item C<toString>

Returns the entire hash as a string that can be passed to C<eval>. Be
aware that it always returns the I<entire> hash...

=item C<eachString>

Like C<each>, but returns the next item as a string that can be passed
to C<eval>. Two out of three dentists prefer this over C<toString>.

=item C<init_from FH>

Takes the filehandle of an open checkpoint file and initializes the
hash from it.

=item C<undump>

Initializes the database from a string like the ones created by
C<toString>. Try to use C<eachString> and C<eval> rather than
C<toString> and <undump> for large hashes.

=item C<sync KEY>

Sends a sync update of KEY to the sync peers. Useful if C<$db{KEY}> is
a ref to something (i.e., C<$db{KEY} = {name=>"who",phone=>5551234}>),
and something inside that something has changed (i.e.,
C<$db{KEY}{name}="whoelse">). In this case, the value of C<$db{KEY}>
stays the same, and the tie implementation sees no only the access of
KEY, so only a TOUCH sync happens. After making a change like that,
C<$dbobj->sync(KEY)> would make sure that everyone in the sync pool
sees the change.

=back

=head1 OPTIONS

Most of the behavior of a hash C<tie>d to C<DistHash> is controlled
by options. Options are usually set at initialization, when the object
is C<tie>d. Simply pass a hash to C<tie()> after the first two
arguments. See L</SYNOPSIS>.

=over 4

=item v

Verbose flag, warning messages are on if true. Defaults to false.

=item def_expint

The default expiration interval for items in this hash. 0 means
never expire anything. Defaults to 7200 (2 hours).

=item basedir

Working directory. Default location for checkpoint files. Defaults to
the contents of the PWD environment variable, or "/tmp".

=item ckptfile

Checkpoint file. '' means never write checkpoints. Defaults to
F<CH_dump-PID.dat>.

=item gc_at_ckpt

Whether to collect garbage at checkpoints. Defaults to 1.

=item ic_at_gc

Whether to check integrity at garbage-collection. This does I<NOT>
apply to the garbage collection at checkpoints! Defaults to 1.

=item ic_at_ckpt

Whether to check integrity at checkpoints. (gc_at_ckpt has no effect
on this.) Default to 0.

=item sync

An arryref to a hash of hosts. See L</SYNCHRONIZATION>. Takes the
form of 

 sync => [ { host => "hostname1", 
             addr => "1.1.1.1:1234",
             sync_to => 10
           },
           { host => "hostname1", 
             port => 1234, addr => "1.1.1.1",
             sync_to => 10
           }
         ]

The contents of {addr} will overwrite those of {host} (and {port}, if
that part of {addr} is supplied.)

One of the elements of this array will usually specify a local
address, and there will usually be one or more remote
specifications. If there is no local spec that can be bound to, we
will I<only send> syncs (see L</sync_checkint>.)

=item sync_port

Default sync port. Defaults to 8789. See L</SYNCHRONIZATION>.

=item sync_to

Default UDP timeout for syncing. Defaults to 1. See
L</SYNCHRONIZATION>.

=back

=item sync_atime

Minimum time between touches. Defaults to 60. See
L</SYNCHRONIZATION>.

=item sync_checkint

Time between checks for inbound sync commands. Defaults to 3. Set to
0 to enable I<send-only> mode. See L</SYNCHRONIZATION>.

=back

=item sync_state_retrieval_hook

A CODEREF to run if no peers respond to our full sync requests upon
startup. If it returns 0 or doesn't exist, state is retrieved from a
checkpoint file.

=back

=item sync_events

A hashref of whether to sync various events. See
L</SYNCHRONIZATION>. Defaults to

 sync_events => { ACCESS  => 1,	# to refresh access time for timeouts 
                  DELETE  => 1,
                  MODIFY  => 1,
		  ADD     => 1,
		  SETTO   => 1, # when setting a timeout
                  CLEAR   => 0, # eg, before assigning to the hash
		  DESTROY => 0, # end of tie
		  CKPT    => 0, # at checkpoints
		  INTCK   => 0, # at integrity checks
		  GC      => 0  # at garbage collection
                }

B<NOTE>: CKPT, INTCK and GC are not yet implemented.

=back

=head1 SYNCHRONIZATION

At startup, the C<DistHash> tries to retrieve state via a full sync
from one of the servers in the pool. It will accept sync commands from
other servers while this full sync completes. If no server in the pool
is reachable, the CODEREF C<sync_state_retrieval_hook> is run. If it
returns 0 or doesn't exist, then state is retrieved from the
checkpoint file. Only after a full state retrieval is complete will
the tie operation return.

When an event occurs that requires synchronization per sync_events, a
command is sent to each server in the sync pool.

I<Note:> Be aware that only events actually handled by the tie can
sync automatically. If a value is a reference, and something in what's
referenced changes, there's no way for a tie to know
automatically. See L</sync> for a way to tell C<Tie::DistHash>
manually.

For fetches and other ATIME-only updates, a touch is sent; no more
than one every C<sync_atime> seconds. For data inserts, updates and
deletes, an appropriate command is sent, basically duplicating the
command received by the server performing the sync.

Sync commands are checked for every C<sync_checkint> seconds. This
mechanism is built using SIGALRM, so if you use sync, I<be careful>
when you set or modify C<$SIG{ALRM}>!

Synchronization at this time is accomplished in the blind UDP way. How
trustworthy this is depends on how trustworthy your network is. This
also implies a limit to the size of a value before you get errors
about too-large message sizes. Messages as large as 11K work fine, but
this limitiation will hopefully be lifted in the future.

=cut


use strict;
use Data::Dumper;
use Fcntl;
use IO::File();
use IO::Socket;
use Carp;

# columns for the database array
use constant ATIME => 0;
use constant TTIME => 1;
use constant EXPINT => 2;
use constant KEY => 3;
use constant VALUE => 4;

# Some useful defaults
use constant DEF_EXPINT => 0;
use constant DEF_BASEDIR => ($ENV{'PWD'} || "/tmp");
use constant DEF_CKPTFILE => "CH_dump-$$.dat";

# FIXME: is this a good idea? I *think* it's part of the tie spec, but
# wouldn't that make it hard to have more than one tied var in a
# namespace? 
#my(@EXPORT) = qw(TIEHASH "DESTROY" CLEAR FETCH STORE
#		 DELETE EXISTS FIRSTKEY NEXTKEY);

# our warning printer
sub dh_warn {
    my($this,$msg)=(shift(),shift());
    carp($this->{class}.$msg) if $this->{v};
}

#
# constructor method
sub TIEHASH {
    my($classname) = shift();
    my(%this) = @_;

    # the object
    my($this) = %this ? \%this : {};
    bless($this,$classname);
    # the database array
    $this->{"database"} = [];
    # the keyed index
    $this->{"index"} = {};
    # verbosity
    $this->{v} = undef
	unless defined($this->{v});
    # useful for logging &etc.
    $this->{class} = $classname
	unless defined($this->{class});
    # set default expiration
    $this->{def_expint} = DEF_EXPINT
	unless defined($this->{def_expint});
    # checkpoint file
    $this->{basedir} = DEF_BASEDIR
	unless $this->{basedir};
    $this->{ckptfile} = DEF_CKPTFILE
	unless $this->{ckptfile};
    # initialize last checkpoint time
    $this->{last_ckpt} = time;

    # the DataDumper we'll use for this object
    # FIXME: we should use an object -- what if these class vars are
    # reset elsewhere? 
    $Data::Dumper::Purity   = 1;
    $Data::Dumper::Terse    = 0;
    $Data::Dumper::Indent   = 0;
    $Data::Dumper::Deepcopy = 0;

    # initialize the options on this object
    $this->{gc_at_ckpt} = 1         # incorporate garbage collection at
	unless $this->{gc_at_ckpt}; # checkpoints
    $this->{ic_at_gc} = 1           # do an integrity check at garbage
	unless $this->{ic_at_gc};   # collection (not at checkpoints!)
    $this->{ic_at_ckpt} = 1         # do an integrity check at checkpoints
	unless $this->{ic_at_ckpt};

    # set up sync 
    $this->{"sync"} = []
	unless $this->{"sync"};
    $this->{sync_events} = {}
    unless $this->{sync_events};
    if ( $#{$this->{"sync"}} > 0 ) {
	$this->sync_init();
    } elsif ( exists($this->{ckptfile}) ) {
	# init from ckpt file
	my($fh) = new IO::File;
	if ( open($fh,$this->{ckptfile}) ) {
	    $this->init_from($fh);
	    close($fh);
	}
    } # set up sync
    
    return $this;
} # sub TIEHASH

#
# destructor method - returns ref to destroyed object
# FIXME: checkpoint, close out sockets, etc.
sub DESTROY {
    my($this) = shift();
    $this->sync_bcast("DESTROY\n")
	if $this->{sync_events}{"DESTROY"};
    $this->dh_warn(' Exiting');
    #delete(@$this{"database","index","class","def_expint"});
    #undef(%$this);
    return $this;
} # sub DESTROY

#
# spec(?) method - returns undef, clears all values
sub CLEAR {
    my($this) = shift();
    @{$this->{database}} = ();
    %{$this->{"index"}} = ();
    $this->sync_bcast("CLR\n")
	if $this->{sync_events}{CLEAR};
} # sub clear

#
# spec method - returns a value
sub FETCH {
    my($this,$key) = (shift(),shift());
    my($value);
    eval {
	if ( !exists($this->{'index'}{$key}) ||
	     !defined($this->{'database'}[$this->{'index'}{$key}]) ) {
	    # oops, the row wasn't found
	    $value = undef;
	    die("no such key $key");
	} elsif ( $this->{'database'}[$this->{'index'}{$key}]->[EXPINT]
		  && time >= 
		     $this->{'database'}[$this->{'index'}{$key}][ATIME] + 
		     $this->{'database'}[$this->{'index'}{$key}]->[EXPINT] ) {
	    # oops, the row has timed out
	    $this->DELETE($key);
	    $value = undef;
	    die("$key has expired");
	} else {
	    # hooray, the row exists
	    $value = $this->{"database"}[$this->{"index"}{$key}][VALUE];
	    $this->{"database"}[$this->{"index"}{$key}][ATIME] = time();
	    if ( $this->{sync_events}{ACCESS} &&
		 time > $this->{"database"}[$this->{"index"}{$key}][TTIME]
		 + $this->{sync_atime} ) {
		$this->sync_bcast("TOUCH $key\n");
		$this->{"database"}[$this->{"index"}{$key}][TTIME] = time();
	    }
	}
    };
    $this->dh_warn("::FETCH failed: $@")
	if $@ ne '';
    return $value;
} # sub FETCH

#
# spec method - returns value of $hash{$key} after the STORE attempt
sub STORE {
    my($this,$key,$val) = (shift(),shift(),shift());
    $this->dh_warn("::STORE: $key .$val.");
    eval {
	my($ekey) = $key;
	$ekey =~ s/([\\\'])/\\$1/g;
	if ( exists($this->{"index"}{$key}) ) {
	    $this->{database}[$this->{"index"}{$key}][VALUE] = $val;
	    $this->{database}[$this->{"index"}{$key}][EXPINT] = 
		$this->{def_expint};
	    $this->{database}[$this->{"index"}{$key}][ATIME] = 
		$this->{database}[$this->{"index"}{$key}][TTIME] = time();
	    $this->sync_bcast("MOD '$ekey' ".Data::Dumper->Dump([$val])."\n")
		if $this->{sync_events}{MODIFY};
	} else {
	    push(@{$this->{"database"}},
		 [time(), time(), $this->{def_expint}, $key, $val]);
	    $this->{"index"}{$key} = $#{$this->{"database"}};
	    $this->sync_bcast("ADD '$ekey' ".Data::Dumper->Dump([$val])."\n")
		if $this->{sync_events}{ADD};
	}
    };
    $this->dh_warn("::STORE failed: $@")
	if $@ ne '';
    return $this->{"database"}[$this->{"index"}{$key}][VALUE];
} # sub STORE

#
# spec method - returns $hash{$key} before the DELETE attempt
sub DELETE {
    my($this, $key) = (shift(),shift());
    $this->dh_warn("::DELETE $key");
    my($value);
    eval {
	die "key $key not found.\n"
	    unless exists($this->{"index"}{$key});
	$value = $this->{"database"}[$this->{"index"}{$key}][VALUE];
	$this->{"database"}[$this->{"index"}{$key}] = undef;
	delete($this->{"index"}{$key});
	$this->sync_bcast("DEL $key\n")
	    if $this->{sync_events}{DELETE}; 
    };
    $this->dh_warn("::DELETE failed: $@")
	if $@ ne '';
    return $value;
} # sub DELETE

#
# spec method - returns 1/undef
sub EXISTS {
    my($this,$key) = (shift(),shift());
    $this->dh_warn("::EXIST $key - ".(exists($this->{"index"}{$key})?'y':'n'));
    return exists($this->{"index"}{$key});
} # sub EXISTS

#
# spec method - resets key counter, returns next key
sub FIRSTKEY {
    my($this) = shift();
    keys %{$this->{"index"}};
    my($key,$ind) = each(%{$this->{"index"}});
    return $key;
} # sub FIRSTKEY

#
# spec method - returns next key, advances key counter
sub NEXTKEY {
    my($this,$lastkey) = (shift(),shift());
    my($key,$ind) = each(%{$this->{"index"}});
    return $key;
} # sub NEXTKEY

###################################
####### End tie() interface #######
###################################

#
# special method - returns 1/undef, sets expint
# for an individual row, or object-wide default if ! $key 
sub set_timeout {
    my($this,$to,$key) = (shift(),shift(),shift());
    eval {
	if ( $key ) {
	    $this->{"database"}[$this->{"index"}{$key}][EXPINT] = $to;
	    my($ekey) = $key;
	    $ekey =~ s/([\\\'])/\\$1/g;
	    $this->sync_bcast("TO '$ekey' $to\n")
		if $this->{sync_events}{SETTO};
	} else {
	    $this->{def_expint} = $to;
	    $this->sync_bcast("TO 'ALL' $to\n")
		if $this->{sync_events}{SETTO};
	}
    };
    if ( $@ eq '' ) {
	return 1;
    } else {
	$this->dh_warn(" timeout failed: $@");
	return undef;
    }
} # sub set_timeout

#
# special method - returns timeout/undef
# for an individual row, or default if ! $key
sub get_timeout {
    my($this,$key) = (shift(),shift());
    if ( $key ) {    
	return $this->{database}[$this->{"index"}{$key}][EXPINT];
    } else {
	return $this->{def_expint};
    }
} # sub get_timeout

#
# special method - returns atime/undef
# for an individual row
sub get_atime {
    my($this,$key) = (shift(),shift());
    if ( $key ) {    
	return $this->{database}[$this->{"index"}{$key}][ATIME];
    } else {
	return undef;
    }
} # sub get_timeout

#
# special method - return 1/undef, writes database
# to the checkpoint file. options are:
#   [no_]gc_at_ckpt : override object-wide setting
#   [no_]ic_at_ckpt : override object-wide setting
sub checkpoint {
    my($this) = shift();
    unless ( $this->{ckptfile} ) {
	return 1;
    }
    my(%options) = ( gc_at_ckpt => $this->{gc_at_ckpt},
		     ic_at_ckpt => $this->{ic_at_ckpt} );
    foreach my $opt ( @_ ) {
	if ( substr($opt,4) eq "gc_at_ckpt" ) {
	    $options{gc_at_ckpt} = substr($opt,0,2) eq "no" ? 0 : 1;
	} elsif ( substr($opt,4) eq "ic_at_ckpt" ) {
	    $options{ic_at_ckpt} = substr($opt,0,2) eq "no" ? 0 : 1;
	}
    }
    eval {
	my($file) = new IO::File;
	unless ( $file->open("> ".$this->{ckptfile}."tmp") ) {
	    die("Can't open data file: $!\n");
	}

	if ( $options{gc_at_ckpt} ) {
	    $this->collect_garbage();
	}
	if ( $options{ic_at_ckpt} ) {
	    die("Integrity error during checkpoint.\n")
		unless defined($this->integrity_check());
	}

	# FIXME: file slurping in reverse
	die("Write failed during checkpoint: $!\n")
	    unless $file->print(Data::Dumper->Dump([$this->{database}]),"\n");

	$file->close;
	unlink($this->{ckptfile})
	    if -e $this->{ckptfile};
	rename($this->{ckptfile}."tmp", $this->{ckptfile});
	$this->sync_bcast("CKPT\n")
	    if $this->{sync_events}{CKPT};
    };
    if ( $@ eq '' ) {
	return 1;
    } else {
	$this->dh_warn(" Checkpoint failed: $@");
	return undef;
    }
} # sub checkpoint

#
# special method - returns the number of items were garbage-collected.
sub collect_garbage {
    my($this) = shift();
    my($count) = $#{$this->{"database"}};
    eval {
	map {
	    # skip undef or non-timed out rows
	    if ( $_ && $_->[EXPINT] && time > $_->[ATIME] + $_->[EXPINT] ) {
		delete($this->{"index"}{$_->[KEY]});
		$_ = undef;
	    }
	} @{$this->{"database"}};
	$this->sync_bcast("GC\n")
	    if $this->{sync_events}{GC};
    };

    $this->dh_warn( '::collect_garbage '.
	  ($@ eq '' ? "succeeded\n" : 'failed') );
    return $count - $#{$this->{database}};
} # sub collect_garbage

#
# special method - returns undef on error, or
# the number of integrity anomalies corrected
sub integrity_check {
    my($this) = shift();
    my($count,$dbcnt) = (0,0);
    my($key,$ind,$row);
    eval {

	# find missing/corrupt index entries
      ROW: foreach $row ( @{$this->{database}} ) {
	  unless ( $row ) {
	      $dbcnt++;
	      next ROW;
	  }	
	  # add to index if missing/wrong
	  $key = $row->[KEY];
	  unless ( exists($this->{"index"}{$key}) ) {
	      $this->{"index"}{$key} = $dbcnt;
	      $count++;
	      $this->dh_warn(" Fixed missing index");
	  }
	  unless ( $this->{"index"}{$key} == $dbcnt ) {
	      $this->{"index"}{$key} = $dbcnt;
	      $count++;
	      $this->dh_warn(" Fixed corrupted index");
	  }      
	  # AHH! something is awry...
	  unless ( $this->{database}[$this->{'index'}{$key}] == $row ) {
	      die('Row in db not equivalent to row pointed to by idx!');
	  }
	  $dbcnt++;
      } # ROW

	# Now go through the index, checking each row for corruption
      KEY: foreach $key ( keys %{$this->{'index'}} ) {
	  $ind = $this->{'index'}{$key};
	  unless ( defined($this->{database}[$ind]) ) {
	      $this->dh_warn(' Nuking idxkey pointing to null entry');
	      delete($this->{'index'}{$key});
	      $count++;
	      next KEY;
	  }
	  unless ( defined($this->{database}[$ind][EXPINT]) ) {
	      $this->dh_warn(' Setting undef expint to def_expint');
	      $this->{database}[$ind][EXPINT] = $this->{def_expint};
	      $count++;
	  }
	  unless ( defined($this->{database}[$ind][KEY]) ) {
	      $this->dh_warn('KEY not stored in row?? Restoring from idx');
	      $this->{database}[$ind][KEY] = $key;
	      $count++;
	  }
	  unless ( $this->{database}[$ind][KEY] eq $key ) {
	      $this->dh_warn('Stale KEY stored in row?? Restoring from idx');
	      $this->{database}[$ind][KEY] = $key;
	      $count++;
	  }
	  unless ( defined($this->{database}[$ind][ATIME]) ) {
	      $this->dh_warn('Undefined ATIME!! Nuking row.');
	      $this->{database}[$ind] = undef;
	      delete($this->{"index"}{$key});
	      $count++;
	  }
	  unless ( defined($this->{database}[$ind][TTIME]) ) {
	      $this->dh_warn('Undefined TTIME!! Resetting to ATIME.');
	      $this->{database}[$ind][TTIME] = $this->{database}[$ind][ATIME];
	      $count++;
	  }
      } # KEY

	$this->sync_bcast("IC\n")
	    if $this->{sync_events}{INTCK};
    };

    if ( $@ eq '' ) {
	return $count;
    } else {
	$this->dh_warn(' integrity check failed');
	return undef;
    }
} # sub integrity_check

#
# special method - like each(), but value is DataParsed
sub eachString {
    my($this) = shift();
    my($key,$ind) = each(%{$this->{'index'}});
    if ( $key ) {
	return ($key, Data::Dumper->Dump([$this->{database}[$ind][VALUE]]));
    } else {
	return (undef,undef);
    }
} # sub eachString

#
# special method - returns the entire database, but parsed
sub toString {
    my($this) = shift();
    my($key,$ind,$string);
    keys(%{$this->{'index'}});
    $string = "(\n";
    # FIXME: can we just call Dump once?
    while ( ($key,$ind) = each(%{$this->{'index'}}) ) {
	$string .= "$key => ".
	  Data::Dumper->Dump([$this->{database}[$ind][VALUE]]).",\n";
    }
    chop($string);chop($string);
    $string .= "\n);\n";
    return $string;  
} # sub toString

#
# internal method - returns stringified database dump
sub dodump {
    my($this) = shift();
    return Data::Dumper->Dump([$this->{database}]);
} # sub dump

#
# internal method - returns 1/undef, initializes db&index from dump
# string 
sub undump {
    my($this,$string) = (shift(),shift());

    $this->{database} = [];
    $this->{'index'} = {};
    $this->sync_bcast("CLR\n")
	if $this->{sync_events}{'CLEAR'};

    unless ( $string ) {
	return 1;
    }    

    eval {
	$this->{database} = eval $string;
	for (my $count=0; $count <= $#{$this->{database}}; $count++ ) {
	    next
		unless $this->{database}[$count];
	    my($ekey) = $this->{database}[$count][KEY];
	    $ekey =~ s/([\\\'])/\\$1/g;
	    $this->{'index'}{$this->{database}[$count][KEY]} = $count;
	    $this->sync_bcast
		( "ADD '$ekey' " .
		Data::Dumper->Dump([$this->{database}[$count][VALUE]])."\n" )
		    if $this->{sync_events}{ADD};
	}
    };
    if ( $@ eq '' ) {
	return 1;
    } else {
	$this->dh_warn(" undump error: $@");
	return undef;
    }
} # sub undump

#
# special method - returns 1/undef, initializes the hash from a file 
# containing a dump
sub init_from {
    my($this,$fh) = (shift(),shift());
    my(@undump) = $fh->getlines();           # FIXME: slurping entire
    return $this->undump(join(' ',@undump)); # files is bad, m'kay?
} # sub init_from

###############################
## methods to implement sync ##
###############################

#
# special method - returns undef/# of dgrams broadcasted, broadcasts a
# sync of the given key
sub sync {
    my($this,$key) = (shift(),shift());
    return undef
	unless ( $key && exists $this->{'index'}{$key} );
    my($ekey) = $key;
    $ekey =~ s/([\\\'])/\\$1/g;
    return $this->sync_bcast
	("MOD '$ekey' " .
       Data::Dumper->Dump([$this->{database}[$this->{'index'}{$key}][VALUE]]) .
	 "\n");
} # sub sync

#
# internal method - processes any datagrams waiting for us from the
# servers in the sync pool, returns what the alarm reset returns
sub sync_refresh {
    my($this) = shift();
    my($c,$count,$line,$raddr,$iaddr,$port,$host,$rport,$rhost,$peer);
    $this->dh_warn(' Checking for pending sync commands');

    # read manually (as opposed to getline or some such) so we can check
    # that the source is in our pool
    $count = 1;
    $c = $line= '';
  READ: for ( $this->{sync_listener}->recv($c,$count,MSG_PEEK);
	      length($c) == $count;
	      $this->{sync_listener}->recv($c,$count,MSG_PEEK) ) {
      unless ( $c =~ /\n$/ ) {
	  $count++;
	  next READ;
      }

      $raddr = $this->{sync_listener}->recv($line,$count,0);
      unless ( defined($raddr) ) {
	  $this->dh_warn("Error recv'ing: ".$!);
	  $count++;
	  next READ;
      }

      ($rport,$iaddr) = sockaddr_in($raddr);
      $rhost = inet_ntoa($iaddr);

      $c = 0;
      foreach $peer ( @{$this->{"sync"}} ) {
	  next
	      if $peer->{"local"};
	  if ( !defined($peer->{sockaddr}) ) {
	      $this->dh_warn(" $peer->{host}:$peer->{port} has null socket");
	      next;
	  }
	  ($port,$iaddr) = sockaddr_in($peer->{sockaddr});
	  $host = inet_ntoa($iaddr);
	  unless ( $rhost eq $host && $rport == $port ) {
	      next;
	  }
	  $c = 1;
	  last;
      }

      if ( $c ) {
	  $this->dh_warn(" Got sync command from $rhost:$rport - $line");
      } else {
	  $this->dh_warn( ' Got sync command from host NOT IN POOL - '.
			  "$rhost:$rport - $line" );
	  $count++;
	  next READ;
      }

      $this->sync_process_cmd($line,$raddr);
  } # READ

    $this->dh_warn(" Checkint is .$this->{sync_checkint}.");
    if ( $this->{sync_checkint} ) {
	return alarm($this->{sync_checkint});
    } else {
	return 0;
    }
} # sub sync_refresh

#
# internal method - initializes the sync options, socket, peer list;
# inits the database from a peer or from ckptfile
sub sync_init {
    my($this) = shift();

    # set up sync options
    $this->dh_warn(' setting up sync');
    $this->{sync_port} = 8789
	unless $this->{sync_port};
    $this->{sync_to} = 1
	unless $this->{sync_to};    
    $this->{sync_checkint} = 3
	unless $this->{sync_checkint};    
    $this->{sync_atime} = 60
	unless $this->{sync_atime};
    $this->{sync_state_retrieval_hook} = sub { 0; }
    unless ref($this->{sync_state_retrieval_hook}) eq 'CODE';
    my($event);
    foreach $event ( qw( CLEAR 'DESTROY' CKPT INTCK GC ) ) {
	$this->{sync_events}{$event} = 0
	    unless $this->{sync_events}{$event};
    }
    foreach $event ( qw( ACCESS DELETE MODIFY ADD SETTO ) ) {
	$this->{sync_events}{$event} = 1
	    unless $this->{sync_events}{$event};
    }
    
    # set up sync pool
    $this->{sync_listener} = undef;
    my ($server,$sock,$line,$rin,$rout);
    foreach $server ( @{$this->{'sync'}} ) {
	# find a local spec to bind to -- we'll use this same socket to
	# do all our writing, so peers can tell that we're in their pool
	# (not that it's non-trivial to spoof udp...)
	die("No 'host' or 'addr' in sync server spec!")
	    unless exists($server->{host}) || exists($server->{addr});
	$server->{host} = $server->{addr}
	unless exists($server->{host});
	$server->{port} = $this->{sync_port}
	unless exists($server->{port});
	if ( exists($server->{addr}) ) {
	    if ( $server->{addr} =~ /([^:]):([0-9]+)/ ) {
		$server->{addr} = $1;
		$server->{port} = $2;
	    }
	} else {
	    $server->{addr} = $server->{host};
	}
	$server->{sync_to} = $this->{sync_to}
	unless exists($server->{sync_to});
	$! = 0;
	if ( eval 'local($SIG{__WARN__}) = sub {;;};
               $sock = new IO::Socket::INET
                       ( Domain    => "INET",
			 Proto     => "udp",
			 LocalAddr => $server->{addr},
                         LocalPort => $server->{port},
			 Timeout   => $server->{sync_to} )'
	     ) {
	    $this->{sync_listener} = $sock;
	    $server->{sockaddr} = 
		sockaddr_in($server->{port},inet_aton($server->{addr}));
	    $server->{'local'} = 1;
	    fcntl($sock,F_SETFL,
		  fcntl($sock,F_GETFL,0) | O_NONBLOCK | O_SYNC);
	    $sock->bind($sock->sockname());
	    last;
	} # try binding
    } # find suitable local spec
    
    if ( ! $this->{sync_listener} ) {
	$this->dh_warn('Couldn\'t find a local sync spec');
	$this->{sync_checkint} = 0;
	$this->{sync_listener} = 
	    new IO::Socket::INET( Domain => 'INET',
				  Proto  => 'udp',
				  LocalPort => $this->{sync_port},
				  Timeout => $this->{sync_to}
				  );
    } # set up null listener
    
    # set up list of remote peers
    foreach $server ( @{$this->{'sync'}} ) {
	next if $server->{'local'}; # this was set up upstairs
	die('No "host" or "addr" in sync server spec')
	    unless exists($server->{host}) || exists($server->{addr});
	$server->{host} = $server->{addr}
	unless exists($server->{host});
	$server->{port} = $this->{sync_port}
	unless exists($server->{port});
	if ( exists($server->{addr}) ) {
	    if ( $server->{addr} =~ /([^:]):([0-9]+)/ ) {
		$server->{addr} = $1;
		$server->{port} = $2;
	    }
	} else {
	    $server->{addr} = $server->{host};
	}
	$server->{sync_to} = $this->{sync_to}
	unless exists($server->{sync_to});
	$server->{sockaddr} = 
	    sockaddr_in($server->{port},inet_aton($server->{addr}));
	$server->{"local"} = 0;      
    } # set up remote peers
    
    foreach my $server ( @{$this->{'sync'}} ) {
	if ( $server->{'local'} ) {
	    $this->dh_warn(" LOCAL: $server->{addr}:$server->{port}");
	} else {
	    $this->dh_warn(" REMOTE: $server->{addr}:$server->{port}");
	}
    }
    
    if ( $this->{sync_checkint} ) {
	$this->sync_full();
	$main::SIG{ALRM} = sub {$this->sync_refresh()};
	$this->dh_warn( ' Sync initializations complete (' .
			alarm($this->{sync_checkint}).$this->{sync_checkint} .
			')' );
    } else {
	$this->dh_warn(' syncing *only* outbound');
    }  
} # sub sync_init  

#
# internal method - initializes the contents of the database from
# state information, trying first to request a full sync from one of
# the servers, or from the state/checkpoint file if no servers want to
# feed us. Returns undef on error, -1 on file init, -2 on successful
# sync_state_retrieval_hook, or number of keys added to the db from
# peer.
sub sync_full {
    my($this) = shift();
    my( $count,$icount,$client,$c,$line,$now,$fh,
        $raddr,$iaddr,$host,$port,$rhost,$rport,$peer );

    return -2
	if &{$this->{sync_state_retrieval_hook}};

    $count = $client = $c = $fh = 0;
  REQSYNC: foreach $client ( @{$this->{'sync'}} ) {
      next REQSYNC
	  if $client->{'local'};
      $count++
	  if $this->{sync_listener}->send
	      ( "FULL SYNC REQ\n", 0, $client->{sockaddr} );
      ($port,$iaddr) = sockaddr_in($client->{sockaddr});
      $host = inet_ntoa($iaddr);
      #$this->dh_warn(" Sent FSreq #$count to $client->{host} ($host:$port)");
      # we don't want multiple FULL SYNCs, so let's give peers time...
      # selects don't seem to work for udp under linux -- timeouts by hand
      $now = time;
      while ( time < $now + $this->{sync_checkint} + $this->{sync_to} ) {
	  # we've been heard...
	  $c = 1
	      if $this->{sync_listener}->recv($line,1,MSG_PEEK);
      }
      next REQSYNC
	  unless $c;

      $now = time;
    READ: while ( time < $now + $this->{sync_to} ) {
	$c = $line = '';
	$icount = 1;
	while (1) {
	    $this->{sync_listener}->recv($c,$icount,MSG_PEEK);
	    next READ
		if length($c) != $icount;
	    last
		if $c =~ /\n$/;
	} continue {
	    $icount++;
	}
	
	$raddr = $this->{sync_listener}->recv($line,$icount,0);
	unless ( defined($raddr) ) {
	    $this->dh_warn(" Error recv'ing: $!");
	    next READ;
	}

	($rport,$iaddr) = sockaddr_in($raddr);
	$rhost = inet_ntoa($iaddr);
	
	$c = 0;
	foreach $peer ( @{$this->{'sync'}} ) {
	    next
		if $peer->{'local'};
	    if ( !defined($peer->{sockaddr}) ) {
		$this->dh_warn(" $peer->{host}:$peer->{port} has null socket");
		next;
	    }
	    ($port,$iaddr) = sockaddr_in($peer->{sockaddr});
	    $host = inet_ntoa($iaddr);
	    next
		unless ( $rhost eq $host && $rport == $port );
	    $c = 1;
	    last;
	}
	
	unless ( $c ) {
	    $this->dh_warn
		(" Got sync cmd from host NOT IN POOL ($rhost:$rport) $line");
	    next READ;
	}
	
	$now = time; $fh++;
	last REQSYNC
	    if $line =~ /^END FULL SYNC$/;
	$this->sync_process_cmd($line,$raddr);
    } # READ
  } # REQSYNC

    unless ( $fh ) {
	# init from ckpt file
	return undef unless
	    (exists($this->{ckptfile})) && ($fh = new IO::File($this->{ckptfile}));
	$this->dh_warn(' No FS responses -- falling back to checkpoint file');
	$this->init_from($fh);
	close($fh);
	return -1;
    }

    return $count;
} # sub sync_full

#
# internal method - returns 1/undef, processes a sync command
sub sync_process_cmd {
    my($this,$cmd) = (shift(),shift());
    local($_) = $cmd;
    my($key);
    chomp;
    if ( /TOUCH\s+(\')?((?:\S.*)?\S)\'?/i ) {
	eval {
	    $key = $2;
	    if ( defined($1) ) {
		$key =~ s/\\(.)/$1/g;
	    }
	    die(' TOUCHing new item')
		unless exists($this->{'index'}{$key});
	    $this->{'database'}[$this->{'index'}{$key}][ATIME] = 
		$this->{'database'}[$this->{'index'}{$key}][TTIME] = time();
	    $this->dh_warn
		( " TOUCHEDed key --$key-- " .
		  $this->{database}[$this->{'index'}{$key}][VALUE] );
	};
	if ( $@ ne '' ) {
	    return undef;
	}
    } elsif ( /MOD\s+(?:'(.*?)'||(\S+))\s+(.+)$/i ) {
	eval {
	    if ( defined($1) ) {
		$key = $1;
		$key =~ s/\\(.)/$1/g;
	    } else {
		$key = $2;
	    }
	    if ( exists($this->{'index'}{$key}) ) {
		my($VAR1);
		eval $3;
		if ( $@ ) {
		    die(" Error in eval -$3-: $@");
		}
		$this->{database}[$this->{'index'}{$key}][VALUE] = $VAR1;
		$this->{database}[$this->{'index'}{$key}][EXPINT] =
		    $this->{def_expint};
		$this->{database}[$this->{'index'}{$key}][ATIME] = 
		    $this->{database}[$this->{'index'}{$key}][TTIME] = time();
		$this->dh_warn
		    ( " MODed key -$key- " .
		      $this->{database}[$this->{'index'}{$key}][VALUE] );
	    } else {
		die(' MODing new item');
	    }
	};
	if ( $@ ne '' ) {
	    $this->dh_warn($@);
	    return undef;
	}
    } elsif ( /ADD\s+(?:'(.*?)'||(\S+))\s+(.+)$/i ) {
	eval {
	    if ( defined($1) ) {
		$key = $1;
		$key =~ s/\\(.)/$1/g;
	    } else {
		$key = $2;
	    }
	    if ( exists($this->{'index'}{$key}) ) {
		die(' ADDing existing item');
	    } else {
		my($VAR1);
		eval $3;
		if ( $@ ) {
		    die(" Error in eval -$3-: $@");
		}
		$this->dh_warn(" Evaled -$3- to be -".join(',',@$VAR1).'-');
		push( @{$this->{'database'}},
		      [time(), time(), $this->{def_expint}, $key, $VAR1] );
		$this->{'index'}{$key} = $#{$this->{'database'}};
		$this->dh_warn
		    ( " Added key -$key- " .
		      $this->{database}[$this->{'index'}{$key}][VALUE] );
	    }
	};
	if ( $@ ne '' ) {
	    return undef;
	}
    } elsif ( /DEL\s+(\')?((?:\S.*)?\S)\'?/ ) { 
	eval {
	    $key = $2;
	    if ( defined($1) ) {
		$key =~ s/\\(.)/$1/g;
	    }
	    die ' DELing absent item'
		unless exists($this->{'index'}{$key});
	    $this->dh_warn
		( "DELing key -$key- " .
		  $this->{database}[$this->{'index'}{$key}][VALUE] );
	    $this->{'database'}[$this->{'index'}{$key}] = undef;
	    delete($this->{'index'}{$key});
	};
	if ( $@ ne '' ) {
	    return undef;
	}
    } elsif ( /TO\s+(?:'(.*?)'||(\S+))\s+([0-9]+||DEFAULT)/ ) {
	eval {
	    if ( defined($1) ) {
		$key = $1;
		$key =~ s/\\(.)/$1/g;
	    } else {
		$key = $2;
	    }
	    if ( $key eq 'ALL' ) {
		if ( $3 eq 'DEFAULT' ) {
		    foreach $key ( keys %{$this->{'index'}} ) {
			$this->{database}[$this->{'index'}{$key}][EXPINT] =
			    $this->{def_expint};
		    }
		} else {
		    foreach $key ( keys %{$this->{'index'}} ) {
			$this->{database}[$this->{'index'}{$key}][EXPINT] = $3;
		    }
		}
	    } elsif ( $3 eq 'DEFAULT' ) {
		$this->{'database'}[$this->{'index'}{$key}][EXPINT] =
		    $this->{def_expint};
	    } else {
		$this->{'database'}[$this->{'index'}{$key}][EXPINT] = $3;
	    }
	};
	if ( $@ ne '' ) {
	    return undef;
	}
    } elsif ( /CLR/ ) {
	eval {
	    @{$this->{database}} = ();
	    %{$this->{'index'}} = ();
	};
	if ( $@ ne '' ) {
	    return undef;
	}
    } elsif ( /DESTROY/ ) {
	eval {
	    delete(@$this{'database','index','class','def_expint'});
	    undef(%$this);
	};
	if ( $@ ne '' ) {
	    return undef;
	}
    } elsif ( /NOOP/ ) {
	$this->dh_warn(' No-op');
    } elsif ( /PING/ ) {
	$this->{sync_listener}->send("PONG\n",0,$_[0])
	    if $_[0];
	$this->dh_warn( 'PING');
    } elsif ( /FULL SYNC REQ/ ) {
	foreach my $key ( keys %{$this->{"index"}} ) {
	    $key =~ s/([\\\'])/\\$1/g;
	    $this->{sync_listener}->send
		( "ADD '$key' " .
	 	  Data::Dumper->Dump
		      ( [$this->{database}[$this->{'index'}{$key}][VALUE]] )
		  ."\n",
		  0, $_[0] );
	}
	$this->{sync_listener}->send("END FULL SYNC\n", 0, $_[0]);
    } elsif ( /END FULL SYNC/ ) {
	$this->dh_warn(' Someone took too long to fullfil my FULL SYNC REQ');
    } elsif ( /CKPT|GC|IC/ ) {
	$this->dh_warn(" Use of unimplemented sync command: $_");
    } else {
	$this->dh_warn(" Unknown sync command: $_");
	return undef;
    } 
    return 1;
} # sub sync_process_cmd

#
# internal method - returns the number of datagrams sent.
sub sync_bcast {
    my($this,$msg) = (shift(),shift());
    my($count,$chars) = (0,0);
    return 0
	unless $this->{sync_listener};
    $this->dh_warn(" Bcasting: $msg");
    foreach my $peer ( @{$this->{'sync'}} ) {
	next if $peer->{'local'};
	# FIXME: why does IO::Socket::send return (apparently) ""?
	if ( $chars = send( $this->{sync_listener},
			    $msg,0,$peer->{sockaddr} ) &&
	     $chars == length($msg) ) {
	    $count++;
	    $this->dh_warn
		( "\tbcasted $chars bytes to $peer->{addr}:$peer->{port}" );
	} elsif ( defined($chars) ) {
	    $count++;
	    $this->dh_warn("\tSent $chars/".length($msg)." bytes of $msg");
	} else {
	    $this->dh_warn
		( " ERROR sending to $peer->{addr}:$peer->{port} $@:$!" );
	}
    }
    $this->dh_warn("\tBcasted to $count peers");
    return $count;
} # sub sync_bcast

1;
