#!/usr/bin/perl -w
# $Id: LMSd.pl,v 1.9 2002/04/16 21:56:09 santana Exp $
#
# the Login Method Server daemon
#

package LMSd;
use strict;
use lib "../lib";
use Net::muxDaemon;
# the DB driver, and retrieve types
use DBI qw(:sql_types);

#
# Get configured
use Config::IniFiles;
my($INI) = (shift() || "../macs.cfg");
my($cfg,%cfg);
die("Config file $INI unreadable.\n")
    unless -r $INI;
$cfg = tie( %cfg, 'Config::IniFiles',
	    -file => $INI,
	    -default => "default" );
die("Couldn't configure: $!\n")
    unless $cfg->Sections();
for ( keys %{$cfg{env}} ) {
    $ENV{$_} = $cfg{env}{$_};
}

# make sure we're running in the background
FORK: {
    # the process id variable
    my($pid);
    # spawn a child
    if($pid = fork()) {
	# parent process here, bye bye
	exit(0);
    } elsif(defined $pid) {
	# child process here, we continue
	last FORK;
    } elsif($! =~ m/[Nn]o more process/) {
	# can't fork, sleep and try again
	sleep(2);
	redo FORK;
    } else {
	# unrecoverable fork error
	die("Couldn't fork: $!\n");
    }
}

# has the Net::muxDemon opened syslog yet?
my($syslog_open) = 0;

# database specific variables
my($dbh, $mapname_sth, $mapuser_sth, $findloginmethodid_sth,
   $adduser_sth, $authuser_sth, $muidfromemail_sth, $finduserid_sth,
   $getusernames_sth);

# connect to the database
&db_setup();
# aus specific variables
my($aus,$timeout);
# connect to the aus
&aus_setup();

#
# The hosts in our distributed setup
my($hosts) = [];
for ( $cfg->val("LMS","p") ) {
    push(@$hosts, {addr => $_});
}

# Where the LMC connections are stored
my(%connections) = ();

# starting Net::Demon, syslog will be open
$syslog_open = 1;
# vroom...
new Net::muxDaemon( POOL    => $hosts,
		    LOGTAG  => $cfg{LMS}{logtag},
		    PIDFILE => $cfg{LMS}{pidf},
		    TIMEOUT => $cfg{LMS}{timeout}
		    );
die("Net::muxDaemons don't return!\n\t$!\n\t$@\n");

# process incoming commands to the Net::muxDaemon
sub process_command {
    # the the Net::muxDaemon handle and the command
    my($this,$buf,$client) = (shift(),shift(),shift());
    syslog("debug","Entering process_command: $buf");
    # if the command is empty
    if ( $buf =~ m/^\s*$/ ) {
	# send back an error string
	return "OK NOOP\n";
    }  
    # used all over
    my($reply, $operation);
    # split out the operation from the command
    ($operation,$buf) = split(' ', $buf, 2);
    $operation = uc($operation);
    # eat surrounding whitespace
    $buf =~ s/^\s+|\s+$//g;

  OP: {

      # REG method
      # register a new LMC
      if ($operation eq "REG") {
	  # if there are no arguments
	  unless($buf) {
	      # send back an error
	      $reply = "ERR REG: not enough arguments\n";
	      last OP;
	  }
	  if ( $buf eq "internal" || exists($connections{$buf})) {	
	      # you can't override a current LMC
	      $reply = "NO\n";
	      last OP;
	  }
	  # registered in the database?
	  $findloginmethodid_sth->bind_param( 1, $buf,
					      { TYPE => SQL_VARCHAR } );
	  $findloginmethodid_sth->execute();
	  if ( $#{$findloginmethodid_sth->fetchall_arrayref()} == -1 ) {
	      # no dynamic login methods -- set them up by hand
	      $reply = "NO $buf unknown method\n";
	      last OP;
	  }
	  # store the LMC socket connection
	  $connections{$buf} = $client;
	  # send acknowledgement to the LMC
	  $reply = "OK\n";
      }

      # DEREG method
      # unregister an old LMC
      elsif ($operation eq "DEREG") {
	  # if there ar no arguments
	  unless($buf) {
	      # send back an error
	      $reply = "ERR DEREG: not enough arguments\n";
	      last OP;
	  }
	  unless (exists($connections{$buf})) {
	      $reply = "NO Method $buf not registered\n";
	      last OP;
	  }
	  # only the socket that registered a meth can unregister it
	  unless ( $client == $connections{$buf} ) {
	      $reply = "NO\n";
	      last OP;
	  }
	  # delete the key in %connections, and
	  # close the socket unless it's the socket for other methods
	  my($om) = 0;
	  for ( keys %connections ) {
	      next if $_ eq $buf;
	      if ( $connections{$_} == $connections{$buf} ) {
		  $om++;
		  last;
	      }
	  }
	  delete $connections{$buf};
	  $reply = $om
	      ? "OK\n"
		  : { CLOSE => '', reply => 'OK' };
	  last OP;
      }

      # AUTH method username password ip port
      elsif($operation eq "AUTH") {
	  my($method, $username, $password, $ip, $port) = split(/ /, $buf, 5);
	  unless(defined($method) and defined ($username) and 
		 defined($password) and defined($ip) and defined($port)) {
	      $reply = "ERR AUTH: not enough arguments\n";
	      last OP;
	  }
	  if ( exists($connections{$method})) {
	      $connections{$method}->print("AUTH $method $username $password ip port\n");
	      my($approval);
	      $this->getreq($connections{$method}, \$approval);
	      unless($approval =~ m/^YES/) {
		  $reply = ($approval =~ m/^(NO|ERR)/)
		      ? $approval
		      : "ERR LMC: $approval\n";
		  last OP;
	      }
	  } elsif ( $method eq 'internal' ) {
	      unless ( do_internal_auth($username,$password,$ip,$port) ) {
		  $reply = "NO\n";
		  last OP;
	      }
	  } else {
	      $reply = "ERR AUTH: method not available\n";
	      last OP;
	  }
	  unless($ip =~ m/^([0-9]+)\.([0-9]+)\.([0-9]+)\.([0-9]+)$/) {
	      $reply = "ERR AUTH: bad IP address\n";
	      last OP;
	  }
	  srand();
	  my($skey) = sprintf( "%03d%05d%03d%03d%05d%03d", 
			       $3, $port, $1, $4, rand(time()), $2 ) . time();
	  my($uid) = db_get_uid($this,$method, $username);
	  # check for uid
	  unless ( defined($uid) && length($uid) > 0 ) {
	      $reply = "ERR cannot unmap $username by $method.\n";
	      last OP;
	  }
	  # if the aus is crashed, we go on. :-/
	  &aus_setup() unless (defined($aus) && $aus->sockport());
	  if ( defined($aus) && $aus->sockport() ) {
	      syslog('debug',"setting up aus session for $uid");
	      $aus->print("PUT $skey $uid $timeout $ip\n");
	      my($ok) = $aus->getline();
	      unless(defined($ok)) {
		  $reply = 'ERR no reply from AUS';
		  last OP;
	      }
	      unless($ok =~ m/^OK/) {
		  $reply = $ok;
		  last OP;
	      }
	  }
	  $reply = "YES $skey\n";
      }

      # METH
      elsif ( $operation eq "METH" ) {
	  $reply = "OK " . join(' ',keys(%connections)) . "\n";
      }

      # PING
      elsif ( $operation eq "PING" ) {
	  $reply = "PONG\n";
      }

      # EVAL
      elsif ( $operation eq "EVAL" ) {
	  $reply = "OK ";
	  $reply .= eval $buf;
	  $reply .= "\n";
      }

      # QUIT
      elsif ( $operation eq "QUIT" ) {
	  unlink("AppCache.pid");
	  $reply = "OK\n";
	  $KILLED = 1;
      }

      # menospeakie
      else {
	  $reply = "ERR unknown command: $operation $buf\n"
	  }

  } # OP

    syslog("debug","Exiting process_command: $reply");
    return $reply;
} # sub process_command


# set up the AUS connection
sub aus_setup {
    # connect to the AUS
    $aus = new IO::Socket::INET(PeerAddr => $cfg{AUS}{s});
    # the AUS timeout
    $timeout = $cfg{LMS}{session_timeout};
} # sub aus_setup


# set up the database connection
sub db_setup {
    # make sure the database isn't already connected
    eval { $dbh->disconnect(); };
    # create the database handle
    $dbh = DBI->connect($cfg{db}{d}, $cfg{db}{l}, $cfg{db}{p});
    # make sure the database is connected
    unless($dbh) {
	# log connection error and exit
	if($syslog_open) {
	    syslog('err', "Couldn't connect: $DBI::errstr");
	    $KILLED = 1;
	} else {
	    die("Couldn't connect: $DBI::errstr");
	}
    }
    # the database should wine if there are any errors
    $dbh->{RaiseError} = 1;
    # prepare the query to retrieve login method information
    $@ = "";
    # trap errors
    eval {
	$mapname_sth = $dbh->prepare($cfg{"sql LMS"}{map_name});
	$findloginmethodid_sth = $dbh->prepare($cfg{"sql LMS"}{find_login_method_id});
	$finduserid_sth = $dbh->prepare($cfg{"sql LMS"}{find_user_id});
	$mapuser_sth = $dbh->prepare($cfg{"sql LMS"}{map_user});
	$adduser_sth = $dbh->prepare($cfg{"sql LMS"}{add_user});
	$authuser_sth = $dbh->prepare($cfg{"sql LMS"}{auth_user});
	$muidfromemail_sth = $dbh->prepare($cfg{"sql LMS"}{muid_from_email});
	$getusernames_sth = $dbh->prepare($cfg{"sql LMS"}{get_usernames});
    };
    # exit if there are errors
    if($@){
	# log connection error and exit
	if($syslog_open) {
	    syslog('err', "$@");
	    $KILLED = 1;
	} else {
	    die("$@");
	}
    }
} # sub db_setup



# retreive UID of this user from the database
# (otherwise known as unmapping a login)
sub db_get_uid {
    my($this,$method,$login) = (shift(),shift(),shift());
    my($username,$email);
    
    # is the standard mapping there?
    return $username
	if defined($username = unmap($method,$login));
    
    # can we find an existing user to map to?
    $connections{$method}->print("MAIL $method $login\n");
    $this->getreq($connections{$method}, \$email);
    $username = finduser($method,$login,$email)
	if $email =~ s/^YES\s([^\s]*)\s*(.*)\s*$/$1/;
    return $username
	if defined($username);
    
    # create a new user w/ mapping
    return $username
	if defined($username = adduser($method,$login,$email,$2));

    # we couldn't even add a user!
    return undef;
} # sub db_get_uid



sub unmap {
    my($method,$login) = (shift(),shift());
    my($t) = [];

    # if a database operation fails, reconnect and try again
    my($second_try) = 0;
  DB: {
      $@ = "";
      # trap errors
      eval {
	  # run the query
	  $mapname_sth->bind_param(1, $method, { TYPE => SQL_VARCHAR });
	  $mapname_sth->bind_param(2, $login, { TYPE => SQL_VARCHAR });
	  $mapname_sth->execute();
	  $t = $mapname_sth->fetchall_arrayref();
      };
      # if there are errors
      if ($@) {
	  # try again
	  unless ($second_try) {
	      syslog("err", "$@");
	      $second_try = 1;
	      &db_setup();
	      redo DB;
	  }
	  # bye bye
	  syslog("err", "Second try: $@");
	  $KILLED = 1;
	  return undef;
      }  
  } # DB block
    return $t->[0][0];
} # sub unmap



sub finduser {
    my($method,$login,$email) = (shift(),shift(),shift());
    my($p,$m,$u);

    my($second_try) = 0;
  DB: {
      $@ = "";
      eval {
	  # check that the mapping exists
	  $muidfromemail_sth->bind_param(1,$p->[0][0],{TYPE => SQL_VARCHAR});
	  $muidfromemail_sth->execute();
	  $u = $muidfromemail_sth->fetchall_arrayref();
	  # FIXME: we should resolve multi-matches like the URA does
	  return undef if $#$u != 1;

	  # map it!
	  $findloginmethodid_sth->bind_param(1,$method,{TYPE => SQL_VARCHAR});
	  $findloginmethodid_sth->execute();
	  $m = $findloginmethodid_sth->fetchall_arrayref();
	  if ($#$m == -1) {
	      # no such login method
	      return undef;
	  }
	  $mapuser_sth->bind_param(1,$login,{TYPE => SQL_VARCHAR});
	  $mapuser_sth->bind_param(2,$m->[0][0],{TYPE => SQL_VARCHAR});
	  $mapuser_sth->bind_param(3,$u->[0][0],{TYPE => SQL_VARCHAR});
	  $mapuser_sth->execute();
      };
      # if there are errors
      if ($@) {
	  # try again
	  unless ($second_try) {
	      syslog("err", "$@");
	      $second_try = 1;
	      &db_setup();
	      redo DB;
	  }
	  # bye bye
	  syslog("err", "Second try: $@");
	  $KILLED = 1;
      }  
  } # DB block

    return $p->[0][0];
} # sub finduser



sub adduser {
    my($method,$login,$mail,$name) = (shift(),shift(),shift(),shift());
    my($username) = $login;
    my($u);

    my($second_try) = 0;
  DB: {
      $@ = "";
      # trap errors
      eval {
	  # load similar usernames in core
	  $getusernames_sth->bind_param(1, $login."%");
	  $getusernames_sth->execute();
	  $u = $getusernames_sth->fetchall_arrayref();
      };
      # if there are errors
      if ($@) {
	  # try again
	  unless ($second_try) {
	      syslog("err", "$@");
	      $second_try = 1;
	      &db_setup();
	      redo DB;
	  }
	  # bye bye
	  syslog("err", "Second try: $@");
	  $KILLED = 1;
	  return undef;
      }  
  } # DB block

    # stay in this block until we make a unique username
  MUNGE: {
      for ( @$u ) {
	  if ( $$_[0] eq $username ) {
	      $username =~ s/(\d*)$/1 + $1/e;
	      redo MUNGE;
	  }
      }
  } # MUNGE
    
    undef $u; # free()
    $second_try = 0;
  DB: {
      $@ = "";
      # trap errors
      eval {
	  # do the doo
	  $adduser_sth->bind_param(1,$username);
	  $adduser_sth->bind_param(2,$name);
	  $adduser_sth->bind_param(3,$mail);
	  $adduser_sth->execute();

	  $findloginmethodid_sth->bind_param(1, $method);
	  $findloginmethodid_sth->execute();
	  my($mid) = $findloginmethodid_sth->fetchall_arrayref();
	  if ( $#$mid == -1 ) {
	      # no such login method???
	      warn("no such method?"); 
	      return undef;
	  }
	  
	  $finduserid_sth->bind_param(1, $username);
	  $finduserid_sth->execute();
	  my($uid) = $finduserid_sth->fetchall_arrayref();
	  if ( $#$uid == -1 ) {
	      # no such user???
	      warn("no such user?"); 
	      return undef;
	  }
	  if ( ! $uid->[0][0] ) { warn("null uid"); }
	  
	  $mapuser_sth->bind_param(1, $login);
	  $mapuser_sth->bind_param(2, $mid->[0][0]);
	  $mapuser_sth->bind_param(3, $uid->[0][0]);
	  $mapuser_sth->execute();
      };
      # if there are errors
      if ($@) {
	  # try again
	  unless ($second_try) {
	      syslog("err", "$@");
	      $second_try = 1;
	      &db_setup();
	      redo DB;
	  }
	  # bye bye
	  syslog("err", "Second try: $@");
	  $KILLED = 1;
	  return undef;
      }  
  } # DB block
    return $username;
} # sub adduser


# FIXME: we need to grab the passwd and crypt() using a real salt.
sub do_internal_auth {
    my($username,$password,$ip,$port) = (shift(),shift(),shift(),shift());
    $authuser_sth->bind_param(1, $username);
    $authuser_sth->bind_param(2, crypt($password,substr($password,0,2)));
    $authuser_sth->execute();
    return ($#{$authuser_sth->fetchall_arrayref()} == 0);
} # sub do_internal_auth

