/* $Id: mod_macs_upc.c,v 1.5 2002/04/16 04:20:18 santana Exp $
 *
 * 05.00,
 * .mario
 *
 * An implementation of the upc piece of MACS. Starting off from a perl
 * module with the following comments
 *
 * ---------
 * WACA.pm
 *
 * This module is the heart of WACA and performs all the dificult "in
 * the server" type stuff, any serious problem here could crash the
 * whole server, so be careful!
 *
 * 05-04-99 Blake Mitchell
 * ---------
 *
 * Installation and Use
 * --------------------
 *
 * Make sure the mod_macs_upc.so is in your apache's modules/ dir,
 * that the following line is in httpd.conf:
 *   LoadModule macs_upc_module modules/mod_macs_upc.so
 *
 * There are two ways to configure this upc. The best is to use
 * the MacsConfigFile directive in your base and/or <virtualhost>
 * sections, and the MacsProfile directive in your <dir>,
 * <location>, or <file> sections:
 *   MacsConfigFile /path/to/cf
 *   <location /appneedsloginandemail>
 *   MacsProfile MACS_LOGIN LOGIN 1
 *   MacsProfile MACS_EMAIL EMAIL 1
 *   </location>
 *   <location /appneedsloginandemail/andnamebutnoemail>
 *   MacsProfile MACS_FIRST FIRST 1
 *   MacsProfile MACS_LAST LAST 1
 *   MacsProfile MACS_EMAIL
 *   </location>
 * The MacsProfile directive maps an element of the user's profile
 * to a request header, and therefore a CGI environment variable
 * in all caps and prefixed by 'HTTP_'. It also spoofs a cookie, with
 * the same name as the request header (no HTTP_ prefix.) The syntax
 * for MacsProfile is:
 *   MacsProfile <request_header> [ <user_profile_element> <app_id> ]
 * where <user_profile_element> is the tag for this element in the
 * UPS, <app_id> is the application namespace, and <request_header> is
 * the spoofed request header that will contain the element's
 * value. If the <app_id> and <user_profile_element> are left out, the
 * directive unmaps the element's (probably inherited from parent
 * directories) mappings. The format of the MACS configuration file is
 * out of the scope of these comments, see the docs for MACS
 * configuration.
 */

/******************
 ** begin header **
 ******************/

/* identify yourself */
const char macs_upc_rcs_id[] = "$Id: mod_macs_upc.c,v 1.5 2002/04/16 04:20:18 santana Exp $";

/* apxs includes */
#include "httpd.h"
#include "http_log.h"
#include "http_config.h"
#include "http_protocol.h"
#include "ap_config.h"

#define MACS_UPC_BUFSIZ 1024
#define MACS_UPC_ELEMENT_PREFIX "MACS_UPC_ELEMENT_"
#define MACS_UPC_DEFAULT_PORT 0
#define MACS_UPC_DEFAULT_TIMEOUT 3

/* sockets to UPS servers */
struct macs_upc_sock_struct {
  char *host;
  int port;
  int isup;
  int sock;
  int in_use;
  struct macs_upc_sock_struct *next;
};
typedef struct macs_upc_sock_struct macs_upc_sock;

/* The list of profile elements to expose is kept in an apache table
 * in the configuration block for a request. The keys are elements and
 * the values are request headers to spoof.
 */

/* server configuration data */
struct macs_upc_cfg_struct {
  int port,        /* default port (not used)              */
    timeout;       /* timeout for various ops              */
  char *uidhdr;    /* auc-spoofed request header w/ uid    */
  macs_upc_sock *socks; /* linked list of UPC sockets      */
};
typedef struct macs_upc_cfg_struct macs_upc_cfg;

/* directory configuration data */
struct macs_upc_dir_cfg_struct {
  table *elts; /* per-directory hashes of elements to expose */
  table *aids; /* and their aids, keyed by header name.      */
};
typedef struct macs_upc_dir_cfg_struct macs_upc_dir_cfg;

/* upc initializations (logs a line) */
void macs_upc_init(server_rec *,pool *);

/* Initialize upc_cfg
 * There won't be any *server* merging routines because if we give
 *   different servers different config files, we're plugging those
 *   server into different, isolated instances of macs
 * Directory merges will inherit the parent's stuff, and treat the
 *   directives as modifications to that configuration.
 */
void *macs_upc_svrcfg_init(pool *,server_rec *);
void *macs_upc_dircfg_init(pool *,char *);
void *macs_upc_dircfg_merge(pool *,void *,void *);
int macs_upc_dirmerge_cb(void *,const char *,const char *);

/* New-style configuration directive handler
 *   populate upsspec lists
 *   set global variables
 */
const char *macs_upc_cfdirective(cmd_parms *,void *,char *);
char *macs_upc_parse_cf(char *,cmd_parms *,macs_upc_cfg *);

/* Configuration directive handler to create this directory's profile
 * exposing configuration
 */
const char *macs_upc_map_element( cmd_parms *,void *,
				  const char *,const char *,const char *);

/* opaque struct for macs_upc_get_element callback */
struct macs_upc_callback_data_struct {
  request_rec *r;
  macs_upc_cfg *c;
  macs_upc_dir_cfg *dc;
  char *buf;
  char *uid;
};
typedef struct macs_upc_callback_data_struct macs_upc_callback_data;

/* Fixup handler -- grabs elements from the UPS and spoofs the request
 * headers
 */
int macs_upc_expose(request_rec *);
int macs_upc_expose_cb(void *,const char *,const char *);
char *macs_upc_get_element(const char *,const char *,macs_upc_callback_data *);

/* add a UPS server
 * errors into arg1 w/ rv != 0
 */
int macs_upc_add_ups(char *,cmd_parms *,macs_upc_cfg *);

/* returns a null-initted macs_upc_sock */
macs_upc_sock *new_upc_sock(pool *);

/* connect to the UPS specs */
void macs_upc_child_init(server_rec *,pool *);

/* close all UPS connections so garbage collection goes smoothly */
void macs_upc_child_exit(server_rec *,pool *);

/* Config directive */
static const command_rec macs_upc_commands[] = {
  { "MacsConfigFile", macs_upc_cfdirective, NULL, RSRC_CONF, TAKE1,
    "MACS configuration file: MacsConfigFile /path/to/macs.cfg" },
  { "UPCConfigFile", macs_upc_cfdirective, NULL, RSRC_CONF, TAKE1,
    "UPC-specific MACS configuration file: UPCConfigFile /path/to/upc.cfg" },
  { "MacsProfile", macs_upc_map_element, NULL, OR_OPTIONS, TAKE13,
    "Map MACS user profile element to header: MacsProfile element header" },
  {NULL}
};

/* Dispatch list for API hooks */
module MODULE_VAR_EXPORT macs_upc_module = {
  STANDARD_MODULE_STUFF,
  macs_upc_init,         /* module initializer                  */
  macs_upc_dircfg_init,  /* create per-dir    config structures */
  macs_upc_dircfg_merge, /* merge  per-dir    config structures */
  macs_upc_svrcfg_init,  /* create per-server config structures */
  NULL,                  /* merge  per-server config structures */
  macs_upc_commands,     /* table of config file commands       */
  NULL,                  /* [#8] MIME-typed-dispatched handlers */
  NULL,                  /* [#1] URI to filename translation    */
  NULL,                  /* [#4] validate user id from request  */
  NULL,                  /* [#5] check if the user is ok _here_ */
  macs_upc_expose,       /* [#3] check access by host address   */
  NULL,                  /* [#6] determine MIME type            */
  NULL,                  /* [#7] pre-run fixups                 */
  NULL,                  /* [#9] log a transaction              */
  NULL,                  /* [#2] header parser                  */
  macs_upc_child_init,   /* child_init                          */
  macs_upc_child_exit,   /* child_exit                          */
  NULL                   /* [#0] post read-request              */
};

/****************
 ** end header **
 ****************/

void macs_upc_init(server_rec *s,pool *p) {
  ap_add_version_component("macs/upc");
} /* macs_upc_init */


/*****************************************************************
            config block initializers & mergers
 *****************************************************************/
void *macs_upc_svrcfg_init(pool *p, server_rec *s) {
  macs_upc_cfg *c = (macs_upc_cfg *)ap_palloc(p,sizeof(macs_upc_cfg));
  c->port = MACS_UPC_DEFAULT_PORT;
  c->timeout = MACS_UPC_DEFAULT_TIMEOUT;
  c->socks = NULL;
  return (void *)c;
} /* macs_upc_svrcfg_init */

void *macs_upc_dircfg_init(pool *p, char *dir) {
  macs_upc_dir_cfg *dc = 
    (macs_upc_dir_cfg *)ap_palloc(p,sizeof(macs_upc_dir_cfg));
  dc->elts = ap_make_table(p,0);
  dc->aids = ap_make_table(p,0);
  return (void *)dc;
} /* macs_upc_dircfg_init */

void *macs_upc_dircfg_merge(pool *p, void *base, void *new) {
  macs_upc_dir_cfg
    *parent = (macs_upc_dir_cfg *)base,
    *child  = (macs_upc_dir_cfg *)new,
    *merged = (macs_upc_dir_cfg *)ap_palloc(p,sizeof(macs_upc_dir_cfg));
  merged->elts = ap_copy_table(p,parent->elts);
  ap_table_do(macs_upc_dirmerge_cb, (void *)merged->elts,
	      child->elts, NULL);
  merged->aids = ap_copy_table(p,parent->aids);
  ap_table_do(macs_upc_dirmerge_cb, (void *)merged->aids,
	      child->aids, NULL);
  return (void *)merged;
} /* macs_upc_dircfg_merge */

/* Adds exposures in the parent to the child, or clears all exposures
 * of an element if it's been explicitly cleared in the child.
 */
int macs_upc_dirmerge_cb( void *tablev,
			  const char *header,
			  const char *elt_or_aid ) {
  ap_table_set((table *)tablev,header,elt_or_aid);
  return 1;
} /* macs_upc_dircfg_merge_cb */


/*****************************************************************
             new-style directive handlers & co
 *****************************************************************/

/* expose this element as this request-header, or clear its exposures */
const char *macs_upc_map_element( cmd_parms *parms, void *mconfig,
				  const char *header,
				  const char *elt,
				  const char *aid ) {
  macs_upc_dir_cfg *dc = (macs_upc_dir_cfg *)mconfig;
  /* ap_get_module_config(parms->r->per_dir_config, &macs_upc_module); 
   */
  if ( ap_table_get(dc->elts,header) ||
       ap_table_get(dc->aids,header) ) {
    ap_log_error( APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, parms->server,
		  "overwriting app %s element %s to %s for %s", 
		  aid, elt, header, parms->path );
  } else {
    ap_log_error( APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, parms->server,
		  "mapping app %s element %s to %s for %s", 
		  aid, elt, header, parms->path );
  }
  ap_table_set(dc->elts,header,elt);
  ap_table_set(dc->aids,header,aid);
  return NULL;
} /* macs_upc_map_element */

/* go through the macs-style config file */
const char *macs_upc_cfdirective( cmd_parms *parms, 
				  void *mconfig, 
				  char *cfname ) {
  FILE *cf;
  macs_upc_cfg *c = (macs_upc_cfg *)
    ap_get_module_config(parms->server->module_config,&macs_upc_module);
  char *rv,
    *line = (char *)ap_palloc(parms->temp_pool,sizeof(char)*MACS_UPC_BUFSIZ);

  /* we should inherit a pre-configured macs_upc_cfg by now */
  if ( !c )
    return "Internal error initializing configs";
  cf = ap_pfopen(parms->temp_pool,cfname,"r");
  if ( !cf ) {
    ap_snprintf( line,MACS_UPC_BUFSIZ,"Error opening %s: %s",
		 cfname,strerror(errno) );
    return line;
  }
  while ( fgets(line,MACS_UPC_BUFSIZ,cf) ) {
    if ( (rv=macs_upc_parse_cf(line,parms,c)) ) {
      /* errs in line */
      return rv;
      break;
    }
  } /* while fgets */
  if ( ferror(cf) && !feof(cf) ) {
    /* bad */
    ap_snprintf( line,MACS_UPC_BUFSIZ,"Error reading from %s: %s",
		 cfname,strerror(errno) );
    return line;
  }
  return (char *)NULL;
} /* macs_upc_cfdirective */

/* changes the config block, or feeds a line to add_ups */
char *macs_upc_parse_cf(char *line, cmd_parms *parms, macs_upc_cfg *c) {
  char *cdir, *val;

  cdir = ap_getword_white_nc(parms->temp_pool,&line);
  if (cdir[0] == '#')
    return (char *)NULL; /* skip comments */
  /* there *has* to be a better way to do this, but this isn't a
   *   speed-sensitive section...
   * for now we'll just check each directive against the ones we care
   *   about
   */
  if (!strcasecmp(cdir,"UPS")) {
    val = ap_getword_white_nc(parms->temp_pool,&line);
    if (!val) {
      line = ap_psprintf(parms->pool,"Null UPS spec in config: %s",line);
      return line;
    }
    if ( line &&
	 (cdir = ap_getword_white_nc(parms->temp_pool,&line)) &&
	 cdir[0] != '#' )
      val = ap_pstrcat(parms->temp_pool,val,":",cdir,NULL);
    line = macs_upc_add_ups(val,parms,c) ?
      ap_psprintf(parms->pool,"Error adding UPS %s",val) :
      (char *)NULL;
    return line;
  }
  if (!strcasecmp(cdir,"HTTP_UID_HEADER")) {
    c->uidhdr = ap_getword_conf_nc(parms->pool,&line);
    line = c->uidhdr ?
      (char *)NULL :
      ap_psprintf(parms->pool,"Error adding UPS %s",val);
    return line;
  }
  if (!strcasecmp(cdir,"TIMEOUT")) {
    val = ap_getword_conf_nc(parms->pool,&line);
    if (!val)
      return ap_psprintf(parms->pool,"Null timeout spec in config: %s",line);
    c->timeout = atoi(val);
    return (char *)NULL;
  }
  return (char *)NULL;
} /* macs_upc_parse_cf */


/*****************************************************************
                     config-block modifiers
 *****************************************************************/

/* adds a ups to the config block */
int macs_upc_add_ups(char *hostparm, cmd_parms *parms, macs_upc_cfg *c) {
  char *host, *charpt;
  long port;
  macs_upc_sock *curr;
  /* parse hostparm */
  charpt = host = ap_pstrdup(parms->temp_pool,hostparm);
  if ( !host ) {
    /* malloc error */
    hostparm--;
    ap_snprintf( hostparm,MACS_UPC_BUFSIZ,"Memory allocation error: %s",
		 strerror(errno) );
    return -1;
  }
  while ( charpt && *charpt != ':' )
    charpt++;
  if ( charpt ) {
    *charpt = (char)NULL;
    port = atol(++charpt);
  } else {
    ap_log_error( APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, parms->server,
		  "No port in config line: %s", hostparm );
  }
  if ( !c->socks ) {
    /* first item in list */
    curr = c->socks = new_upc_sock(parms->pool);
  } else {
    /* append an item to list */
    for ( curr=c->socks; curr->next; curr=curr->next );
    curr = curr->next = new_upc_sock(parms->pool);
  }
  if ( !curr ) {
    /* malloc error */
    hostparm--;
    ap_snprintf(hostparm,MACS_UPC_BUFSIZ,"Memory allocation error: %s",
		strerror(errno));
    return -1;
  }
  /* init the bare necessities; the client init actually connects */
  curr->host = ap_pstrdup(parms->pool,host);
  if ( !curr->host ) {
    /* malloc error */
    hostparm--;
    ap_snprintf(hostparm,MACS_UPC_BUFSIZ,"Memory allocation error: %s",
		strerror(errno));
    return -1;
  }
  curr->port = port;
  ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, parms->server,
	       "Configured UPS: %s:%d",curr->host,curr->port);
  return 0;
} /* macs_upc_add_ups */

macs_upc_sock *new_upc_sock(pool *p) {
  macs_upc_sock *s = (macs_upc_sock *)ap_palloc(p,sizeof(macs_upc_sock));
  s->host = NULL;
  s->port = 0;
  s->isup = 0;
  s->sock = -1;
  s->in_use = 0;
  s->next = NULL;
  return s;
} /* new_upc_sock */


/*****************************************************************
                       child startup/exit
 *****************************************************************/

/* connect to UPSs on child startup (and at need) */
void macs_upc_child_init(server_rec *server, pool *p) {
  macs_upc_cfg *c = (macs_upc_cfg *)
    ap_get_module_config(server->module_config, &macs_upc_module);
  macs_upc_sock *s;
  struct hostent *h;
  struct sockaddr_in addr;
  unsigned int host_addr;
  /* we don't get any real tmp pools, so
     we'll just do this buffer by hand */
  char *buf = (char *)malloc(sizeof(char)*MACS_UPC_BUFSIZ);
  /* I hate inheriting this... */
  errno = 0;
  /* multi-process rather than multi-threaded...  this might change
   * someday, but til then we don't want each child to connect to
   * every UPS. One is probably enough.
   *
   * FIXME: there should be a way to evenly spread which server entry
   * the children connect to. Right now, the first entry gets hit til
   * it won't accept any more, then the second, etc.
   */
  for (s = c->socks; s; s = s->next) {
    if ( s->in_use ) {
      /* in use means ok */
      ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, server,
		   "UPS %s:%i in use",s->host,s->port);
      continue;
    }
    s->in_use = 1;
    errno = 0;
    /* let's just ping and see if we get a response */
    if ( s->isup && ! ( write(s->sock,"PING\n",5) == 5   &&
			read(s->sock,buf,MACS_UPC_BUFSIZ)     &&
			strncasecmp(buf,"PONG\n",5) == 0 ) ) {
      ap_log_error(APLOG_MARK, APLOG_DEBUG, server,
		   "UPS %s:%i down", s->host, s->port);
      s->isup = 0;
    } else if ( s->isup ) {
      ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, server,
		   "Already connected to UPS %s:%i",s->host,s->port);
      s->in_use = 0;
      /* this shouldn't happen in multi-process mode */
      break;
    }
    /* if down, bring back up */
    close(s->sock);
    if ( !(h = gethostbyname(s->host)) ) {
      ap_log_error(APLOG_MARK, APLOG_WARNING, server,
		   "Couldn't get hostname for %s", s->host);
      s->in_use = 0;
      continue;
    }
    host_addr = *(unsigned int *)h->h_addr;
    addr.sin_family = AF_INET;
    addr.sin_port = htons(s->port);
    addr.sin_addr.s_addr = host_addr;
    s->sock = socket(AF_INET,SOCK_STREAM,IPPROTO_TCP);
    if ( s->sock < 0 ) {
      ap_log_error(APLOG_MARK, APLOG_WARNING, server,
		   "Couldn't open socket to %s:%i", s->host,s->port);
      s->in_use = 0;
      continue;
    }
    if ( connect(s->sock,(struct sockaddr *)&addr,sizeof(addr)) < 0 ) {
      ap_log_error(APLOG_MARK, APLOG_DEBUG, server,
		   "Failed to connect to UPS %s:%i", s->host, s->port);
      close(s->sock);
      s->in_use = 0;
      continue;
    }
    /* let's just ping and see if we get a response */
    if ( s->isup && ! ( write(s->sock,"PING\n",5) == 5   &&
			read(s->sock,buf,MACS_UPC_BUFSIZ)     &&
			strncasecmp(buf,"PONG\n",5) == 0 ) ) {
      ap_log_error(APLOG_MARK, APLOG_WARNING, server,
		   "Communication with UPS %s:%i failed", s->host, s->port);
      s->in_use = 0;
      continue;
    } else {
      ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, server,
		   "Connected to UPS %s:%i", s->host, s->port);
      s->isup = 1;
      s->in_use = 0;
      /* we got one, that's enough */
      break;
    }
  } /* foreach socket */
  free(buf);
  return;
} /* macs_upc_child_init */

void macs_upc_child_exit(server_rec *r, pool *p) {
  macs_upc_cfg *c = (macs_upc_cfg *)
    ap_get_module_config(r->module_config, &macs_upc_module);
  macs_upc_sock *s;
  for (s = c->socks; s; s = s->next) {
    errno = 0;
    if (s->sock)
      close(s->sock);
    ap_log_error(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		 "Disconnecting from %s:%i", s->host, s->port);
    s->in_use = 1;
    s->isup = 0;
  }
  return;
} /* macs_upc_child_exit */


/*****************************************************************
          the actual expose routines (finally)
 *****************************************************************/

int macs_upc_expose(request_rec *r) {
  macs_upc_callback_data *cd;
  macs_upc_cfg *c;
  macs_upc_dir_cfg *dc;

  c = (macs_upc_cfg *)
    ap_get_module_config(r->server->module_config, &macs_upc_module);
  cd = (macs_upc_callback_data *)
    ap_palloc(r->pool,sizeof(macs_upc_callback_data));
  cd->uid = (char *)ap_table_get(r->headers_in,c->uidhdr);
  if ( !cd->uid ) {
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		  "No MACS uid detected");
    return DECLINED;
  }
  dc = (macs_upc_dir_cfg *)
    ap_get_module_config(r->per_dir_config, &macs_upc_module);
  if ( ap_is_empty_table(dc->elts) ) {
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		  "No profile exposure configured for %s", r->uri);
    return DECLINED;
  }
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, r,
		"Exposing profile in %s for uid %s", r->uri, cd->uid);
  /* iterate through profile exposure list and expose */
  cd->buf = (char *)ap_palloc(r->pool,sizeof(char)*MACS_UPC_BUFSIZ);
  cd->r = r; cd->c = c; cd->dc = dc;
  ap_table_do(macs_upc_expose_cb,(void *)cd,dc->elts,NULL);
  return OK;
} /* macs_upc_expose */

/* callback function for profile exposure list iteration */
int macs_upc_expose_cb( void *vcd,
			const char *header,
			const char *elt ) {
  char *v, *c;
  const char *aid;
  macs_upc_callback_data *cd = (macs_upc_callback_data *)vcd;
  aid = ap_table_get(cd->dc->aids,header);
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, cd->r,
		"Exposing aid %s elt %s as %s", aid, elt, header);
  snprintf( cd->buf, MACS_UPC_BUFSIZ, "%s%s%s",
	    MACS_UPC_ELEMENT_PREFIX, aid, elt );
  /* we'll cache these in the request notes */
  /* FIXME: does this make sense? connection notes maybe... */
  if ( !(v = (char *)ap_table_get(cd->r->notes,cd->buf)) ) {
    v = macs_upc_get_element(aid,elt,cd);
    ap_table_set(cd->r->notes,cd->buf,v);
  }
  /* expose as spoofed request header, env var, spoofed cookie */
  if ( v ) {
    ap_table_add(cd->r->headers_in,header,v);
    ap_table_set(cd->r->subprocess_env,header,v);
    /* FIXME: setting the cookies this way is redundant */
    c = (char *)ap_table_get(cd->r->headers_in,"Cookie");
    ap_table_unset(cd->r->headers_in,"Cookie");
    ap_table_set(cd->r->headers_in, "Cookie",
		 ap_pstrcat(cd->r->pool,c,"; ",header,"=",v,"\0",NULL));
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, cd->r,
		  "exposed aid %s elt %s as %s", aid, elt, header);
  } else {
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, cd->r,
		  "got nothing for aid %s elt %s", aid, elt);
  }
  return 1;
} /* macs_upc_expose_cb */

char *macs_upc_get_element( const char *aid,
			    const char *elt,
			    macs_upc_callback_data *cd ) {
  macs_upc_sock *s = cd->c->socks;
  int upsvrs, rv, len;
  char *query = (char *)ap_palloc(cd->r->pool,sizeof(char)*MACS_UPC_BUFSIZ),
    *reply;
  /* create the query */
  rv = ap_snprintf(query,MACS_UPC_BUFSIZ,"GET %s %s %s\n",
		   aid, cd->uid, elt);
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, cd->r,
		"Query is %s",query);
  /* if there are servers up, keep trying them until one shows up
   * unused, then grab it
   */
  for (upsvrs=0; s||upsvrs; s=s->next) {
    if ( !s ) {
      /* start over */
      s = cd->c->socks;
      upsvrs = 0;
    }
    if ( s->isup )
      upsvrs++;
    else
      continue;
    if ( !s->in_use ) {
      s->in_use = 1;
      break;
    }
  }
  if ( !s ) {
    /* no UPS's up! */
    if ( ap_table_get(cd->r->notes,"macs_upc_childreinit") ) {
      ap_log_rerror(APLOG_MARK, APLOG_ERR | APLOG_NOERRNO, cd->r,
		    "No UPS's up after reinit!");
      return NULL;
    }
    ap_log_rerror(APLOG_MARK, APLOG_NOTICE | APLOG_NOERRNO, cd->r,
		  "No UPS's up! Reinitting");
    /* only reinit once */
    ap_table_setn(cd->r->notes,"macs_upc_childreinit","1");
    macs_upc_child_init(cd->r->server,(pool *)NULL);
    return macs_upc_get_element(aid,elt,cd);
  }

  /* send the query to the UPS */
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, cd->r,
		"sending query to ups at %s:%i", s->host, s->port);
  if ( (upsvrs=write(s->sock,query,rv)) != rv ) {
    ap_log_rerror(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, cd->r,
		  "Retry on error writing to UPS %s:%i: %s",
		  s->host, s->port, strerror(errno));
    s->isup = 0;
    s->in_use = 0;
    return macs_upc_get_element(aid,elt,cd);
  }

  /* get the length */
  rv = upsvrs = errno = 0;
  reply = (char *)ap_palloc(cd->r->pool,sizeof(char)*MACS_UPC_BUFSIZ);
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, cd->r,
		"Reading reply");
  while (1) {
    /* upsvrs := bytes read */
    upsvrs = read(s->sock,&reply[rv],MACS_UPC_BUFSIZ-rv);
    ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, cd->r,
		  "Read a chunk: %s", reply);
    if ( upsvrs > 0 ) rv += upsvrs;
    for ( len=0; reply[len] && reply[len] != '\n'; len++ );
    if ( reply[len] == '\n' ) {
      if ( strncasecmp("YES ",reply,4) ) {
	/* error string from server */
	ap_log_rerror(APLOG_MARK, APLOG_WARNING | APLOG_NOERRNO, cd->r,
		      "Error string from UPS: %s", reply);
	return NULL;
      }
      /* found length */
      reply[len] = '\0';
      break;
    }
    switch (errno) {
    case 0: /* fine, read more */
      break;
    case EAGAIN: /* temporary error condition - clear & read more */
      errno = 0;
      break;
    case ETIMEDOUT: /* timed out waiting for response */
      ap_log_rerror(APLOG_MARK, APLOG_WARNING, cd->r,
		    "Timeout waiting for UPS %s:%i", s->host, s->port);
    default: /* read error */
      ap_log_rerror(APLOG_MARK, APLOG_WARNING, cd->r,
		    "Retry on read error from UPS %s:%i", s->host, s->port);
      s->isup = 0;
      s->in_use = 0;
      return macs_upc_get_element(aid,elt,cd);
    }
  } /* read length loop */

  rv = len;
  len = atoi(&reply[4]);
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, cd->r,
		"Read lenline from ups: %s(%i)", reply, len);
  reply = &reply[rv+1];
  rv = strlen(reply);
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, cd->r,
		"reading item. have %s(%i/%i)", reply, rv, len);
  while (rv < len) {
    /* upsvrs := bytes read */
    upsvrs = read(s->sock,&reply[rv],len-rv);
    if ( upsvrs > 0 )
      rv += upsvrs;
    if ( rv == len ) {
      /* found item */
      break;
    }
    switch (errno) {
    case 0: /* fine, read more */
      break;
    case EAGAIN: /* temporary error condition - clear & read more */
      errno = 0;
      break;
    case ETIMEDOUT: /* timed out waiting for response */
      ap_log_rerror(APLOG_MARK, APLOG_WARNING, cd->r,
		    "Timeout waiting for UPS %s:%i", s->host, s->port);
    default: /* read error */
      ap_log_rerror(APLOG_MARK, APLOG_WARNING, cd->r,
		    "Retry on read error from UPS %s:%i", s->host, s->port);
      s->isup = 0;
      s->in_use = 0;
      return macs_upc_get_element(aid,elt,cd);
    }
  } /* read item loop */
  reply[rv]='\0';
  s->in_use = 0;
  ap_log_rerror(APLOG_MARK, APLOG_DEBUG | APLOG_NOERRNO, cd->r,
		"Got %s's %s %s: %s",cd->uid,aid,elt,reply);
  return reply;

} /* macs_upc_get_element */
